Package: srcpkgs
Title: R Source Packages Manager
Version: 0.2.1
Authors@R: 
    person("Karl", "Forner", , "karl.forner@gmail.com", role = c("aut", "cre", "cph"))
Description: Manage a collection/library of R source packages. Discover, document, load, test 
  source packages. Enable to use those packages as if they were actually installed. Quickly reload
  only what is needed on source code change. Run tests and checks in parallel.
License: GPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.3
URL: https://github.com/kforner/srcpkgs
BugReports: https://github.com/kforner/srcpkgs/issues
Imports: cli, clitable, devtools, pkgload, testthat, stats, utils
Suggests: knitr, rmarkdown, withr
Config/testthat/edition: 3
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-20 11:01:33 UTC; vscode
Author: Karl Forner [aut, cre, cph]
Maintainer: Karl Forner <karl.forner@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-20 13:00:02 UTC
Built: R 4.5.2; ; 2026-04-20 15:06:55 UTC; unix

Package: srcpkgs
Title: R Source Packages Manager
Version: 0.2.1
Authors@R: 
    person("Karl", "Forner", , "karl.forner@gmail.com", role = c("aut", "cre", "cph"))
Description: Manage a collection/library of R source packages. Discover, document, load, test 
  source packages. Enable to use those packages as if they were actually installed. Quickly reload
  only what is needed on source code change. Run tests and checks in parallel.
License: GPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.3
URL: https://github.com/kforner/srcpkgs
BugReports: https://github.com/kforner/srcpkgs/issues
Imports: cli, clitable, devtools, pkgload, testthat, stats, utils
Suggests: knitr, rmarkdown, withr
Config/testthat/edition: 3
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-20 11:01:33 UTC; vscode
Author: Karl Forner [aut, cre, cph]
Maintainer: Karl Forner <karl.forner@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-20 13:00:02 UTC
Built: R 4.5.2; ; 2026-04-20 23:46:11 UTC; unix

Package: dummy.srcpkg
Title: Dummy Srcpkg To Use in Examples
Description: A dummy srcpkg to use in examples.
Version: 3.1416
Authors@R: person("Karl", "Forner", , "karl.forner@gmail.com", role = c("aut", "cre"))
License: GPL-2
Encoding: UTF-8
Suggests: 
  testthat

Package: elves
Title: elves dummy package for scrpkgs
Description: elves dummy package for scrpkgs.
Version: 0.1
Authors@R: person("Karl", "Forner", , "karl.forner@gmail.com", role = c("aut", "cre"))
Imports: galadriel, legolas
License: MIT
Encoding: UTF-8
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.3.2
Suggests: testthat

# Generated by roxygen2: do not edit by hand

S3method(as.data.frame,pkg_check)
S3method(as.data.frame,pkg_test)
S3method(as.data.frame,pkgs_check)
S3method(as.data.frame,pkgs_test)
S3method(as.data.frame,srcpkgs)
S3method(as.logical,pkg_check)
S3method(as.logical,pkg_test)
S3method(as.logical,pkgs_check)
S3method(as.logical,pkgs_test)
S3method(print,pkg_test)
S3method(print,pkgs_check)
S3method(print,pkgs_test)
S3method(print,srcpkg)
S3method(print,srcpkgs)
S3method(summary,pkg_check)
S3method(summary,pkg_test)
S3method(summary,pkgs_check)
S3method(summary,pkgs_test)
export(find_srcpkgs)
export(get_srcpkgs)
export(hack_r_loaders)
export(pkg_check)
export(pkg_list_attached)
export(pkg_load)
export(pkg_roxygenise)
export(pkg_test)
export(pkg_unload)
export(pkgs_check)
export(pkgs_deps)
export(pkgs_install)
export(pkgs_test)
export(reset)
export(settings)
export(setup_and_get_dummy_srcpkg)
export(srcpkgs)
export(unhack_r_loaders)
importFrom(cli,cli_text)
importFrom(clitable,cli_table)
importFrom(testthat,test_dir)
importFrom(utils,getFromNamespace)

srcpkgs
=======

<!-- badges: start -->
[![R-CMD-check](https://github.com/kforner/srcpkgs/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/kforner/srcpkgs/actions/workflows/R-CMD-check.yaml)
[![Codecov test coverage](https://codecov.io/gh/kforner/srcpkgs/branch/main/graph/badge.svg)](https://app.codecov.io/gh/kforner/srcpkgs?branch=main)
[![CRAN status](https://www.r-pkg.org/badges/version/srcpkgs)](https://CRAN.R-project.org/package=srcpkgs)
<!-- badges: end -->

The aim of `srcpkgs` is to make development and use of source packages easy, and thus encourage to  continuously develop a shared library of reusable code. 

## Synopsis

Developing a library of packages can be really tedious, especially if there are dependencies between those packages. Each time you change a package, you have to **install** it before being able to use it.

The advent of [devtools](https://devtools.r-lib.org/) enabled the use of source packages directly, bypassing the installation step, which considerably streamlined the development of ONE package, also making it possible to reload a package within a R session.

`srcpkgs` *extends* devtools to manage the development and use of a  collection or library of source packages as easily and efficiently as possible.
In particular, it understands the dependencies between those packages and upon changes, determines the optimal
course of action to implement them.

See the vignettes:

-  [Why would you need srcpkgs?](https://kforner.github.io/srcpkgs/articles/demo.html)
- [Getting Started](https://kforner.github.io/srcpkgs/articles/getting_started.html)
- [Testing and checking packages with srcpkgs](https://kforner.github.io/srcpkgs/articles/testing_and_checking.html))

## intended target

The main use case for `srcpkgs` is for an organization with many R users, that wants to share and reuses code across projects and users. This reusable code will be implemented as a collection of R source packages. The project code itself should also be implemented as a source package. 
This way: 
- each project has it own version of the collection
- the project devs can easily adapt the collection for their project (add features, fix bugs) in the same R session used for the project
- at any time, the project devs can **push** their changes to the collection


The recommended setup is:

  - one git repository for the (interdependent) collection of source packages
    * of course you could have a git repo per source package, but if you have dozends of them it's cumbersome to add them as git sub modules to all your projects. Moreover it's harder to select compatible versions.
  - one git repository for a project
    * with the collection add the the project as a **git submodule**


This can also be useful for a single developer. This will allow to easily reuse code across projects, avoiding the duplication of code. 

**N.B**: wrapping the code as a source package is for convenience, and is really easy. It's useful even if you have no plan of releasing or distributing the package. 


## Features

- fast and automatic **discovery** of the source packages 
- fast and automatic detection of **changes**: quickly unload/reload/load only what is needed.
- apply roxygen ([cf roxygen2](https://roxygen2.r-lib.org/)) automatically and only when needed
- can **hack** the R loaders (`library()` and `loadNamespace()`) to use `srcpkgs` in development, 
  so that your scripts still work if you install your packages (e.g. in *Production* mode)
- **testing** the whole collection at once, and output an aggregated report (using [testthat](https://testthat.r-lib.org/) of course).
- same for `R CMD check`ing the collection
- extensively tested (~ 450 tests) and 100% test coverage

### forthcoming

- testing and checking the collection in **parallel**
- test coverage of the collection


## Installation

The package is now available on CRAN!

Install if from CRAN:
```
install.packages('srcpkgs') 
```

Install it from github: 
```
### using devtools
# install devtools from CRAN if needed
install.packages('devtools') 

install_github('kforner/srcpkgs')

### or using pak
# install pak from CRAN if needed
install.packages('pak') 
pak::pak("kforner/srcpkgs")
```

## developer notes

### how to quickly test on R-devel

```
docker run --rm -ti -v $(pwd):/$(pwd) -w $(pwd) rocker/tidyverse:devel bash

R

library(devtools)
load_all()
check()

```

demo library `LOTR` to exemplify srcpkgs
=========================================

This dummy library of R source packages to exemplify the R package 
[srcpkgs](https://github.com/kforner/srcpkgs)

# prerequisites

install the `srcpkgs` package! (cf https://github.com/kforner/srcpkgs/blob/main/README.md)

# usage

## checkout this repository

```
cd "whereever_you_want_to_work"
git checkout https://github.com/kforner/srcpkgs_lotr_demo.git`
cd srcpkgs_lotr_demo
```

## basic usage


```
# start clean (remove generated roxygen documentation)
make clean

# run R
R

```

### using the hacked R loaders

```
library(srcpkgs)

### using hacked library
# be sure to hack the loaders.
# TODO: describe how it could be done automatically
hack_r_loaders()

library(lotr)
# everything was loaded
str(characters_by_race())

```

### using srcpkgs::pkg_load()

```
library(srcpkgs)


library(lotr)
# everything was loaded
str(characters_by_race())

```

#######################################################################
# wraps a devtools package object into a derived S3 "srcpkg" object
######################################################################

# creates a new "srcpkg" object from a devtools::package, or a path, or a srcpkg (noop)
srcpkg <- function(pkg = devtools::as.package(path), path = NULL) {
  force(pkg)
  if (inherits(pkg, 'srcpkg')) return(pkg) 
  stop_unless(devtools::is.package(pkg), 'pkg is not a devtools package object')

  # fix deps: either NULL, or with EOL and spaces/tabs
  for (dep in c('imports', 'depends', 'suggests')) {
    pkg[[dep]] <- gsub("\\s*\\s*", "", pkg[[dep]]  %||% '')  
  }


  class(pkg) <- c("srcpkg", "package")

  pkg
}



# makes sure we get a srcpkg instance
as_srcpkg <- function(x, src_pkgs = get_srcpkgs()) {
  stop_unless(length(x), 'bad arg: empty')
  if (inherits(x, 'srcpkg')) return(x)
  if (devtools::is.package(x)) return(srcpkg(x))

  stop_unless(is.character(x), 'bad arg: not an instance nor character')

  # check if it is a srcpkg name from src_pkgs
  pkg <- src_pkgs[[x]]
  if (length(pkg)) return(pkg)

  # assume it is a path
  stop_unless(dir.exists(x), 'bad arg: should be a path, but it does not exist')

  srcpkg(path = x)
}



#' @export
print.srcpkg <- function(x, ...) {
  cli::cli_text("{.pkg {x$package}}@{.url {x$version}} source package [{.file {x$path}}]")
}

# coerces a pkg object or a package name to a package name
# useful to accept multiple input type in a function
as_pkg_name <- function(pkg_or_name) {
  if (is.character(pkg_or_name)) {
    pkg_or_name
  } else {
    stop_unless(devtools::is.package(pkg_or_name), 'bad arg pkg_or_name')
    pkg_or_name$package
  }
}

test_that("srcpkg", {
  setup_temp_dir()
  srcpkg <- pkg_create('.', 'minimal')
  ###
  expect_s3_class(srcpkg, 'srcpkg')
  expect_true(devtools::is.package(srcpkg))

  # works also directly on a srcpkg
  srcpkg2 <- srcpkg(srcpkg)
  expect_identical(srcpkg2, srcpkg)

  # with a path
  srcpkg2 <- srcpkg(path = srcpkg$path)
  expect_identical(srcpkg2, srcpkg)

  ### with dependencies
  pkg <- pkg_create('.', 'withdeps', imports = c("i1", "i2"), depends = "d1", suggests = c("s1", "s2"))

  srcpkg <- srcpkg(pkg)

  .comma <- function(x) paste0(x, collapse = ",")
  expect_identical(srcpkg$imports, .comma(c("i1", "i2")))
  expect_identical(srcpkg$depends, .comma("d1"))
  expect_identical(srcpkg$suggests, .comma(c("s1", "s2")))
})


test_that("as_srcpkg", {
  setup_temp_dir()
  srcpkg <- pkg_create('.', 'minimal')

  ### empty
  expect_error(as_srcpkg(NULL), 'empty')
  expect_error(as_srcpkg(character()), 'empty')

  expect_error(as_srcpkg(1), 'bad arg')

  ### srcpkg
  expect_identical(as_srcpkg(srcpkg), srcpkg)

  ### "package" object
  pkg <- devtools::as.package(srcpkg$path)
  expect_identical(as_srcpkg(pkg), srcpkg)

  ### srcpkg name
  src_pkgs <- srcpkgs(list(srcpkg))
  expect_identical(as_srcpkg('minimal', src_pkgs = src_pkgs), srcpkg)

  ### path
  expect_identical(as_srcpkg(srcpkg$path, src_pkgs = src_pkgs), srcpkg)
})

test_that("print.srcpkg", {
  setup_temp_dir()
  srcpkg <- pkg_create('.', 'minimal')

  out <- capture.output(print(srcpkg), type = 'message')
  expect_match(out, "minimal")
})

test_that("as_pkg_name", {
  setup_temp_dir()
  srcpkg <- pkg_create('.', 'minimal')

  # from name
  expect_identical(as_pkg_name('AA'), 'AA')

  # from devtools package
  pkg <- devtools::as.package(srcpkg$path)
  expect_identical(as_pkg_name(pkg), srcpkg$package)

  # from srcpkg
  expect_identical(as_pkg_name(srcpkg), srcpkg$package)

  # from something else
  expect_error(as_pkg_name(1),'bad arg')
})

#' @section Features:
#'
#' `srcpkgs` main objective is to ease development on any project
#' that uses a collection of R source packages (a library).
#' It is able to figure out which dependencies are source packages, and is able
#' to quickly detect changes in any of the used source packages.
#'

#'
"_PACKAGE"

NULL

LOADERS_HACK_KEY <- 'LOADERS_HACK'

set_loaders_hack <- function(status) {
  set_config(LOADERS_HACK_KEY, status)
}
is_loaders_hack_enabled <- function() {
  isTRUE(get_config(LOADERS_HACK_KEY))
}

is_loaders_hack_installed <- function() {
  is_traced(library)
}

is_traced <- function(fun) {
  inherits(fun, "functionWithTrace")
}

INHIBIT_ENV_VAR <- 'SRCPKGS.INHIBIT_R_LOADERS_HACK'
INHIBIT_OPTION <- tolower(INHIBIT_ENV_VAR)

inhibit_r_loaders_hack <- function() {
  raw <- get_env(INHIBIT_ENV_VAR, getOption(INHIBIT_OPTION, FALSE))

  #### may be character, numeric or logical
  # more annoying case: "1" and "0"
  if (is.character(raw) && is.na(as.logical(raw)) && !is.na(as.integer(raw))) {
    raw <- as.integer(raw)
  }

  isTRUE(as.logical(raw))
}


#' instruments the R loaders to make them aware of source packages
#'
#' hacks `library()` and `loadNamespace()` using the base R tracer function `trace()`.
#' `library(pkg)` will basically call `pkg_load(pkg)` if the source package `pkg` 
#' is managed by **srcpkgs**
#'
#' N.B: usually you do not need to call that function explicitly. The function is reentrant.
#' 
#' @section Package startup:
#' 
#' At package startup (actually `.OnAttach()`), `hack_r_loaders()` will be automatically called to hack
#' the R loaders, UNLESS this is inhibited via the option `srcpkgs.inhibit_r_loaders_hack` or the 
#' environment variable `SRCPKGS.INHIBIT_R_LOADERS_HACK`. You may set any value like TRUE, "TRUE", 1 or "1".
#'
#' @return no return value, called for side-effects
#' @export
#' @importFrom utils   getFromNamespace
#' @seealso [unhack_r_loaders()]
#' @examples
#' \dontrun{
#' # hack library
#' hack_r_loaders()
#' 
#' # unhack
#' unhack_r_loaders()
#' 
#' # prevent automatic hacking when srcpkgs is loaded
#' options(srcpkgs.inhibit_r_loaders_hack=TRUE)
#' # or
#' Sys.setenv(SRCPKGS.INHIBIT_R_LOADERS_HACK="1")
#' library(srcpkgs)
#' }
hack_r_loaders <- function() {
  hack_library()
  hack_loadNamespace()
  set_loaders_hack(TRUE)
  invisible()
}

#' untraces library() and loadNamespace()
#'
#' The function is reentrant.
#' @return no return value, called for side-effects
#' @export
#' @seealso [hack_r_loaders()]
#' @examples
#' \dontrun{
#' unhack_r_loaders()
#' }
unhack_r_loaders <- function() {
  suppressMessages(untrace(library))
  suppressMessages(untrace(loadNamespace))
  set_loaders_hack(FALSE)
}


hack_library <- function() {
  tracer <- quote({
      # if ('srcpkgs' %in% loadedNamespaces()) { # in case srcpkgs is unloaded after hacking library
        ### fool R CMD check which does not like :::
        get_from_srcpkgs <- function(x) getFromNamespace(x, 'srcpkgs')
        enabled <- get_from_srcpkgs('is_loaders_hack_enabled')()
        if (enabled) {
          set_loaders_hack <- get_from_srcpkgs('set_loaders_hack')
          # N.B: we disable the hack temporarily to avoid recursive calls
          # e.g. library(pkg1) ---> library(depends1), library(depends2)
          # everything we need will be done by this hack. so we do not need to subsequent calls
          # to trigger that hack and its overhead for nothing
    
          set_loaders_hack(FALSE)
          on.exit(set_loaders_hack(TRUE), add = TRUE)

          # figure out the library() function arguments
          if (!missing(package)) {
            if (!character.only)
              pkg <- as.character(substitute(package))
            else
              pkg <- package
          
            src_pkgs <- get_srcpkgs() # may take some I/O time
            src_pkg <- src_pkgs[[pkg]]
            if (length(src_pkg)) {
              ### this is actually a source package that we manage, so we take care of loading it
              pkg_load(src_pkg, src_pkgs, quiet = TRUE)
            }
          }
        }
    # }
  })

  suppressMessages(trace(library, tracer, print = FALSE))
}

hack_loadNamespace <- function() {
  tracer <- quote({
      # if ('srcpkgs' %in% loadedNamespaces()) { # in case srcpkgs is unloaded after hacking library
        ### fool R CMD check which does not like :::
        get_from_srcpkgs <- function(x) getFromNamespace(x, 'srcpkgs')
        enabled <- get_from_srcpkgs('is_loaders_hack_enabled')()
        if (enabled) {
          set_loaders_hack <- get_from_srcpkgs('set_loaders_hack')
          # N.B: we disable the hack temporarily to avoid recursive calls
          # e.g. loadNamespace(pkg1) ---> loadNamespace(imp1), loadNamespace(imp2)
          # everything we need will be done by this hack. so we do not need to subsequent calls
          # to trigger that hack and its overhead for nothing
    
          set_loaders_hack(FALSE)
          on.exit(set_loaders_hack(TRUE), add = TRUE)

          # figure out the loadNamespace() function arguments
          pkg <- as.character(package)[[1L]] # sometimes contain the version as element #2 ?!

          src_pkgs <- get_srcpkgs() # may take som I/O time
          src_pkg <- src_pkgs[[pkg]]
          if (length(src_pkg)) {
            ### this is actually a source package that we manage, so we take care of loading it
            pkg_load(src_pkg, src_pkgs, attach = FALSE, quiet = TRUE)
          }
        }
    # }
  })

  suppressMessages(trace(loadNamespace, tracer, print = FALSE))
}

.pkg_test_errors <- 
test_that(".pkg_test_errors", {
  src_pkgs <- examples_srcpkgs_basic()
  add_dummy_test_to_srcpkgs(src_pkgs)

  ### an exception in the setup
  pkga <- src_pkgs$AA
  writeLines(r"{ stop("aie aie aie") }", file.path(pkga$path, "tests/testthat/setup.R"))
 
  expect_error(pkg_test('AA', src_pkgs = src_pkgs, reporter = "silent"), "aie aie aie")
})

.pkg_test <- 
test_that("pkg_test", {
  ### trivial example A->B
  src_pkgs <- examples_srcpkgs_basic()

  ### pkg with no tests
  is_empty_pkg_test <- function(x) inherits(x, "pkg_test") && is.list(x) && length(x) == 0

  res <- pkg_test('AA', src_pkgs = src_pkgs, reporter = "silent")
  expect_true(is_empty_pkg_test(res))
  expect_true(is_empty_pkg_test(pkg_test('BB', src_pkgs = src_pkgs, reporter = "silent")))

  # as.data.frame
  df <- as.data.frame(res)

  expect_identical(names(df), c("file", "test", "nb", "failed", "passed", "skipped", "error", "warning", "time"))
  expect_equal(nrow(df) ,0)

  # no tests --> no failure
  expect_true(as.logical(res))
  expect_match(capture.output(print(res), type = "message"), "package AA has no tests", fixed = TRUE, all = FALSE)

  sdf <- summary(res)
  expect_identical(names(sdf), c("file", "nb", "failed", "passed", "skipped", "error", "warning", "time"))
  expect_equal(nrow(sdf) ,0)

  ###### with tests
  add_dummy_test_to_srcpkgs(src_pkgs)
  ### one package with a dep

  resAA <- res <- pkg_test('AA', src_pkgs = src_pkgs, reporter = "silent")

  expect_s3_class(res, "pkg_test")
  df <- as.data.frame(res)
  expect_equal(sum(df$failed), 6)
  pkg <- attr(res, 'pkg')
  expect_identical(pkg, src_pkgs$AA)

  ### pkg with no deps
  res <- pkg_test('BB', src_pkgs = src_pkgs, reporter = "silent")
  expect_s3_class(res, "pkg_test")

  ####### filter
  res <- pkg_test('AA', filter = "success", src_pkgs = src_pkgs, reporter = "silent")
  df <- as.data.frame(res)
  expect_equal(sum(df$failed), 0)

  res <- pkg_test('AA', filter = "failure", src_pkgs = src_pkgs, reporter = "silent")
  df <- as.data.frame(res)
  expect_true(all(df$failed > 0))

  ### if loaders_hack_enabled
  if (! is_loaders_hack_enabled()) {
    hack_r_loaders()
    on.exit(unhack_r_loaders(), add = TRUE)
  }
  res <- pkg_test('AA', src_pkgs = src_pkgs, reporter = "silent")
  expect_identical(as.data.frame(fix_test_result_timings(res)), as.data.frame(fix_test_result_timings(resAA)))
})



.pkg_test_s3_methods <- 
test_that("pkg_test_s3_methods", {
  local_reproducible_output(crayon = TRUE, unicode = TRUE)

  src_pkgs <- examples_srcpkgs_complex_deps()
  add_dummy_test_to_srcpkgs(src_pkgs)

  res <- pkg_test('AA', src_pkgs = src_pkgs, reporter = "silent")

  ### as.data.frame
  # fetch original data frame from testthat
  tr <- res
  class(tr) <- class(res)[-1]
  df0 <- as.data.frame(tr)

  df <- as.data.frame(res)

  # agree on common cols
  common_cols <- setdiff(intersect(names(df0), names(df)), "file")
  expect_identical(df0[common_cols], df[common_cols])

  # extra cols
  extra_cols <- setdiff(names(df), names(df0))
  expect_true("time" %in% extra_cols)


  
  ### summary
  df <- as.data.frame(res)
  # by file (default)
  sdf <- summary(res)
  expect_setequal(names(sdf), setdiff(names(df), "test"))

  expect_equal(anyDuplicated(sdf$file), 0)
  expect_equal(sum(sdf$nb), sum(df$nb))
  expect_equal(sum(sdf$failed), sum(df$failed))

  # overview: col = NULL
  sdf <- summary(res, col = NULL)
  
  expect_equal(nrow(sdf), 1)
  expect_setequal(names(sdf), setdiff(names(df), c("test", "file")))
  expect_equal(sum(sdf$nb), sum(df$nb))
  expect_equal(sum(sdf$failed), sum(df$failed))

  ### as.logical
  expect_false(as.logical(res))

  # keep only successful tests
  good <- res[df$failed == 0 & !df$error]
  class(good) <- class(res)

  expect_true(as.logical(good))

  ### print
  withr::local_options(list(cli.num_colors = 256))
  expect_snapshot(print(fix_test_result_timings(res)))
})

.untrace <- function(...) {
  suppressMessages(untrace(...))
}

test_that("hack_r_loaders", {

  ### try to restore state
  old <- set_loaders_hack(TRUE)
  on.exit(set_loaders_hack(old), add = TRUE)
  TRACED <- is_traced(library)
  if (TRACED) on.exit(hack_r_loaders(), add = TRUE) else on.exit(unhack_r_loaders(), add = TRUE)

  ### hack_r_loaders
  hack_r_loaders()

  expect_true(is_traced(library))
  expect_true(is_traced(loadNamespace))
  expect_true(is_loaders_hack_enabled())

  ### unhack_r_loaders
  unhack_r_loaders()
  
  expect_false(is_traced(library))
  expect_false(is_traced(loadNamespace))
  expect_false(is_loaders_hack_enabled())

})


test_that("hack_library", {
  setup_temp_dir()
  pkg_name <- 'hack.library'
  pkg_create('.', pkg_name)


  ### try to restore state
  on.exit(unhack_r_loaders(), add = TRUE)
  # old <- set_loaders_hack(TRUE)
  # on.exit(set_loaders_hack(old), add = TRUE)
  # TRACED <- is_traced(library)
  # if (TRACED) on.exit(hack_library(), add = TRUE) else on.exit(.untrace(library), add = TRUE)
  paths <- find_srcpkgs_paths('.')
  PATHS <- set_srcpkgs_paths(paths)
  on.exit(set_srcpkgs_paths(PATHS), add = TRUE) 

  ### default library() behaviour --> does NOT know about our source packages 
  pkg_unload(pkg_name, quiet = TRUE)
  .untrace(library)

  expect_error(library(hack.library), 'there is no package called')

  ### now hack it

  hack_library()
  set_loaders_hack(TRUE)

  expect_error(library(hack.library), NA)
  expect_true(pkg_is_attached(pkg_name))

  ### reentrant
  expect_error(library(pkg_name, character.only = TRUE), NA)

  ### let's modify it a bit: add a new function/symbol
  writeLines('new_fun <- function(x) { x + 1}', file.path(pkg_name, 'R/new_fun.R'))

  expect_error(library(hack.library), NA)
  expect_true(is.function(utils::getFromNamespace('new_fun', pkg_name)))

  ### set_loaders_hack / is_loaders_hack_enabled
  expect_true(is_loaders_hack_enabled())
  set_loaders_hack(FALSE)
  expect_false(is_loaders_hack_enabled())

  # unload
  pkg_unload(pkg_name, quiet = TRUE)
  # suddenly library() does not load our source package anymore
  expect_error(library(pkg_name, character.only = TRUE), 'there is no package called')
})

test_that("hack_loadNamespace", {
  setup_temp_dir()
  pkg_name <- 'hack.loadNamespace'
  pkg_create('.', pkg_name)

  ### try to restore state
  old <- set_loaders_hack(TRUE)
  on.exit(set_loaders_hack(old), add = TRUE)
  TRACED <- is_traced(loadNamespace)
  if (TRACED) on.exit(hack_loadNamespace(), add = TRUE) else on.exit(.untrace(loadNamespace), add = TRUE)
  paths <- find_srcpkgs_paths('.')
  PATHS <- set_srcpkgs_paths(paths)
  on.exit(set_srcpkgs_paths(PATHS), add = TRUE) 

  ### default library() behaviour --> does NOT know about our source packages 
  pkg_unload(pkg_name, quiet = TRUE)
  .untrace(loadNamespace)

  expect_error(loadNamespace(pkg_name), 'there is no package called')

  ### now hack it
  hack_loadNamespace()

  expect_error(loadNamespace(pkg_name), NA)
  
  expect_true(pkg_is_loaded(pkg_name))
  expect_false(pkg_is_attached(pkg_name))
  ### reentrant
  expect_error(loadNamespace(pkg_name), NA)

  ### let's modify it a bit: add a new function/symbol
  writeLines('new_fun <- function(x) { x + 1}', file.path(pkg_name, 'R/new_fun.R'))

  expect_error(loadNamespace(pkg_name), NA)
  expect_true(is.function(utils::getFromNamespace('new_fun', pkg_name)))

  ### unload
  pkg_unload(pkg_name, quiet = TRUE)
  set_loaders_hack(FALSE)
  # suddenly library() does not load our source package anymore
  expect_error(loadNamespace(pkg_name), 'there is no package called')
})


test_that("inhibit_r_loaders_hack", {

  ### try to restore state
  old_env <- get_env(INHIBIT_ENV_VAR)
  old_option <- getOption(INHIBIT_OPTION)
  on.exit({set_env(INHIBIT_ENV_VAR, old_env); set_option(INHIBIT_OPTION, old_option)}, add = TRUE)

  ### defaults
  set_env(INHIBIT_ENV_VAR, '')
  set_option(INHIBIT_OPTION, NULL)

  expect_false(inhibit_r_loaders_hack())

  ### option
  # TRUE
  set_option(INHIBIT_OPTION, 'TRUE')
  expect_true(inhibit_r_loaders_hack())

  set_option(INHIBIT_OPTION, 1)
  expect_true(inhibit_r_loaders_hack())

  # FALSE
  set_option(INHIBIT_OPTION, 'FALSE')
  expect_false(inhibit_r_loaders_hack())

  set_option(INHIBIT_OPTION, 0)
  expect_false(inhibit_r_loaders_hack())

  set_option(INHIBIT_OPTION, '')
  expect_false(inhibit_r_loaders_hack())

  ### env
  set_option(INHIBIT_OPTION, '')
  # TRUE
  set_env(INHIBIT_ENV_VAR, 'TRUE')
  expect_true(inhibit_r_loaders_hack())

  set_env(INHIBIT_ENV_VAR, '1')
  expect_true(inhibit_r_loaders_hack())

  # FALSE
  set_env(INHIBIT_ENV_VAR, 'FALSE')
  expect_false(inhibit_r_loaders_hack())

  set_env(INHIBIT_ENV_VAR, '0')
  expect_false(inhibit_r_loaders_hack())

  set_env(INHIBIT_ENV_VAR, '')
  expect_false(inhibit_r_loaders_hack())
})

#' tests a list of source packages
#'
#' @inheritParams params
#' @param filter  filter out the packages to test using this pattern
#' @param ...   passed to `pkg_test`
#' @return the results as a `pkgs_test` object
#' @export
#' @examples
#' ## create a dummy collection of srcpkgs by replicating the dummy srcpkg
#'  pkg <- setup_and_get_dummy_srcpkg()
#'  pkgs <- srcpkgs(list(pkg, pkg))
#' 
#'  res <- pkgs_test(pkgs, error_on = "never")
#'  print(res)
pkgs_test <- function(pkgids = names(filter_srcpkgs(src_pkgs, filter)), src_pkgs = get_srcpkgs(), filter = NULL, quiet = TRUE, ...)
{
  force(src_pkgs)
  if (!length(pkgids)) stop('No package to test')
  pkgs <- as_srcpkgs(pkgids, src_pkgs)

  .test_pkg <- function(pkg) {
    try(pkg_test(pkg, src_pkgs = src_pkgs, quiet = quiet, ...), silent = quiet)
  }
  lst <- lapply(pkgs, .test_pkg)
  class(lst) <- 'pkgs_test'
  invisible(lst)
}

#' @export
as.data.frame.pkgs_test <- function(x, ...) {
  .row <- function(test_res) {
    if (!length(test_res)) {
      return(data.frame(nb = 0L, failed = 0L, passed = 0L, skipped = 0L, error = 0L, warning = 0L, time = 0))
    }

    if (is_error(test_res)) {
      # trick to create an empty data frame like test_results

      dummy_res <- testthat::ListReporter$new()$get_results()
      row <- as.data.frame(dummy_res)
      row[1,] <- NA

      row <- fortify_pkg_test(row)
      row$error <- 1
      row$file <- row$test <- NULL
      return(row)
    }
 
    summary(test_res, col = NULL)
  }
  rows <- lapply(x, .row)
  df <- do.call(rbind, rows)

  ## add package column
  df <- cbind(package = names(x), df, stringsAsFactors = FALSE)

  df
}

#' @export
as.logical.pkgs_test <- function(x, ...) {
   df <- as.data.frame(x)
   all(df$error == 0) && all(df$failed == 0)
}

#' @export
summary.pkgs_test <- function(object, ...) {
  df <- as.data.frame(object)
  .sum_rm_na <- function(...) sum(..., na.rm = TRUE)
  df$package <- 1 # N.B: package column is converted to counts
  sdf <- stats::aggregate(df, by = list(.dummy = rep(TRUE, nrow(df))), .sum_rm_na)
  sdf$.dummy <- NULL
 
  sdf
}

#' @export
print.pkgs_test <- function(x, ...) {
  df <- as.data.frame(x)

  # remove NAs to avoid warning in print_text_table(): replace them by 0
  df$time[is.na(df$time)] <- 0

  # only keep ms
  df$time <- format(df$time, digits = 3)

  ### by package
  bad <- which(df$failed > 0 | df$error > 0)

  print_text_table(df, 
    title = "Test results by package",
    hilite_rows = bad, heatmap_columns = 'time')

  ### overview
  sdf <- summary(x)
  sdf$time <- format(sdf$time, digits = 3)

  print_text_table(sdf, title = "Test results overview")

  ### overall result (parsable)
  cat('\n')
  ok <- as.logical(x)
  if (ok) cat("SUCCESS\n") else cat("FAILED\n")

  invisible()
}

## ----setup, include = FALSE---------------------------------------------------
## N.B: we need to setup and make available the demo collection of packages "srcpkgs_lotr_demo"
# setup a temp dir
root_dir <- tempfile()
dir.create(root_dir)
file.copy('srcpkgs_lotr_demo', root_dir, recursive = TRUE)
workdir <- file.path(root_dir, 'srcpkgs_lotr_demo')
knitr::opts_knit$set(root.dir = workdir) # set directory for other chunks
withr::local_envvar(R_USER_CACHE_DIR = tempfile())

## ----srcpkgs_setup, results = "markup"----------------------------------------

library(srcpkgs)
reset(root_dir)
print(names(get_srcpkgs()))
# cat(clitable::cli_table(as.data.frame(get_srcpkgs())), sep = "\n")

## ----no_tests, results = "markup"---------------------------------------------
print(pkgs_test(reporter = "silent"))

## ----dummy_tests--------------------------------------------------------------
add_dummy_test_to_srcpkg <- function(srcpkg, with_failures = TRUE, with_errors = TRUE, with_warnings = TRUE) {
  withr::local_dir(srcpkg$path)
  dir.create("tests/testthat", recursive = TRUE, showWarnings = FALSE)

  .write_test <- function(name, code, test = name) {
    writeLines(sprintf(r"-----{
    test_that("%s", {
      %s
    })
    }-----", name, code), sprintf("tests/testthat/test-%s.R", test))
  }

  .write_test("success", "expect_true(TRUE)")
  if (with_failures) {
    .write_test("failure", "expect_true(FALSE)")
    .write_test("mixed", "expect_true(FALSE);expect_true(TRUE)")
  }
  .write_test("skip", 'skip("skipping");expect_true(FALSE)')
  if (with_errors) .write_test("errors", 'expect_true(TRUE);stop("Arghh");expect_true(TRUE)')
  if (with_warnings)  .write_test("warning", 'expect_true(FALSE);warning("watch out");expect_true(FALSE)')
  if (with_failures && with_errors)
    writeLines(r"-----{
    test_that("misc1", {
      expect_true(FALSE)
      expect_true(TRUE)
    })
    test_that("misc2", {
      expect_true(FALSE)
      skip("skipping")
    })
    test_that("misc3", {
      expect_true(TRUE)
      expect_true(TRUE)
    })
    test_that("misc4", {
      expect_true(TRUE)
      warning("fais gaffe")
      stop("aie")
      expect_true(TRUE)
    })
    }-----", "tests/testthat/test-misc.R")

  writeLines(sprintf(r"-----{
    library(testthat)
    library(%s)

    test_check("%s")
  }-----", srcpkg$package, srcpkg$package), "tests/testthat.R")
}
i <- 0
for (pkg in get_srcpkgs()) {
  add_dummy_test_to_srcpkg(pkg, i %% 3 == 1, i %% 7 == 1, i %% 5 == 1)
  i <- i + 1
}

## ----with_tests, results = "markup"-------------------------------------------
# N.B: we use the silent testthat reporter because we only want to get the results as tables
test_results <- pkgs_test(reporter = "silent")
print(test_results)

## ----pkgs_test_methods, results = "markup"------------------------------------
print(as.data.frame(test_results))
print(summary(test_results))
print(as.logical(test_results))

print(test_results$bilbo)
print(as.data.frame(test_results$lotr))
print(summary(test_results$lotr))
print(as.logical(test_results$aragorn))

## ----subset-------------------------------------------------------------------
small_collection <- get_srcpkgs(filter = "elves|galadriel|legolas")

## ----fixing-------------------------------------------------------------------
.fix_description <- function(path, lst) {
  df <- read.dcf(path, all = TRUE)
  df2 <- utils::modifyList(df, lst)
  write.dcf(df2, path)
}
for (pkg in small_collection) {
  .fix_description(file.path(pkg$path, "DESCRIPTION"), list(Suggests = "testthat"))
}

## ----checking, results = "markup"---------------------------------------------
check_results <- pkgs_check(small_collection, quiet = TRUE)
print(check_results)

## ----pkgs_check_methods, results = "markup"-----------------------------------
print(as.data.frame(check_results))
print(summary(check_results))
print(as.logical(check_results))

print(check_results$bilbo)
print(as.data.frame(check_results$lotr))
print(summary(check_results$lotr))
print(as.logical(check_results$aragorn))

.pkgs_test_when_no_tests <- 
test_that("pkgs_test_when_no_tests", {
  src_pkgs <- examples_srcpkgs_basic()

  ### edge cases
  expect_error(pkgs_test(NULL), 'No package to test')

  ### no tests at all
  res <- pkgs_test(src_pkgs = src_pkgs, reporter = "silent", quiet = TRUE)

  expect_s3_class(res, "pkgs_test")
  expect_identical(names(res), c("AA", "BB"))

  ## all empty since no tests
  expect_s3_class(res$AA, "pkg_test")
  expect_s3_class(res$BB, "pkg_test")
  expect_length(res$AA, 0)
  expect_length(res$BB, 0)

  # as.data.frame
  df <- as.data.frame(res)

  expect_identical(df$package, c("AA", "BB"))
  expect_true(all(df[, -1] == 0))

  # as.logical
  expect_true(as.logical(res))

  # summary
  sdf <- summary(res)
  expect_equal(sdf$package, 2)
  expect_true(all(sdf[, -1] == 0))

  # print
  local_reproducible_output(crayon = TRUE, unicode = TRUE)
  expect_snapshot(print(res))

  ### only BB has tests
  add_dummy_test_to_srcpkg(src_pkgs$BB)
  
  res <- pkgs_test(src_pkgs = src_pkgs, reporter = "silent", quiet = TRUE)
  expect_s3_class(res, "pkgs_test")

  # as.data.frame
  df <- as.data.frame(res)

  expect_identical(df$package, c("AA", "BB"))
  expect_true(all(df[1, -1] == 0))
  expect_equal(df["BB", "nb"], 17)

  # as.logical
  expect_false(as.logical(res))

  # summary
  sdf <- summary(res)
  expect_equal(sdf$package, 2)
  expect_equal(sdf[, -1], df[2, -1], ignore_attr = TRUE)

  # print
  local_reproducible_output(crayon = TRUE, unicode = TRUE)
  expect_snapshot(print(fix_pkg_tests_results_timings(res)))
})


.pkgs_test <- 
test_that("pkgs_test", {
  src_pkgs <- examples_srcpkgs_basic()


  ###### with tests
  add_dummy_test_to_srcpkgs(src_pkgs)
  # make BB to fail in the tests setup
  writeLines(r"{ stop("aie aie aie") }", file.path(src_pkgs$BB$path, "tests/testthat/setup.R"))

  ### 
  res <- pkgs_test(src_pkgs = src_pkgs, reporter = "silent", quiet = TRUE)

  expect_s3_class(res, "pkgs_test")
  expect_true(is.list(res))
  expect_length(res, 2)
  expect_s3_class(res[[1]], "pkg_test")
  expect_s3_class(res[[2]], "try-error")

  res1 <- subset_s3_list(res, 1)
  res2 <- subset_s3_list(res, 2)

  ## filter
  subres <- pkgs_test(src_pkgs = src_pkgs, reporter = "silent", quiet = TRUE, filter = "A")
  expect_identical(names(subres), "AA")

  ## alternate inputs
  # names
  res2 <- pkgs_test("AA", src_pkgs = src_pkgs, reporter = "silent", quiet = TRUE)

  .cmp_res <- function(res1, res2) {
    df1 <- as.data.frame(res1)
    df2 <- as.data.frame(res2)
    df1$time <- df2$time <- NULL
    identical(df1, df2)
  }
  expect_true(.cmp_res(res2, subset_s3_list(res, "AA")))
  
  # srcpkgs objects
  resbis <- pkgs_test(src_pkgs, src_pkgs = src_pkgs, reporter = "silent", quiet = TRUE)
  expect_identical(names(resbis), c("AA", "BB"))
  
  res22 <- pkgs_test(subset_s3_list(src_pkgs, "BB"), src_pkgs = src_pkgs, reporter = "silent", quiet = TRUE)
  expect_identical(names(res22), "BB")

  ####### =============== S3 methods ================================
  res <- pkgs_test(src_pkgs = src_pkgs, reporter = "silent", quiet = TRUE)
  res1 <- subset_s3_list(res, 1)
  res2 <- subset_s3_list(res, 2)

  ### as.data.frame
  df <- as.data.frame(res)

  expect_true(is.data.frame(df))
  expect_true(all(c("package", "nb", "failed", "passed", "skipped", "error", "warning", "time") %in% names(df)))
  expect_identical(df$package, names(src_pkgs))

  expect_equal(df[1, "nb"], sum(as.data.frame(res[[1]])$nb))
  # 2nd pkg BB has crashed
  expect_equal(df[2, "error"], 1)
  other_cols <- setdiff(names(df), c("package", "error"))
  expect_true(all(is.na(df[2, other_cols, drop = TRUE])))
  # 1 pkg no error

  df1 <- as.data.frame(res1)
  df1$package <- NULL
  rownames(df1) <- NULL
  expect_identical(df1, summary(res1[[1]], col = NULL))
  
  # 1 pkg with error
  df2 <- as.data.frame(res2)
  df2$skipped <- as.integer(df2$skipped)
  expect_identical(df2, df[2,])

  ### as.logical
  expect_false(as.logical(res))
  expect_false(as.logical(res1))
  expect_false(as.logical(res2))

  # keep only successful tests
  res11_good <- res1[[1]]
  df11 <- as.data.frame(res11_good)
  res11_good <- res11_good[df11$failed == 0 & !df11$error]
  class(res11_good) <- class(res1[[1]])
  good <- list(AA = res11_good);  class(good) <- class(res)

  expect_true(as.logical(good))

  ### summary
  sdf <- summary(res)
  expect_true(is.data.frame(sdf))
  expect_equal(nrow(sdf), 1)
  
  expected <- df[1,]
  expected$error <- expected$error + 1
  rownames(expected) <- NULL
  expected$package <- 2
  expect_identical(sdf, expected)

  ### print
  local_reproducible_output(crayon = TRUE, unicode = TRUE)
  expect_snapshot(print(fix_pkg_tests_results_timings(res)))
})

#' list the available hobbits
#' @return elves names as character
#' @export
#' @importFrom galadriel galadriel
elves <- function() {
  list(
    galadriel = galadriel(), # importFrom
    legolas = legolas::legolas() # import
  )
}

examples_srcpkgs_basic <- function(.local_envir = parent.frame()) {
  # CRITICAL: must be BEFORE setup_temp_dir(), so that loaded packages are actually dangling
  withr::defer({
    # message("examples_srcpkgs_basic --> cleanup")
    cleanup_dangling_srcpkgs()
  }, .local_envir)
  setup_temp_dir(.local_envir = .local_envir)


  pkg_create('.', 'AA', imports = 'BB', suggests = 'testthat')
  pkg_create('.', 'BB', suggests = c('stats', 'testthat'))

  find_srcpkgs('.')
}

examples_srcpkgs_star <- function(.local_envir = parent.frame()) {
  # CRITICAL: must be BEFORE setup_temp_dir(), so that loaded packages are actually dangling
  withr::defer({cleanup_dangling_srcpkgs()}, .local_envir)
  setup_temp_dir(.local_envir = .local_envir)


  pkg_create('.', 'AA', suggests = 'roxygen2')
  pkg_create('.', 'BB', suggests = 'stats')
  pkg_create('.', 'CC')
  pkg_create('.', 'DD')
  pkg_create('.', 'EE', 
    imports = c('AA', 'BB'), 
    depends = c('BB', 'DD'), 
    suggests = c('CC', 'roxygen2'))

  find_srcpkgs('.')
}

examples_srcpkgs_complex_imports <- function(.local_envir = parent.frame()) {
  # CRITICAL: must be BEFORE setup_temp_dir(), so that loaded packages are actually dangling
  withr::defer({cleanup_dangling_srcpkgs()}, .local_envir)
  setup_temp_dir(.local_envir = .local_envir)

  # G-->C-->B-->A--->stats, F-->D-->B, E-->A, Z, G--->utils
  # N.B: we use namespace = TRUE because for unloading, only the ns-imports are considered
  pkg_create('.', 'AA', imports = 'stats', namespace = TRUE)
  pkg_create('.', 'BB', imports = 'AA', namespace = TRUE)
  pkg_create('.', 'CC', imports = 'BB', namespace = TRUE)
  pkg_create('.', 'DD', imports = 'BB', namespace = TRUE)
  pkg_create('.', 'FF', imports = 'DD', namespace = TRUE)
  pkg_create('.', 'EE', imports = 'AA', namespace = TRUE)
  pkg_create('.', 'GG', imports = 'CC', depends = "utils", namespace = TRUE)
  pkg_create('.', 'ZZ')

  find_srcpkgs('.')
}

examples_srcpkgs_complex_deps <- function(.local_envir = parent.frame()) {
  # CRITICAL: must be BEFORE setup_temp_dir(), so that loaded packages are actually dangling
  withr::defer({
    # message("examples_srcpkgs_complex_deps --> cleanup")
    cleanup_dangling_srcpkgs()
  }, .local_envir)

  setup_temp_dir(.local_envir = .local_envir)


  # A->B->C->D, B->D->E, Z
  # N.B: we use namespace = TRUE because for unloading, only the ns-imports are considered
  pkg_create('.', 'AA', imports = c('stats', 'BB'), depends = 'CC')
  pkg_create('.', 'BB', imports = c('CC', 'DD'))
  pkg_create('.', 'CC', imports = 'DD', suggests = 'EE')
  pkg_create('.', 'DD', depends = 'EE')
  pkg_create('.', 'EE')
  pkg_create('.', 'ZZ')

  find_srcpkgs('.')
}
 

.pkg_check <- 
test_that("pkg_check", {
  testthat::skip_on_cran()

  src_pkgs <- examples_srcpkgs_basic()
  setup_temp_dir()

  ###### with tests
  add_dummy_test_to_srcpkgs(src_pkgs)
  # make BB to fail
  writeLines(r"{ stop("aie aie aie") }", file.path(src_pkgs$BB$path, "tests/testthat/setup.R"))

  ### error_on = "never"
  expect_error(chk <- pkg_check("BB", src_pkgs = src_pkgs, error_on = "never", quiet = TRUE), NA)
  expect_s3_class(chk, "pkg_check")

  # s3 generic: summary
  df <- summary(chk)
  
  expect_equal(nrow(df), 1)
  expect_type(df$time, "double")

  df$time <- 1
  expect_identical(df, data.frame(package = "BB", errors = 1L, warnings = 0L, notes = 0L, time = 1))

  # as.data.frame - N.B: currently same as summary()
  expect_identical(as.data.frame(chk), summary(chk))

  # as.logical
  expect_false(as.logical(chk))

  # BB should error if e.g. error_on=="error"
  capture_output(expect_error(pkg_check("BB", src_pkgs = src_pkgs, error_on = "error", quiet = TRUE), 
    "R CMD check found ERRORs"))

  ### print
  # local_reproducible_output(crayon = TRUE, unicode = TRUE)
  # N.B: since the print output includes a stacktrace, which depends on testthat/rlang versions
  # we can not expect_snapshot()
  # anyway the print method is provided by the rcmdcheck package/object
  
  mute(expect_error(print(chk), NA))
#   # fix time to get a reproducible output
#   chk$duration <- 1.5
#   # fix non deterministic path to get a reproducible output
#   chk$test_fail[[1]] <- sub(".*setup.R", "in path /dummy/setup.R", chk$test_fail[[1]])
#   expect_snapshot(print(chk))
})

.pkgs_deps <- 
test_that("pkgs_deps", {
  # G-->C-->B-->A--->stats, F-->D-->B, E-->A, Z, G--->utils
  pkgs <- examples_srcpkgs_complex_imports()

  expect_identical(pkgs_deps("CC", pkgs), c("BB", "AA", "stats"))
  expect_identical(pkgs_deps("FF", pkgs), c("DD", "BB", "AA", "stats"))
  expect_identical(pkgs_deps(c("FF", "CC"), pkgs), c("DD", "BB", "AA", "stats"))

  expect_identical(pkgs_deps("GG", pkgs), c("CC", "utils", "BB", "AA", "stats"))
  expect_identical(pkgs_deps(c("FF", "GG"), pkgs), c("CC", "DD", "utils", "BB", "AA", "stats"))

  expect_identical(pkgs_deps(c("FF", "GG"), pkgs, imports = FALSE), "utils")
  expect_identical(pkgs_deps(c("FF", "GG"), pkgs, depends = FALSE), c("CC", "DD", "BB", "AA", "stats"))
  expect_identical(pkgs_deps(c("FF", "GG"), pkgs, installed = FALSE), c("CC", "DD", "BB", "AA"))
  expect_identical(pkgs_deps(c("FF", "GG"), pkgs, source = FALSE), c("stats", "utils"))

  ### reverse
  expect_identical(pkgs_deps("AA", pkgs, reverse = TRUE), c("BB", "DD", "CC", "GG", "FF", "EE"))

  expect_identical(pkgs_deps("CC", pkgs, reverse = TRUE), "GG")
  expect_identical(pkgs_deps("DD", pkgs, reverse = TRUE), "FF")
  expect_identical(pkgs_deps(c("CC", "DD"), pkgs, reverse = TRUE), c("GG", "FF"))
})



.pkgs_deps_1pkg <- 
test_that(".pkgs_deps_1pkg", {
  # A->B->C->D, B->D->E, Z
  examples_srcpkgs_complex_deps()

  pkgs <- find_srcpkgs('.')

  NADA <- character()

  ### CC
  # N.B: the order matters (topological)
  expect_identical(pkgs_deps('CC', pkgs), c("DD", "EE"))
  expect_identical(pkgs_deps('CC', pkgs, source = FALSE), NADA)
  expect_identical(pkgs_deps('CC', pkgs, suggests = FALSE), c("DD", "EE"))
  expect_identical(pkgs_deps('CC', pkgs, depends = FALSE), c("DD", "EE"))
  expect_identical(pkgs_deps('CC', pkgs, depends = FALSE, suggests = FALSE), "DD")
  
  ### A
  expect_identical(pkgs_deps('AA', pkgs), c("BB", "stats", "CC", "DD", "EE"))
  expect_identical(pkgs_deps('AA', pkgs, installed = FALSE), c("BB", "CC", "DD", "EE"))
  expect_identical(pkgs_deps('AA', pkgs, source = FALSE), "stats")
  expect_identical(pkgs_deps('AA', pkgs, imports = FALSE), c("CC", "EE"))
  expect_identical(pkgs_deps('AA', pkgs, imports = FALSE, depends = FALSE), NADA)

  ## E
  expect_identical(pkgs_deps('EE', pkgs), NADA)

  ### reverse
  expect_identical(pkgs_deps('EE', pkgs, reverse = TRUE), c("DD", "CC", "BB", "AA"))
  expect_identical(pkgs_deps('AA', pkgs, reverse = TRUE), NADA)

  ### external/installed package
  expect_identical(pkgs_deps('stats', pkgs), NADA)
  expect_identical(pkgs_deps('stats', pkgs, reverse = TRUE), "AA")
})



.get_srcpkg_dependencies <- 
test_that("get_srcpkg_dependencies", {
  setup_temp_dir()
  
  ### pkg with all types of deps
  pkg <- pkg_create('.', 'AA', imports = c('i1', 'i2'), depends = 'd1', suggests = c('s1', 's2'))
  deps <- get_srcpkg_dependencies(pkg)
  expect_identical(deps, list(imports = c("i1", "i2"), depends = "d1", suggests = c("s1", "s2")))

  ### package with no deps
  pkg <- pkg_create('.', 'nodeps')
  deps <- get_srcpkg_dependencies(pkg)
  expect_identical(deps, list(imports = NULL, depends = NULL, suggests = NULL))
})

#' creates and populates a R package-like folder programmatically, useful for writing tests
#' 
#' basically a wrapper around `utils::package.skeleton()`
#'
#' @param dir the directory in which to create the package, as a string
#' @param name  the package name, as a string
#' @param functions a named list of functions to add to the package
#' @param imports the "imports" dependencies
#' @param depends the "depends" dependencies
#' @param suggests the "suggests" dependencies
#' @param namespace whether to write the namespace file (currently only applicable to the imports.
#'        N.B: if the namespace file is generated, roxygen will refuse to update it
#' @param roxygen_imports whether to write the roxygen statements to defined the imports
#' @param ignore_hidden_files whether to create a `.Rbuildignore` file to ignore hidden files.
#' 
#' @return the srcpkg instance, invisibly
#' @keywords internal
pkg_create <- function(dir, name, functions = list(dummy = function() 'DUMMY'),
   imports = NULL, depends = NULL, suggests = NULL, namespace = FALSE, roxygen_imports = FALSE,
   ignore_hidden_files = TRUE)
{

  dir.create(dir, recursive = TRUE, showWarnings = FALSE)

  env <- list2env((functions))
  mute(utils::package.skeleton(name = name, path = dir, environment = env, encoding = 'UTF-8'), 
    warnings = FALSE, output = FALSE)
  

  pkg_path <- file.path(dir, name)

  desc_path <- file.path(pkg_path, 'DESCRIPTION')
  stop_unless(file.exists(desc_path),  'can not find DESCRIPTION file at "%s"', desc_path)

  fix_description(desc_path, list(License = "GPL-3"))

  ns_path <- file.path(pkg_path, 'NAMESPACE')
  if (!namespace) {
    unlink(ns_path)
  }

  .add_section_to_description <- function(section, contents) {
    spacer <- '  '
    line <- paste0(paste0(section, ': \n', spacer), paste0(contents, collapse = paste0(",\n", spacer)), '\n')
    cat(line, file = desc_path, append = TRUE)
  }

  if (!is.null(imports)) {
    .add_section_to_description('Imports', imports)

    if (namespace) {
      cat(sprintf('import(%s)', imports), file = ns_path, sep = '\n', append = TRUE)
    }
    if (roxygen_imports) {
      roxygen_import_file <- file.path(pkg_path, 'R/imports.R')
      lines = c(sprintf("#' @import %s", imports), 'NULL')
      cat(lines, file = roxygen_import_file, sep = '\n')
    }
  }
  if (!is.null(depends)) .add_section_to_description('Depends', depends)
  if (!is.null(suggests)) .add_section_to_description('Suggests', suggests)

  if (ignore_hidden_files) {
    cat("^[.].*$\n^.*/[.].*$\n", file = file.path(pkg_path, ".Rbuildignore"))
  }

  ### man
  unlink(file.path(pkg_path, "man"), recursive = TRUE)

  invisible(srcpkg(path = pkg_path))
}


fix_description <- function(path, lst) {
  df <- read.dcf(path, all = TRUE)
  df2 <- utils::modifyList(df, lst)
  write.dcf(df2, path)
}

# tests if a package is loaded (but maybe not attached)
pkg_is_loaded <- function(pkg_or_name) {
  isNamespaceLoaded(as_pkg_name(pkg_or_name))
}

# tests if a package is attached
pkg_is_attached <- function(pkg_or_name) {
  as_pkg_name(pkg_or_name) %in% pkg_list_attached()
}

#' lists the packages that are attached, i.e. present in the R search() path
#' 
#' @return the names of attached package name as a character vector
#' @export
#' @examples
#' print(sort(pkg_list_attached()))
pkg_list_attached <- function() {
  pattern <- '^package:'
  sub(pattern, '', grep(pattern, search(), value = TRUE))
}

pkg_detach <- function(pkg_or_name) {
  pkg_name <- as_pkg_name(pkg_or_name)
  if (pkg_is_attached(pkg_name)) {
    pn <- paste0('package:', pkg_name)
    detach(pn, character.only = TRUE)
  }
}

# pkg_attach <- function(pkg_name) {
#   if (pkg_is_attached(pkg_name)) return()
#   # we must trick R CMD check, that does not like us messing with package from inside a package
#   get('attach', envir = baseenv())(NULL, name = paste0('package:', pkg_name))
# }

# for a loaded package/namespace pkg, find the packages that are namespace-imported by this package
#  this is fast and do not require any I/O, everything happens in memory
# @return the package names
pkg_list_ns_imports <- function(pkg_name) {
  ns <- getNamespace(pkg_name)
  lst <- getNamespaceImports(ns)
  # currently the output is list with 2 parts:
  # 1. the imported package names
  # 2. the imported functions, as lists named after the package
  pkgs <- unique(names(lst))

  pkgs[nzchar(pkgs)]
}

test_that("pkg_create - simple", {
  setup_temp_dir()
  cleanup_dangling_srcpkgs()
  on.exit(cleanup_dangling_srcpkgs(), add = TRUE)

  ### simple (no deps)
  pkg_create('.', 'AA')

  expect_false(pkg_is_loaded('AA'))
  expect_true(file.exists('AA'))

  # can be loaded
  on.exit(unloadNamespace('AA'), add = TRUE)
  devtools::load_all('AA', export_all = FALSE, attach = FALSE, quiet = TRUE)
  expect_true(pkg_is_loaded('AA'))
  expect_false(pkg_is_attached('AA'))

  # inspect some properties
  expect_length(pkg_list_ns_imports('AA'), 0)

  # can be unloaded
  unloadNamespace('AA')
  expect_false(pkg_is_loaded('AA'))

  # attach it
  devtools::load_all('AA', export_all = FALSE, attach = TRUE, quiet = TRUE)
  expect_true(pkg_is_loaded('AA'))
  expect_true(pkg_is_attached('AA'))
})


test_that("pkg_create - imports", {
  setup_temp_dir()

  ### declared imports (i.e. DESCRIPTION-only imports)
  pkg_create('.', 'descimports', imports = c('utils', 'tools'))

  on.exit(unloadNamespace('descimports'), add = TRUE)
  devtools::load_all('descimports', export_all = FALSE, attach = FALSE, quiet = TRUE)

  # N.B: no namespace-imports!!!
  expect_length(pkg_list_ns_imports('descimports'), 0)

  unloadNamespace('descimports')
  expect_error(pkg_list_ns_imports('descimports'), 'no package')

  ### namespace-imports
  pkg_create('.', 'nsimports', imports = c('utils', 'tools'), namespace = TRUE)

  on.exit(unloadNamespace('nsimports'), add = TRUE)
  devtools::load_all('nsimports', export_all = FALSE, attach = FALSE, quiet = TRUE)

  # !! now, the ns-imports are listed
  expect_setequal(pkg_list_ns_imports('nsimports'), c('utils', 'tools'))

  unloadNamespace('nsimports')
  expect_error(pkg_list_ns_imports('nsimports'), 'no package')

  ### namespace-imports via roxygen
  pkg_create('.', 'roxyimports', imports = c('utils', 'tools'), roxygen_imports = TRUE)

  mute(devtools::document('roxyimports', quiet = TRUE))

  on.exit(unloadNamespace('roxyimports'), add = TRUE)
  devtools::load_all('roxyimports', export_all = FALSE, attach = FALSE, quiet = TRUE)

  expect_setequal(pkg_list_ns_imports('roxyimports'), c('utils', 'tools'))

  unloadNamespace('roxyimports')
  expect_error(pkg_list_ns_imports('roxyimports'), 'no package')
})


test_that("pkg_create - depends", {
  setup_temp_dir()

  pkg_create('.', 'AA')
  pkg_create('.', 'BB', depends = 'AA')

  on.exit({
    unloadNamespace('BB')
    unloadNamespace('AA')
  }, add = TRUE)

  # load_all() is not able to load BB
  expect_error(devtools::load_all('BB', export_all = FALSE,  quiet = TRUE), 'required')

  # load AA explicitly and ATTACH it!! Otherwise load_all() will fail
  devtools::load_all('AA', export_all = FALSE, attach = TRUE, quiet = TRUE)

  # then we can load Bb
  devtools::load_all('BB', export_all = FALSE, attach = FALSE, quiet = TRUE)
  expect_true(pkg_is_loaded('BB'))

  # "depends" are not listed as imports!
  expect_length(pkg_list_ns_imports('BB'), 0)
})


# test_that("multiplication works", {
# .create_package <- function() {
#   setup_temp_dir()

#   ###  package with no objects
#   pkg0 <- create_package('.', 'pkg0')
#   expect_true(file.exists('pkg0'))

#   ### normal package
#   unload_pkg('pkg1', quiet = TRUE)
#   pkg1 <- create_package('.', 'pkg1', functions = list(f = function(x,y) x+y))
#   expect_true(file.exists('pkg1'))
#   expect_true(file.exists('pkg1/R'))
#   expect_true(file.exists('pkg1/DESCRIPTION'))

#   expect_is(pkg1, 'package')
#   expect_equal(pkg1$package, 'pkg1')

#   # try to load it
#   devtools::load_all(pkg1, export_all = TRUE)
#   expect_equal( pkg1::f(1,2), 3 )

#   ### package with 1 import)
#   unload_pkg('pkg2', quiet = TRUE)
#   pkg2 <- create_package('.', 'pkg2', imports = 'pkg1')
#   imp2 <- devtools::parse_deps(pkg2$imports)
#   expect_equal(imp2$name, 'pkg1')

#   # check roxygen imports
#   roxygenise_pkg(pkg2, quiet = TRUE)
#   pkg2bis <- as_srcpkg(pkg2$path)

#   deps <- srcpkgs:::compute_srcpkgs_dependencies

#   srcpkgs <- find_srcpkgs('.')
#   expect_identical(deps(pkg2bis, src_pkgs = srcpkgs), deps(pkg2, src_pkgs = srcpkgs))

#   devtools::load_all(pkg2)
#   deps <- srcpkgs:::make_deps_list_of_loaded_pkgs()
#   expect_identical(deps[['pkg2']], 'pkg1')

#   ### package with 2 imports
#   pkg3 <- create_package('.', 'pkg3', imports = c('pkg1', 'pkg2'))
#   imp3 <- devtools::parse_deps(pkg3$imports)
#   expect_equal(imp3$name, c('pkg1', 'pkg2'))

#   ### package with depends and imports and suggests
#   pkg4 <- create_package('.', 'pkg4', imports = 'pkg1', depends = c('pkg3', 'pkg2'), suggests = 'pkg0')

#   imp4 <- devtools::parse_deps(pkg4$imports)
#   expect_equal(imp4$name, 'pkg1')
#   dep4 <- devtools::parse_deps(pkg4$depends)
#   expect_equal(dep4$name, c('pkg3', 'pkg2'))
#   sugg4 <- devtools::parse_deps(pkg4$suggests)
#   expect_equal(sugg4$name, 'pkg0')

#   ### recursive creation of package directory
#   pkg <- create_package('dir1/dir2', 'pkg')
#   expect_true(file.exists('dir1/dir2/pkg/DESCRIPTION'))

#   on.exit({unload_pkg(pkg2, quiet = TRUE); unload_pkg(pkg1, quiet = TRUE)}, add = TRUE)
# }
# test_that('create_package', .create_package())


# })

# this function is automatically called when the package is attached
# we use that to hack the R loaders unless inhibit_r_loaders_hack()
.onAttach <- function(libname, pkgname) {
  if (!inhibit_r_loaders_hack() && !is_loaders_hack_installed()) {
    hack_r_loaders <- try(getFromNamespace("hack_r_loaders", "srcpkgs"), silent = TRUE)
    if (!is_error(hack_r_loaders)) {
      hack_r_loaders()
      packageStartupMessage('hacked R loaders (cf srcpkgs::hack_r_loaders()).')
    }
  }
}

.onDetach <- function(libpath) {
  srcpkgs::unhack_r_loaders()
}

# incidence matrix helper to easily create graph examples
graph_from_strings <- function(...) {
  dots <- list(...)
  seqs <- strsplit(as.character(dots), '->')

  .parse_seq <- function(seq) {
    lapply(seq_len(length(seq) - 1), function(x) c(seq[x], seq[x + 1])) 
  }
  edges <- fast_unlist(lapply(seqs,  .parse_seq))
  nodes <- fast_unlist(seqs)
  graph_from_edges(edges, nodes)
}

graph_from_edges <- function(edges_lst, nodes = fast_unlist(edges_lst)) {
  nodes <- sort(unique(nodes))
  nb <- length(nodes)
  mat <- matrix(0L, nb, nb, dimnames = list(nodes, nodes))
  for (edge in edges_lst) 
    mat[edge[1], edge[2]] <- 1L
  
  mat
}

graph_from_deps <- function(deps_lst) {
  .dep_to_edges <- function(dep) {
    lapply(deps_lst[[dep]], function(x) c(dep, x))
  }
  edges <- fast_unlist(lapply(names(deps_lst), .dep_to_edges))
  nodes <- sort(unique(c(fast_unlist(deps_lst), names(deps_lst))))
  graph_from_edges(edges, nodes)
}


graph_topo_sort <- function(mat) {
  nb <- nrow(mat)
  if (nb == 0) return(mat)

  is_root <- .colSums(mat, nb, nb) == 0
  roots <- colnames(mat)[is_root]
  stop_unless(length(roots), 'error, not a DAG')
  if (length(roots) == nb) return(roots)

  mat2 <- mat[!is_root, !is_root, drop = FALSE]

  c(roots, graph_topo_sort(mat2))
}

graph_topo_sort_nodes <- function(mat, nodes) {
  if (!length(nodes)) return(nodes)
  graph_topo_sort(sub_graph(mat, nodes))
}

# N.B: the dependents are topologically sorted
graph_get_all_dependents <- function(mat, node) {
  if (!node %in% rownames(mat)) return(node)
  rev(graph_get_all_dependencies(t(mat), node))
}

# depth first search
graph_dfs <- function(mat, node,  processed = new.env(parent = emptyenv())) {
  processed[[node]] <- TRUE
  deps <- colnames(mat)[which(mat[node, ] > 0, useNames = FALSE)]
  deps <- setdiff(deps, names(processed))
  for (dep in deps) graph_dfs(mat, dep, processed)
}

graph_get_all_dependencies <- function(mat, node) {
  if (!node %in% rownames(mat)) return(character())
  processed <- new.env(parent = emptyenv())
  graph_dfs(mat, node, processed)

  deps <- setdiff(names(processed), node)
  graph_topo_sort_nodes(mat, deps)
}



# keep only given nodes
sub_graph <- function(mat, nodes) {
  keep <- intersect(rownames(mat), nodes)
  mat[keep, keep, drop = FALSE]
}

test_that("graph_from_edges", {
  ### A->B, A->C, B->C, B->D
  edges <- list(c('A', 'B'), c('A', 'C'), c('B', 'D'), c('B', 'C'))
  mat <- graph_from_edges(edges)

  expect_identical(mat, graph_from_deps(list(A = c('B', 'C'), B = c('D', 'C'))))
})


test_that("graph_from_deps", {
  ### single node, no edge
  mat <- graph_from_deps(list(A = NULL))
  expect_identical(mat, matrix(0L, 1, 1, dimnames = list('A', 'A')))

  ### A->A
  mat <- graph_from_deps(list(A = 'A'))
  expect_identical(mat, matrix(1L, 1, 1, dimnames = list('A', 'A')))

  ### A->B
  mat <- graph_from_deps(list(A = 'B'))

  expect_equal(dim(mat), c(2, 2))
  expect_equal(mat['A', 'B'], 1)
  expect_equal(sum(mat), 1)

  ### A->B, A->C, B->C, B->D
  mat <- graph_from_deps(list(A = c('B', 'C'), B = c('D', 'C')))

  nodes <- LETTERS[1:4]
  expect_identical(mat, 
    matrix(
      as.integer(c(0, 1, 1, 0,
        0, 0, 1, 1, 
        0, 0, 0, 0,
        0, 0, 0, 0)),
      4, 4, byrow = TRUE, dimnames = list(nodes, nodes)))
})

test_that("graph_from_strings", {
  ###
  mat <- graph_from_strings('A->B', 'B->C', 'D->C')
  expect_identical(mat, graph_from_edges(list(c('A', 'B'), c('B', 'C'), c('D', 'C'))))

  ### combined
  mat <- graph_from_strings('A->B->C', 'B->C->D')
  expect_identical(mat, graph_from_edges(list(c('A', 'B'), c('B', 'C'), c('B', 'C'), c('C', 'D'))))

  ### isolated nodes
  mat <- graph_from_strings('A->B->C', 'D')
  expect_identical(mat, graph_from_deps(list(A = 'B', B = 'C', D = NULL)))
})

.graph_topo_sort <- 
test_that("graph_topo_sort", {
  mat <- matrix(0, 1, 1, dimnames = list('A', 'A'))
  expect_identical(graph_topo_sort(mat), 'A')

  ## A->B
  mat <- graph_from_strings('A->B')
  expect_identical(graph_topo_sort(mat), c('A', 'B'))

  ## A->B->C
  mat <- graph_from_strings('A->B->C')
  expect_identical(graph_topo_sort(mat), c('A', 'B', 'C'))

  # A->B->C, B->D->C
  mat <- graph_from_strings('A->B->C', 'B->D->C')
  expect_identical(graph_topo_sort(mat), c('A', 'B', 'D', 'C'))

  ## cycle
  mat <- graph_from_strings('A->B', 'B->A')
  expect_error(graph_topo_sort(mat), 'not a DAG')

  # empty matrix
  mat <- matrix(0, 0, 0)
  expect_identical(graph_topo_sort(mat), mat)
})


test_that("graph_get_all_dep*", {
  ## A->B, C->D
  mat <- graph_from_strings('A->B', 'C->D')

  expect_identical(graph_get_all_dependencies(mat, 'A'),  'B')
  expect_identical(graph_get_all_dependencies(mat, 'B'), character())
  expect_identical(graph_get_all_dependencies(mat, 'C'),  'D')
  expect_identical(graph_get_all_dependencies(mat, 'D'), character())

  expect_identical(graph_get_all_dependents(mat, 'B'), 'A')
  expect_identical(graph_get_all_dependents(mat, 'D'), 'C')
  expect_identical(graph_get_all_dependents(mat, 'A'), character())
  expect_identical(graph_get_all_dependents(mat, 'C'), character())

  # edge cases: not in graph
  expect_identical(graph_get_all_dependencies(mat, 'Z'),  character())
  expect_identical(graph_get_all_dependents(mat, 'Z'),  'Z')


  ### A->B, C
  mat <- graph_from_strings('A->B', 'C->C')
  mat['C', 'C'] <- 0L

  expect_identical(graph_get_all_dependencies(mat, 'A'),  'B')
  expect_identical(graph_get_all_dependencies(mat, 'B'), character())
  expect_identical(graph_get_all_dependencies(mat, 'C'), character())

  expect_identical(graph_get_all_dependents(mat, 'B'), 'A')
  expect_identical(graph_get_all_dependents(mat, 'A'), character())
  expect_identical(graph_get_all_dependents(mat, 'C'), character())

  ### A->B->C
  mat <- graph_from_strings('A->B->C')

  expect_identical(graph_get_all_dependencies(mat, 'A'),  c('B', 'C'))
  expect_identical(graph_get_all_dependencies(mat, 'B'), 'C')
  expect_identical(graph_get_all_dependencies(mat, 'C'), character())

  expect_identical(graph_get_all_dependents(mat, 'A'), character())
  expect_identical(graph_get_all_dependents(mat, 'B'), 'A')
  expect_identical(graph_get_all_dependents(mat, 'C'), c('A', 'B'))

  ### A->B->C, A->C
  mat <- graph_from_strings('A->B', 'C->B', 'A->C')

  expect_identical(graph_get_all_dependencies(mat, 'A'),  c('C', 'B'))
  expect_identical(graph_get_all_dependencies(mat, 'B'), character())
  expect_identical(graph_get_all_dependencies(mat, 'C'), 'B')

  expect_identical(graph_get_all_dependents(mat, 'A'), character())
  expect_identical(graph_get_all_dependents(mat, 'B'), c('A', 'C'))
  expect_identical(graph_get_all_dependents(mat, 'C'), 'A')

  # let's rotate it B->C, B->A, C->A
  new_names <- c('B', 'A', 'C')
  dimnames(mat) <- list(new_names, new_names)

  expect_identical(graph_get_all_dependencies(mat, 'B'),  c('C', 'A'))
  expect_identical(graph_get_all_dependents(mat, 'A'), c('B', 'C'))

  ### more complex example
  mat <- graph_from_strings('A->B->C', 'A->D->B', 'D->C')

  expect_identical(graph_get_all_dependencies(mat, 'A'), c('D', 'B', 'C'))
  expect_identical(graph_get_all_dependencies(mat, 'D'), c('B', 'C'))

  expect_identical(graph_get_all_dependents(mat, 'C'), c('A', 'D', 'B'))
  expect_identical(graph_get_all_dependents(mat, 'B'), c('A', 'D'))
  expect_identical(graph_get_all_dependents(mat, 'D'), 'A')

  ### not a DAG
  mat <- graph_from_strings('A->B->A')

  expect_identical(graph_get_all_dependencies(mat, 'A'), 'B')
  expect_identical(graph_get_all_dependents(mat, 'A'), 'B')
})

#' list the available hobbits
#' @return hobbit names as character
#' @export
hobbits <- function() {
  res <- list()
  # N.B: easier to edit programmatically in the vignette this way
  res$frodo <- frodo::frodo() # in imports

  res
}

########################################################################
# wraps a list of source package objects (srcpkg/package) as a S3 class
########################################################################

#' creates a new "srcpkgs" object
#' 
#' @param pkgs    an existing srcpkgs object (no op), or a list of 
#'  source package-like objects
#' @param paths   a list of source package paths as a character vector or list
#' @return a `srcpkgs` object (a list named after the package names)
#' @export
#' @examples
#' # build dummy source packages
#' pkg1 <- setup_and_get_dummy_srcpkg()
#' pkg2 <- pkg1
#' pkg2$package <- "dummy.srcpkg2"
#' 
#' print(srcpkgs(list(pkg1, pkg2)))
#' print(srcpkgs(paths = pkg1$path))
srcpkgs <- function(
  pkgs = lapply(paths, devtools::as.package), 
  paths = NULL) 
{
  force(paths)
  force(pkgs)
  if (inherits(pkgs, 'srcpkgs')) return(pkgs)
  stop_unless(length(pkgs), 'bad arg "pkgs": must not be empty')
  stop_unless(is.list(pkgs),'bad arg "pkgs", must be a list')
  stop_unless(inherits(pkgs[[1]], 'package'),'bad arg "pkgs", must be a list of package objects')

  # upgrade to srcpkg if needed
  pkgs <- lapply(pkgs, srcpkg)

  # create the srcpkgs object
  names(pkgs) <- get_elements(pkgs, 'package')
  class(pkgs) <- "srcpkgs"

  pkgs
}

# makes sure the input x is a srcpkgs or can be converted to
# accepted: pkgs names, paths, srcpkgs, srcpkg
as_srcpkgs <- function(x, src_pkgs = get_srcpkgs()) {
  if (!length(x)) stop('bad input')
  
  if (inherits(x, "srcpkgs")) return(x)
  if (inherits(x, "srcpkg")) return(srcpkgs(list(x)))

  stop_unless(is.character(x), 'bad arg: not a srcpkg(s) nor character')

  ## pkg names??
  if (all(x %in% names(src_pkgs))) {
    return(subset_s3_list(src_pkgs, x))
  }

  ### assume there are paths
  srcpkgs(paths = x)
}


graph_from_srcpkgs <- function(src_pkgs, imports = TRUE, depends = TRUE, suggests = FALSE) {
  nb <- length(src_pkgs)
  if (nb == 0) return(NULL)
  nodes <- names(src_pkgs)

  types <- character()
  if (imports) types <- append(types, 'imports')
  if (depends) types <- append(types, 'depends')
  if (suggests) types <- append(types, 'suggests')

  .deps <- function(node) {
    deps <- get_srcpkg_dependencies(src_pkgs[[node]])
    intersect(fast_unlist(deps[types]), nodes)
  }
  deps_lst <- lapply(nodes, .deps)
  names(deps_lst) <- nodes

  graph_from_deps(deps_lst)
}

################### S3 methods#############################################################

#' @export
as.data.frame.srcpkgs <- function(x, ...) {
  # convert the package lists to data frame
  rows <- lapply(x, as.data.frame.list, stringsAsFactors = FALSE)
  # keep columns of interest
  rows <- lapply(rows, '[', c('package', 'version', 'path', 'imports', 'depends', 'suggests'))

  # bind rows to a df
  df <- do.call(rbind, rows)
  # for some reason, it is subclassed as "fs_path"
  df$path <- as.character(df$path)

  df
}

#' @export
print.srcpkgs <- function(x, ...) {
  df <- as.data.frame(x)
  print(df, ...)
  invisible(x)
}

.srcpkgs <- 
test_that("srcpkgs", {
  setup_temp_dir()

  pkg1 <- pkg_create('.', 'pkg1')
  pkg2 <- pkg_create('.', 'pkg2')
  pkgs <- list(pkg1, pkg2)

  ### from a list of "package"
  packages <- pkgs
  for (i in seq_along(packages)) 
    class(packages[[i]]) <- "package"

  res <- srcpkgs(packages)

  expect_length(res, 2)
  expect_s3_class(res, 'srcpkgs')
  expect_identical(names(res), c("pkg1", "pkg2"))
  expect_identical(res[[1]], pkg1)
  expect_identical(res[[2]], pkg2)

  ### from some srcpkgs
  res2 <- srcpkgs(res)
  expect_identical(res2, res)

  ### from a list of "srcpkg"
  res2 <- srcpkgs(pkgs)
  expect_identical(res2, res)

  ### from paths
  res2 <- srcpkgs(paths = c('pkg1', 'pkg2'))
  expect_identical(res2, res)

  ### edge cases
  expect_error(srcpkgs(NULL), 'empty')
  expect_error(srcpkgs(list()), 'empty')
  expect_error(srcpkgs(1), 'list')
  expect_error(srcpkgs(list(list(1))), 'package')
})

.as_srcpkgs <- 
test_that("as_srcpkgs", {
  src_pkgs <- examples_srcpkgs_basic()

  ### srcpkgs
  res <- as_srcpkgs(src_pkgs, src_pkgs)
  expect_identical(res, src_pkgs)

  x <- subset_s3_list(src_pkgs, 2)
  expect_identical(as_srcpkgs(x, src_pkgs), x)
  
  ### srcpkg
  x <- src_pkgs[[2]]
  expect_identical(as_srcpkgs(x, src_pkgs), srcpkgs(list(x)))
  
  ### names
  res <- as_srcpkgs(c("BB", "AA"), src_pkgs)
  expect_identical(res, subset_s3_list(src_pkgs, c("BB", "AA")))
  
  ### paths
  res <- as_srcpkgs(src_pkgs$BB$path)
  expect_identical(res, subset_s3_list(src_pkgs, "BB"))

  ### edge cases
  expect_error(as_srcpkgs(NULL), "bad input")
  expect_error(as_srcpkgs(1), "bad arg")
  expect_error(as_srcpkgs("/toto"), "not a directory")
})


.as.data.frame.srcpkgs <- 
test_that("as.data.frame.srcpkgs", {
  setup_temp_dir()
  pkg1 <- pkg_create('.', 'pkg1', imports = c('i1', 'i2'), depends = c('d1', 'd2'))
  pkg2 <- pkg_create('.', 'pkg2', imports = 'i1', suggests = c('s1', 's2'))

  res <- find_srcpkgs('.')
  
  df <- as.data.frame(res)

  expect_s3_class(df, 'data.frame')
  expect_equal(nrow(df), 2)
  expect_identical(df$package, c("pkg1", "pkg2"))
  # N.B: on windows the path look something like 
  # C:/Users/runneradmin/AppData/Local/Temp/Rtmpc7Jpvu/working_dir/RtmpoVzues/fileab4454c56a1/pkg1
  # or the short version
  # "C:/Users/RUNNER~1/AppData/Local/Temp/Rtmpc7Jpvu/working_dir/RtmpoVzues/fileab4454c56a1/pkg1
  # --> better not to compare the path, just the basename
  expect_setequal(basename(df$path), c("pkg1", "pkg2"))
  expect_true(all(c('imports', 'depends', 'suggests') %in% names(df)))

  expect_identical(df$imports, c("i1,i2", "i1"))
  expect_identical(df$depends, c("d1,d2", ""))
  expect_identical(df$suggests, c("", "s1,s2"))
})

.print.srcpkgs <- 
test_that("print.srcpkgs", {
  setup_temp_dir()
  pkg1 <- pkg_create('.', 'pkg1')
  pkg2 <- pkg_create('.', 'pkg2')

  res <- find_srcpkgs('.')
  out <- capture.output(print(res, width = 200))

  expect_length(out, 3)
  expect_match(out[1], "package")
  expect_match(out[2], "pkg1")
})

.graph_from_srcpkgs <- 
test_that("graph_from_srcpkgs", {
  src_pkgs <- examples_srcpkgs_complex_deps()
  nb <- length(src_pkgs)

  ### default: imports + depends
  mat <- graph_from_srcpkgs(src_pkgs)

  expect_equal(dim(mat), c(nb, nb))
  .deps <- function(node) names(which(mat[node, ] == 1))
  expect_setequal(.deps('AA'), c('BB', 'CC'))
  expect_setequal(.deps('BB'), c('DD', 'CC'))
  expect_setequal(.deps('CC'), 'DD')
  expect_length(.deps('EE'), 0)
  expect_length(.deps('ZZ'), 0)

  ### only depends
  mat2 <- graph_from_srcpkgs(src_pkgs, imports = FALSE, suggests = FALSE)
  expect_equal(sum(mat2), 2)

  ### only suggests
  mat3 <- graph_from_srcpkgs(src_pkgs, imports = FALSE, depends = FALSE, suggests = TRUE)
  expect_equal(sum(mat3), 1)

  ### edge case
  expect_null(graph_from_srcpkgs(list()))
})

CONFIG_INIT_KEY <- 'SRCPKGS_INIT'

is_initialized <- function() { isTRUE(get_config(CONFIG_INIT_KEY)) }
set_initialized <- function(value = NULL) { set_config(CONFIG_INIT_KEY, value) }

# performs the initialization, via reset() if not already done
init_if_needed <- function() {
  if (is_initialized()) return()
  reset()
  invisible()
}



#' resets the `srcpkgs` settings
#' 
#' With this function, you can reset or set precisely the settings. 
#
#' @inheritParams params
#' @return the settings (cf [settings()]) invisibly
#' @export
#' @examples
#' # reset to appropriate defaults based on your current directory
#' old <- reset()
#' 
#' # explictly set the project root
#' reset(root = tempdir())
#' 
#' # explictly set the source package paths (very unlikely)
#' reset(srcpkgs_paths = c('pkgs/mypkg1', 'pkgs/mypkg2'))
#' 
#' # restore previous settings
#' reset(root = old$root, srcpkgs_paths = old$srcpkgs_paths)
reset <- function(
  root = find_project_root(), 
  srcpkgs_paths = find_srcpkgs_paths(root)
  # ,should_hack_r_loaders = should_hack_r_loaders_default(srcpkgs_paths)) 
){
  set_project_root(root)
  set_srcpkgs_paths(srcpkgs_paths)
  # if (should_hack_r_loaders) hack_r_loaders() else unhack_r_loaders()
  set_initialized(TRUE)
  invisible(settings())
}


#' informs about the settings currently used by `srcpkgs`
#' 
#' @return a named list of:
#'   - initialized: whether the settings are initialized (as triggered by [get_srcpkgs()])
#'   - root: the project root
#'   - srcpkgs_paths: the paths of the source packages to manage
#'   - hack_r_loaders_installed: whether the R loaders are hacked
#'   - hack_r_loaders_enabled: whether the R loaded hack is in action (internal use)
#' @export
#' @examples
#' print(settings())
settings <- function() {
  list(
    initialized = is_initialized(),
    root = get_project_root(), 
    srcpkgs_paths = get_srcpkgs_paths(),
    hack_r_loaders_installed = is_loaders_hack_installed(), 
    hack_r_loaders_enabled = is_loaders_hack_enabled()
  )
}

test_that("reset/settings", {
  OLD <- settings()
  on.exit(restore_init(OLD), add = TRUE)
  src_pkgs <- examples_srcpkgs_basic()

  ### no source packages
  dir.create('empty')

  lst <- reset(root = 'empty')

  expect_identical(basename(lst$root), 'empty')
  expect_identical(basename(lst$srcpkgs_paths), character())

  expect_identical(get_project_root(), lst$root)
  expect_identical(get_srcpkgs_paths(),  lst$srcpkgs_paths)

  expect_identical(lst, settings())

  ### source packages available
  lst <- reset()

  expect_identical(basename(lst$root), basename(getwd()))
  expect_identical(basename(lst$srcpkgs_paths), c('AA', 'BB'))
})

# add_on_exit <- function(expr, where = parent.frame()) {
#   do.call("on.exit", list(substitute(expr), add = TRUE), envir = where)
# }

# setup_temp_dir <- function(setwd = TRUE, env = parent.frame(), ...) {
#   dir <- tempfile(...)
#   dir.create(dir, recursive = TRUE)
#   old_dir <- NULL
#   if (setwd) old_dir <- setwd(dir)

#   # on one line because it not seen by the coverage
#   cleanup <- bquote({if (.(setwd)) setwd(.(old_dir));unlink(.(dir), recursive = TRUE)})

#   do.call(add_on_exit, list(cleanup, parent.frame()))

#   invisible(normalizePath(dir))
# }
setup_temp_dir <- function(setwd = TRUE, .local_envir = parent.frame(), ...) {
  dir <- withr::local_tempdir(..., .local_envir = .local_envir)
  if (setwd) withr::local_dir(dir, .local_envir = .local_envir)
  invisible(normalizePath(dir))
}

# add_on_exit <- function(expr, where = parent.frame()) {
#   do.call("on.exit", list(substitute(expr), add = TRUE), envir = where)
# }

find_dangling_srcpkgs <- function() {
  # pkgs loaded via devtools
  pkgs_names <- sort(dev_packages_fixed())
  # metadata stored by devtools
  pkgs <- lapply(pkgs_names, fetch_srcpkg_meta)
  pkgs <- pkgs[lengths(pkgs) > 0] %||% return(NULL)
  pkgs <- srcpkgs(pkgs)
  
  df <- as.data.frame(pkgs)

  df <- fetch_srcpkgs_meta() %||% return(character())
  df <- df[!file.exists(df$path), ]
  intersect(df$package, loadedNamespaces()) 
}

cleanup_dangling_srcpkgs <- function(quiet = TRUE) {
  if (!quiet) message("cleanup_dangling_srcpkgs")
  pkg_names <- find_dangling_srcpkgs() %||% return()
  pkgs <- lapply(pkg_names, fetch_srcpkg_meta)
  src_pkgs <- srcpkgs(pkgs)
  for (pkg_name in pkg_names) pkg_unload(pkg_name, src_pkgs, quiet = quiet)
}

# radical: unload all double-letter loaded package, such as AA, BB, CC...
cleanup_test_packages <- function() {
  pkg_names <- paste0(LETTERS, LETTERS)
  loaded <- intersect(pkg_names, loadedNamespaces())
  for (pkg_name in loaded) devtools::unload(pkg_name, quiet = TRUE)
}


# useful for tests
restore_init <- function(previous) {
  if (previous$initialized) {
    reset(root = previous$root, srcpkgs_paths = previous$srcpkgs_paths)
  } else {
    set_initialized()
    reset(root = NULL, srcpkgs_paths = NULL)
  }
}

fix_pkg_tests_results_timings <- function(res, time = 0) {
  for (i in seq_along(res)) {
    if (is_error(res[[i]])) next
    res[[i]] <- fix_test_result_timings(res[[i]], time)
  }
  res
}

fix_test_result_timings <- function(res, time = 0) {
  for (j in seq_along(res)) {
    res[[j]]$real <- time
  }
  res
}


add_dummy_test_to_srcpkgs <- function(srcpkgs) {
  for (pkg in srcpkgs) add_dummy_test_to_srcpkg(pkg)
}

add_dummy_test_to_srcpkg <- function(srcpkg, with_failures = TRUE, with_errors = TRUE, with_warnings = TRUE) {
  withr::local_dir(srcpkg$path)
  dir.create("tests/testthat", recursive = TRUE, showWarnings = FALSE)

  .write_test <- function(name, code, test = name) {
    writeLines(sprintf(r"-----{
    test_that("%s", {
      %s
    })
    }-----", name, code), sprintf("tests/testthat/test-%s.R", test))
  }

  .write_test("success", "expect_true(TRUE)")
  if (with_failures) {
    .write_test("failure", "expect_true(FALSE)")
    .write_test("mixed", "expect_true(FALSE);expect_true(TRUE)")
  }
  .write_test("skip", 'skip("skipping");expect_true(FALSE)')
  if (with_errors) .write_test("errors", 'expect_true(TRUE);stop("Arghh");expect_true(TRUE)')
  if (with_failures && with_warnings)  .write_test("warning", 'expect_true(FALSE);warning("watch out");expect_true(FALSE)')
  if (with_failures && with_errors)
    writeLines(r"-----{
    test_that("misc1", {
      expect_true(FALSE)
      expect_true(TRUE)
    })
    test_that("misc2", {
      expect_true(FALSE)
      skip("skipping")
    })
    test_that("misc3", {
      expect_true(TRUE)
      expect_true(TRUE)
    })
    test_that("misc4", {
      expect_true(TRUE)
      warning("fais gaffe")
      stop("aie")
      expect_true(TRUE)
    })
    }-----", "tests/testthat/test-misc.R")

  writeLines(sprintf(r"-----{
    library(testthat)
    library(%s)

    test_check("%s")
  }-----", srcpkg$package, srcpkg$package), "tests/testthat.R")
}

# add_dummy_test_to_srcpkg <- function(srcpkg) {
#   withr::local_dir(srcpkg$path)

#   dir.create("tests/testthat", recursive = TRUE, showWarnings = FALSE)

#   writeLines(r"-----{
#   test_that("success", {
#     expect_true(TRUE)
#   })
#   }-----", "tests/testthat/test-success.R")

#   writeLines(r"-----{
#   test_that("failure", {
#     expect_true(FALSE)
#   })
#   }-----", "tests/testthat/test-failure.R")

#   writeLines(r"-----{
#   test_that("skip", {
#     skip("skipping")
#     expect_true(FALSE)
#   })
#   }-----", "tests/testthat/test-skip.R")

#   writeLines(r"-----{
#   test_that("mixed", {
#     expect_true(FALSE)
#     expect_true(TRUE)
#   })
#   }-----", "tests/testthat/test-mixed.R")

#   writeLines(r"-----{
#   test_that("stop", {
#     expect_true(TRUE)
#     stop("Arghh")
#     expect_true(TRUE)
#   })
#   }-----", "tests/testthat/test-error.R")#

#   writeLines(r"-----{
#   test_that("warn", {
#     expect_true(FALSE)
#     warning("watch out")
#     expect_true(FALSE)
#   })
#   }-----", "tests/testthat/test-warning.R")#

#   writeLines(r"-----{
#   test_that("misc1", {
#     expect_true(FALSE)
#     expect_true(TRUE)
#   })
#   test_that("misc2", {
#     expect_true(FALSE)
#     skip("skipping")
#   })
#   test_that("misc3", {
#     expect_true(TRUE)
#     expect_true(TRUE)
#   })
#   test_that("misc4", {
#     expect_true(TRUE)
#     warning("fais gaffe")
#     stop("aie")
#     expect_true(TRUE)
#   })
#   }-----", "tests/testthat/test-misc.R")

#   writeLines(sprintf(r"-----{
#     library(testthat)
#     library(%s)

#     test_check("%s")
#   }-----", srcpkg$package, srcpkg$package), "tests/testthat.R")
# }

#' tests a package - runs its unit tests
#'
#' This function will test a source package using `testthat`, 
#' making sure the package and its source package dependencies are up-to-date and loaded 
#'
#' @inheritParams params
#' @param filter      filter in the tests to run. cf `testthat::test_dir()`
#' @param export_all  passed to [pkg_load()]. Enables the test functions to easily access to non-exported
#'                    functions. Caveat: If the pkg is already loaded and up-to-date with export_all=FALSE, it will not work.
#' @param ...   passed to `testthat::test_dir()`
#' @return the results as a `pkg_test` object, which is an empty listL if no tests were found
#' @importFrom testthat   test_dir
#' @export
#' @examples
#'  root <- tempfile()
#'  pkg <- setup_and_get_dummy_srcpkg(root)
#'  reset(root)
#'  res <- pkg_test(pkg)
#'  print(res)
pkg_test <- function(pkgid, filter = NULL, src_pkgs = get_srcpkgs(), export_all = TRUE, quiet = TRUE, ...)
{
  if (is_loaders_hack_enabled()) {
    unhack_r_loaders()
    on.exit(hack_r_loaders(), add = TRUE)
  }

  force(src_pkgs)
  pkg <- as_srcpkg(pkgid, src_pkgs)
  pkg_name <- pkg$package
  stop_unless(pkg_name %in% names(src_pkgs), 
    'package "%s" is not managed by srcpkgs, check `get_srcpkgs()`', pkg_name)

  # load and document if needed
  pkg_load(pkg, src_pkgs, quiet = quiet, export_all = export_all)

  test_path <- file.path(pkg$path, "tests/testthat")

  if (!dir.exists(test_path) || length(dir(test_path)) == 0) {
    # no tests found, return an empty  pkg_test
    return(invisible(new_pkg_test(pkg)))
  }

  res <- testthat::test_dir(test_path, filter = filter, stop_on_failure = FALSE, ...)

  invisible(new_pkg_test(pkg, res))
}

new_pkg_test <- function(pkg_name, test_results = list()) {
  attr(test_results, 'pkg') <- pkg_name
  class(test_results) <- c('pkg_test', class(test_results))
  test_results
}

# tells if the test is successful
# N.B: this is not a roxygen comment ON PURPOSE
#' @export
as.logical.pkg_test <- function(x, ...) {
  df <- as.data.frame(x)
  !any(df$error) && all(df$failed == 0)
}

# reformat the testthat_results data frame 
fortify_pkg_test <- function(df) {
 # remove context (which is deprecated)
  df$context <- NULL

  # hide user and system (not really informative)
  df$user <- df$system <- NULL

  # hide results
  df$result <- NULL

  # shorten file
  df$file <- sub("^test-", "", df$file, ignore.case = TRUE)
  df$file <- sub("\\.R", "", df$file, ignore.case = TRUE)

  # rename real into time
  names(df)[names(df) == "real"] <- "time"

  # reorder cols --> put "passed" after failed
  cols <- names(df)
  cols <- setdiff(cols, "passed")
  cols <- append(cols, "passed", after = match("failed", cols))
  df <- df[cols]

  df
}

#' @export
as.data.frame.pkg_test <- function(x, ...) {
  if (!length(x)) {
    return(data.frame(
      file = character(0), 
      test = character(0), 
      nb = integer(0), 
      failed = integer(0), 
      passed = integer(0), 
      skipped = logical(0), 
      error = logical(0), 
      warning = integer(0), 
      time = numeric(0))
    )
  }
  fortify_pkg_test(NextMethod())
}

#' @export
print.pkg_test <- function(x, ...) {
  pkg <- attr(x, 'pkg')

  if (!length(x)) {
    cli::cli_h1(paste0("package ", pkg$package, " has no tests"))
    return(invisible())
  }

  df <- as.data.frame(x)
  df$time <- format(df$time, digits = 3) # only keep ms

  results <- df$result

  ### by test
  bad <- which(df$failed > 0 | df$error)
  print_text_table(df, 
    title = paste0("Test results by test for package ", pkg$package),
    hilite_rows = bad, heatmap_columns = 'time')

  ### by file
  sdf <- summary(x)
  sdf$time <- format(sdf$time, digits = 3) # only keep ms
  bad <- which(sdf$failed > 0 | sdf$error)
  print_text_table(sdf, 
    title = paste0("Test results by file for package ", pkg$package),
    hilite_rows = bad, heatmap_columns = 'time')

  ### overview
  sdf <- summary(x, col = NULL)
  final_message <- if (as.logical(x)) "Tests successful :) !" else "Test failed :( ..."
  print_text_table(sdf, title = paste0("Test results overview for package ", pkg$package), footnote = final_message)
  
  invisible()
}

# summarizes by test source file. Use by=NA to summarize all rows to get an overview
# N.B: this is not a roxygen comment ON PURPOSE
#' @export
summary.pkg_test <- function(object, col = 'file', ...) {
  if (!length(object)) {
    return(data.frame(
      file = character(0), 
      nb = integer(0), 
      failed = integer(0), 
      passed = integer(0), 
      skipped = integer(0), 
      error = integer(0), 
      warning = integer(0), 
      time = numeric(0))
    )
  }

  df <- as.data.frame(object)

  stop_unless(length(col) <= 1, "col must be a column name or NULL")
  stop_if(length(col) == 1 && !col %in% names(df), "bad column name")

  # special case for columns "skipped" and "error", which are logical
  # for correct aggregation, must be integer
  df$skipped <- as.integer(df$skipped)
  df$error <- as.integer(df$error)

  ### select non-character columns: summary of numeric columns (sums)
  data_cols <- which(!vapply(df, is.character, TRUE))

  # aggregation grouping elements
  by <- list(rows = rep(TRUE, nrow(df)))
  if (length(col)) {  
    by <- list(df[[col]])
    names(by) <- col
  }

  # # aggregation function
  # .col_stats <- function(x) { 
  #   if (is.logical(x)) any(x) else sum(x) 
  # }
  sdf <- stats::aggregate(df[, data_cols], by = by, sum)
  if (length(col)) sdf <- sdf[order(sdf[[col]]), , drop = FALSE] else sdf$rows <- NULL
  rownames(sdf) <- NULL

  sdf
}

ROOT_KEY <- 'ROOT'

# gets the current project root
get_project_root <- function() {
  get_config(ROOT_KEY)
}

# sets the current project root
set_project_root <- function(path, force = FALSE) {
  if (!force && length(path)) {
    stop_unless(length(path) == 1 && is.character(path), 'bad arg path: "%s", not a string', path)
    stop_unless(dir.exists(path), 'bad arg path: "%s", not a directory', path)
  }
  set_config(ROOT_KEY, path) 
}

# heuristic: search upwards for the first folder with a .git/ folder, or return here
find_project_root <- function(here = getwd()) {
  if (!length(here)) return(here)
  root <- here
  git <- find_git_dir(here)

  if (length(git)) 
    root <- parent_dir(git)

  root
}


# N.B: assumes dir is an absolute path to a dir, not a file!!
# this implementation works on all OS (unix and windows)
parent_dir <- function(dir) {
  parent <- dirname(dir)
  if (parent == dir) return(NULL) # will happen on windows. e.g. with C:\
  
  parent
}


find_git_dir <- function(here = getwd()) {
  if (!length(here)) return(NULL)
  find_file_upwards('.git', here)
}

find_file_upwards <- function(filename, folder = getwd()) {
  root <- '/'
  stop_unless(dir.exists(folder), 'directory does not exist: "%s"', folder)
  folder <- normalizePath(folder, '/')

  while(1) {
    path <- file.path(folder, filename)
    if (file.exists(path)) return(path)
    folder <- parent_dir(folder)
    if (is.null(folder)) break # not found
  }

  NULL # not found
}

test_that("find_project_root", {
  setup_temp_dir()

  # no .git/
  expect_identical(basename(find_project_root()), basename(getwd()))

  # a .git file, but not a dir!
  writeLines('', '.git')
  expect_identical(basename(find_project_root()), basename(getwd()))

  dir.create('dir1/dir2/dir3', recursive = TRUE)
  dir.create('dir1/dir2/.git')

  root2 <- file.path(getwd(), 'dir1/dir2')
  # N.B: we compare only basenames because of Windows... :(
  # dir2 because of dir2/.git in ..
  expect_identical(basename(find_project_root('dir1/dir2/dir3')), basename(root2))
  # dir2 because of dir2/.git in .
  expect_identical(basename(find_project_root('dir1/dir2/')), basename(root2))
  # # dir1 by default
  # expect_identical(basename(find_project_root('dir1/')), 'dir1')
  # root because of .git file
  expect_identical(basename(find_project_root('dir1/')), basename(getwd()))

  root3 <- file.path(getwd(), 'dir1/dir2/dir3')
  # a .git file, not a directory, like for git submodules
  writeLines("", file.path(root3, '.git'))
  # dir.create(file.path(root3, '.git'))
   # dir3 because of dir3/.git in .
  expect_identical(basename(find_project_root('dir1/dir2/dir3')), basename(root3))
  # dir2 because of dir2/.git in .
  expect_identical(basename(find_project_root('dir1/dir2/')), basename(root2))

  ### edge cases
  expect_null(find_project_root(NULL))
})

.find_git_dir <- 
test_that("find_git_dir", {
  setup_temp_dir()

  expect_null(find_git_dir(NULL))


})



test_that("parent_dir", {
  expect_identical(parent_dir('/'), NULL)
  expect_identical(parent_dir('/A'), '/')
  expect_identical(parent_dir('/A/'), '/')
  expect_identical(parent_dir('/A/B'), '/A')
  expect_identical(parent_dir('/A/B/'), '/A')
  
})

test_that("find_file_upwards", {
  setup_temp_dir()

  # file that should not exist all the way to /
  filename <- 'should_not_exist_1234567890_REALLY'
  expect_null(find_file_upwards(filename))

  ## create the file in a subdir
  dir.create('dir1')
  path <- file.path(getwd(), 'dir1', filename)
  writeLines('toto', path)

  # not found here
  expect_null(find_file_upwards(filename))

  # found in the subdir tho
  expect_identical(basename(find_file_upwards(filename, 'dir1')), filename)

  # create more subdirs
  dir.create('dir1/dir2/dir3', recursive = TRUE)

  expect_identical(basename(find_file_upwards(filename, 'dir1/dir2/dir3')), filename)
  expect_identical(basename(find_file_upwards(filename, 'dir1/dir2/')), filename)
  expect_identical(basename(find_file_upwards(filename, 'dir1')), filename)

})

SRCPKGS_PATHS_KEY <- 'SRCPKGS_PATHS'

#' finds all available source packages starting from the project root
#'
#' N.B: the *hidden* files and directories are ignored.
#' In general, this function is not used directly, instead you should use [get_srcpkgs()]
#' 
#' @inheritParams params
#' @param prune   whether to report packages contained inside another package (e.g. in tests/)
#' @return a "srcpkgs" object (or NULL if none found), a named list of "srcpkg" objects, that essentially are 
#' devtools "package" objects. The list is named after the package names.
#' @export
#' @examples
#' pkg <- setup_and_get_dummy_srcpkg()
#' pkgs <- find_srcpkgs(dirname(pkg$path))
#' print(pkgs)
find_srcpkgs <- function(root = get_project_root(), 
  srcpkgs_paths = find_srcpkgs_paths(root, prune = prune), prune = TRUE)
{
  if (!length(srcpkgs_paths)) return(NULL)
  srcpkgs(paths = srcpkgs_paths)
}

#' get the current source packages list
#' 
#' The first call to this function will trigger the initialization of the package ((cf [reset()]).
#' Since it is used by mostly all user-facing load-related functions, this enables a runtime initialization, 
#' as opposed to a load-time initialization. So for example
#' you may load `srcpkgs`, then change the current directory to your project. 
#' Then the first load will setup the settings.
#' 
#' For optimization, the paths to discovered source packages are cached (cf [reset()] and [settings()].
#' This function will reparse the DESCRIPTION for any change. 
#' If you add or delete a source package, you must reset the source package paths using [reset()]
#' 
#' This function is useful for troubleshooting, to understand what are the source packages discovered 
#' and managed by `srcpkgs` 
#' 
#' @param filter    a pattern to filter the source packages
#' @return the source packages as a "scrpkgs" object, cf [find_srcpkgs()], or NULL if none
#' @export
#' @examples
#' # setup a srcpkg. We need reset because it is not discoverable from the current directory
#' pkg <- setup_and_get_dummy_srcpkg()
#' reset(dirname(pkg$path))
#' 
#' print(get_srcpkgs())
get_srcpkgs <- function(filter = NULL) {
  init_if_needed() # triggers initialization
  paths <- get_srcpkgs_paths() %||% return(NULL)

  filter_srcpkgs(srcpkgs(paths = paths), filter)
}

filter_srcpkgs <- function(src_pkgs, filter = NULL) {
  if (!length(filter) || !nzchar(filter)) return(src_pkgs)
  packages <- vapply(src_pkgs, getElement, "", "package")

  subset_s3_list(src_pkgs, grepl(filter, packages))
}


# sets the current paths of source packages
set_srcpkgs_paths <- function(paths) { set_config(SRCPKGS_PATHS_KEY, paths) }

get_srcpkgs_paths <- function() {  get_config(SRCPKGS_PATHS_KEY) }


# finds all source packages inside a directory
#by default not in hidden (.) dirs, cf list.files() all.files= param
find_srcpkgs_paths <- function(dir, src = TRUE,  prune = TRUE, ...) {
  stop_unless(dir.exists(dir), 'bad directory "%s"', dir)

  # look for all folders with a DESCRIPTION file (required for a R package)
  desc_paths <- dir(dir, pattern = '^DESCRIPTION$', full.names = TRUE, recursive = TRUE, no.. = TRUE, 
    include.dirs = FALSE, ...)
  paths <- dirname(desc_paths)

  # remove packages inside other packages (for example in tests/ folder)
  if (prune && length(paths) > 1) {
    paths <- prune_paths(paths)
  }

  if (src) {
    # there could be non-source, i.e. installed packages: remove them
    # Heuristic: installed packages have a Meta/ dir.
    installed <- file.exists(file.path(paths, 'Meta'))
    paths <- paths[!installed]
  }

  unique(sort(paths))
}


prune_paths <- function(paths) {
  # add trailing slash to distinguish between dirs and prefixes
  dirs <- paste0(paths, '/')

  .strictly_contained_by <- function(path) {
    which(path != dirs & startsWith(dirs, path))
  }
  ind <- unlist(lapply(dirs, .strictly_contained_by), use.names = FALSE)
  if (length(ind) > 0)
    paths <- paths[-ind]

  paths
}

.get_srcpkgs <- 
test_that("get_srcpkgs", { 
  setup_temp_dir()
  OLD <- settings()
  on.exit(restore_init(OLD), add = TRUE)
  reset(root = '.')

  ### no pkg yet to be found
  expect_null(find_srcpkgs())
  expect_null(get_srcpkgs())

  ### create a AA package
  pkg_create('.', 'AA', imports = 'stats')
  reset(root = '.')
  ###
  # automatic
  src_pkgs <- find_srcpkgs()
  expect_identical(get_srcpkgs(), src_pkgs)

  ### modify pkg AA
  expect_identical(get_srcpkgs()$AA$imports, 'stats')
  unlink('AA', recursive = TRUE)
  pkg_create('.', 'AA', imports = 'roxygen2')

  expect_identical(get_srcpkgs()$AA$imports, 'roxygen2')

  ### add two packages
  pkg_create('.', 'AB')
  pkg_create('.', 'BB')
  reset()
  src_pkgs <- find_srcpkgs(".")
  expect_identical(get_srcpkgs(), src_pkgs)
  
  expect_identical(get_srcpkgs("AA"), subset_s3_list(src_pkgs, "AA"))
  expect_identical(get_srcpkgs("A"), subset_s3_list(src_pkgs, c("AA", "AB")))
})


.filter_srcpkgs <- 
test_that("filter_srcpkgs", { 
  src_pkgs <- examples_srcpkgs_star()

  expect_identical(filter_srcpkgs(src_pkgs), src_pkgs)
  expect_identical(filter_srcpkgs(src_pkgs, ""), src_pkgs)

  expect_identical(filter_srcpkgs(src_pkgs, "A"), subset_s3_list(src_pkgs, "AA"))
  expect_identical(filter_srcpkgs(src_pkgs, "A|B"), subset_s3_list(src_pkgs, c("AA", "BB")))
})


.prune_paths <- function() {
  prune_paths <- srcpkgs:::prune_paths

  expect_identical(prune_paths('AA'), 'AA')
  expect_identical(prune_paths('AA/'), 'AA/')

  expect_identical(prune_paths(c('AA', 'AA/B')), 'AA')
  expect_identical(prune_paths(c('AA/B', 'AA', 'AA/C')), 'AA')
  expect_identical(prune_paths(c('AA/T/CD/B', 'AA', 'BB')), c('AA', 'BB'))
  expect_identical(prune_paths(c('AA', 'BB/A/C')), c('AA', 'BB/A/C'))

  # prefixes
  expect_identical(prune_paths(c('AA', 'AAA')), c('AA', 'AAA'))

  # circular relations
  expect_identical(prune_paths(c('AA', 'AA/B/C', 'AA/B')), 'AA')
}
test_that("prune_paths", {.prune_paths()})


.find_srcpkgs_paths <- 
test_that("find_srcpkgs_paths", { 
  find_srcpkgs_paths <- srcpkgs:::find_srcpkgs_paths

  dir <- setup_temp_dir(setwd = FALSE)

  ### edge cases
  # no dir
  expect_error(find_srcpkgs_paths('does_not_exist'), 'bad directory')
  # dir empty
  expect_true(length(find_srcpkgs_paths(dir)) == 0)

  ### A package as level 1
  pkg_create(dir, 'AA')

  pkgs <- find_srcpkgs_paths(dir)

  # N.B: we compare only basenames because of Windows... :(
  expect_identical(basename(pkgs), 'AA')
  expect_identical(pkgs, find_srcpkgs_paths(dir)) # reproducible

  #### A and B in same level 1
  pkg_create(dir, 'BB')

  pkgs <- find_srcpkgs_paths(dir)
  
  expect_identical(pkgs, find_srcpkgs_paths(dir))
  expect_identical(basename(pkgs), c('AA', 'BB'))

  ### C is a level 2 pkg compared to A and B
  pkg_create(file.path(dir, 'C_proj'), 'CC')

  pkgs <- find_srcpkgs_paths(dir)
  
  expect_identical(pkgs, find_srcpkgs_paths(dir))
  expect_setequal(basename(pkgs), c('AA', 'BB', 'CC'))

  ### A1 inside A: should be ignored if prune == TRUE
  pkg_create(file.path(dir, 'AA'), 'AA1')

  pkgs <- find_srcpkgs_paths(dir)

  expect_identical(pkgs, find_srcpkgs_paths(dir))
  expect_setequal(basename(pkgs), c('AA', 'BB', 'CC'))

  pkgs <- find_srcpkgs_paths(dir, prune = FALSE)
  expect_setequal(basename(pkgs), c('AA', 'AA1', 'BB', "CC"))

  ### D inside a hidden directory: should be ignored
  pkg_create(file.path(dir, '.hidden'), 'DD')

  pkgs <- find_srcpkgs_paths(dir)

  expect_identical(pkgs, find_srcpkgs_paths(dir))
  expect_setequal(basename(pkgs), c('AA', 'BB', 'CC'))

  pkgs <- find_srcpkgs_paths(dir, all.files = TRUE)
  expect_true('DD' %in% basename(pkgs))

  ### package name is prefix of other package in same level
  pkg_create(dir, 'AAA')

  pkgs <- find_srcpkgs_paths(dir)
  expect_setequal(basename(pkgs), c('AA', 'AAA', 'BB', "CC"))

  ### should not "find" non-source package ==========================
  lib <- file.path(dir, 'lib')
  dir.create(lib)
  
  # create a source package in lib
  pkg_create(lib, 'installed1')

  # it is found by find_srcpkgs_paths()
  pkgs <- find_srcpkgs_paths(dir)
  expect_true('installed1' %in% basename(pkgs))

  # now turn it to a pseudo installed pakage by creating a Meta/ folder
  dir.create(file.path(lib, 'installed1', 'Meta'))
  expect_length(find_srcpkgs_paths(lib), 0)

  pkgs <- find_srcpkgs_paths(dir)
  expect_false('installed1' %in% basename(pkgs))
})

#' computes the dependencies of some (source) packages
#' 
#' @inheritParams params
#' @param source      whether to report source packages
#' @param installed   whether to report installed (non-source) packages
#' @param imports     whether to only consider `imports` dependencies
#' @param depends     whether to only consider `depends` dependencies
#' @param suggests    whether to only consider `suggests` dependencies
#' @param reverse     whether to compute reverse dependencies instead
#' 
#' @return the dependencies, as a character vector, topologically sorted
#' @export
#' @examples
#' pkg <- setup_and_get_dummy_srcpkg()
#' deps_src <- pkgs_deps(pkg, installed = FALSE)
#' deps_inst <- pkgs_deps(pkg, source = FALSE)
#' print(get_srcpkgs())
#' deps_rev <- pkgs_deps(pkg, reverse = TRUE, suggests = FALSE)
pkgs_deps <- function(pkgids, src_pkgs = get_srcpkgs(), 
  source = TRUE, installed = TRUE, 
  imports = TRUE, depends = TRUE, suggests = TRUE, 
  reverse = FALSE) 
{
  .get_pkgs_deps <- function(pkg) {
    lst <- get_srcpkg_dependencies(pkg)
    if (!imports) lst$imports <- NULL
    if (!depends) lst$depends <- NULL
    if (!suggests) lst$suggests <- NULL
    fast_unlist(lst)
  }

  srcpkgs_deps <- lapply(src_pkgs, .get_pkgs_deps)
  mat <- graph_from_deps(srcpkgs_deps)
  
  .fetch_deps <- if (reverse) graph_get_all_dependents else graph_get_all_dependencies
  deps <- fast_unlist(lapply(pkgids, .fetch_deps, mat = mat))

  if (!source) deps <- setdiff(deps, names(src_pkgs))
  if (!installed) deps <- intersect(deps, names(src_pkgs))

  deps <- graph_topo_sort_nodes(mat, deps)

  if (reverse) deps <- rev(deps)

  deps
}


get_srcpkg_dependencies <- function(src_pkg) {
  .parse_deps_field <- function(type) {
    str <- gsub('\n', '', src_pkg[[type]])
    if (length(str) == 0 || !nzchar(str)) return(NULL)
    strsplit(str, ',')[[1]]
  }

  sapply(c('imports', 'depends', 'suggests'), .parse_deps_field, simplify = FALSE)
}

#' @export
print.pkgs_check <- function(x, ...) {
  ### by package
  df <- as.data.frame(x)
  # only keep ms
  df$time <- format(df$time, digits = 3)
  bad <- which(df$errors > 0)
  print_text_table(df, title = "Check results by package", hilite_rows = bad, heatmap_columns = 'time')

  ### overview
  sdf <- summary(x)
  sdf$time <- format(sdf$time, digits = 3)

  print_text_table(sdf, title = "Check results overview")

  ### overall result (parsable)
  cat('\n')
  ok <- as.logical(x)
  if (ok) {
    cat("SUCCESS\n") 
  } else {
    cat("FAILED\n")
  }

  invisible()
}

#' @export
as.logical.pkgs_check <- function(x, ...) {
   df <- summary(x)
   df$error == 0
}


#' @export
as.data.frame.pkgs_check <- function(x, ...) { 
  do.call(rbind, lapply(x, summary.pkg_check)  )
}

#' @export
summary.pkgs_check <- function(object, ...) {
  df <- as.data.frame(object)

  .sum_rm_na <- function(...) sum(..., na.rm = TRUE)
  df$package <- 1 # N.B: package column is converted to counts
  sdf <- stats::aggregate(df, by = list(.dummy = rep(TRUE, nrow(df))), .sum_rm_na)
  sdf$.dummy <- NULL
 
  sdf
}

#' checks a list of source packages
#'
#' @inheritParams params
#' @param filter          filter out the packages to check using this pattern
#' @param fail_on_error   whether to die if there is at least an error or warning in the checks
#' @param ...   passed to `pkg_check`
#' @return the results as a `pkgs_test` object
#' @export
pkgs_check <- function(pkgids = names(filter_srcpkgs(src_pkgs, filter)), src_pkgs = get_srcpkgs(), 
  filter = NULL, lib = ".check",  quiet = FALSE, fail_on_error = FALSE,  ...) 
{  
  force(src_pkgs)
  if (!length(pkgids)) { 
    stop('No package to test')
  }
  pkgs <- as_srcpkgs(pkgids, src_pkgs)
  if (!dir.exists(lib))  dir.create(lib, recursive = TRUE)

  if (!quiet) cli::cli_h1('installing packages in {.file lib}')
  pkgs_install(pkgs, lib, src_pkgs = src_pkgs, quiet = TRUE)

  libpath <- c(lib, .libPaths())
  .check_one_pkg <- function(pkg) {   
    if (!quiet) cli::cli_h1('Checking package {.pkg {pkg$package}}')
    # N.B:  roxygen = FALSE because roxygen was already done by pkgs_install()
    pkg_check(pkg, src_pkgs = src_pkgs, lib = lib, roxygen = FALSE, error_on = "never", 
      quiet = quiet, ...)
  }
  chks <- withr::with_libpaths(libpath, lapply(pkgs, .check_one_pkg))

  chks <- new_pkgs_check(chks, names(pkgs))

  summ <- summary(chks)

  if (fail_on_error && (any(summ$errors > 0) || any(summ$warnings > 0))) {
    cli::cli_abort("checks failed")   
  }
  
  invisible(chks)
}

new_pkgs_check <- function(chks, pkgids) {
  names(chks) <- pkgids
  class(chks) <- 'pkgs_check'
  chks
}

.pkgs_check <- 
test_that("pkgs_check", {
  testthat::skip_on_cran()

  src_pkgs <- examples_srcpkgs_basic()
  setup_temp_dir()

  ###################################### edge cases ###############################
  expect_error(pkgs_check(NULL), 'No package to test')

  ###### with tests
  # add_dummy_test_to_srcpkgs(src_pkgs)
  add_dummy_test_to_srcpkg(src_pkgs[[1]], with_errors = FALSE, with_failures = FALSE)
  add_dummy_test_to_srcpkg(src_pkgs[[2]])

  # make BB to fail
  writeLines(r"{ stop("aie aie aie") }", file.path(src_pkgs$BB$path, "tests/testthat/setup.R"))

  ### 
  chks <- pkgs_check(src_pkgs = src_pkgs, quiet = TRUE)

  expect_s3_class(chks, "pkgs_check")
  expect_true(is.list(chks))
  expect_length(chks, 2)
  expect_s3_class(chks[[1]], "pkg_check")
  expect_s3_class(chks[[2]], "pkg_check")

  ####### =============== S3 methods ================================
  ### as.data.frame - one row per package
  df <- as.data.frame(chks)

  expect_true(is.data.frame(df))
  expect_equal(nrow(df), 2)

  for (i in 1:2) {
    expected <- summary(chks[[i]])
    rownames(expected) <- src_pkgs[[i]]$package
    expect_identical(df[i,], expected)
  }

  ### summary
  sdf <- summary(chks)

  expect_equal(nrow(sdf), 1)
  df$package <- 1
  df2 <- df[1,] + df[2, ]
  rownames(df2) <- NULL
  expect_equal(sdf, df2)
  
  ### as.logical
  expect_false(as.logical(chks))

  #### fail_on_error
  expect_error(pkgs_check(src_pkgs = src_pkgs, quiet = TRUE, fail_on_error = TRUE), "checks failed")

  ### print
  # fix durations for reproducible output
  for (i in seq_along(chks)) chks[[i]]$duration <- pi

  # create a dummy successful pkgs_check by replicating chks[[1]]
  chks_ok <- new_pkgs_check(list(chks[[1]], chks[[1]]), names(chks))
  expect_true(as.logical(chks_ok))

  # call the print method since there seems to be a bug with expect_snapshot() and covr
  mute(expect_error(print(chks), NA))
  mute(expect_error(print(chks_ok), NA))

  local_reproducible_output(crayon = TRUE, unicode = TRUE)
  expect_snapshot(print(chks))
  expect_snapshot(print(chks_ok))
})

.CONFIG <- new.env(parent = emptyenv())

# return the stored config or NULL
get_config <- function(key) {
  get0(key, .CONFIG)
}

set_config <- function(key, value) {
  old <- get_config(key)
  assign(key, value, envir = .CONFIG)
  invisible(old)
}

delete_config <- function(key) {
  old <- get_config(key)
  suppressWarnings(rm(list = key, envir = .CONFIG))
  invisible(old)
}

test_that("config", {
  KEY <- 'toto'

  on.exit(delete_config(KEY), add = TRUE)
  delete_config(KEY)

  ### get_config
  # no config set
  expect_null(get_config(KEY))

  # set a value
  expect_null(set_config(KEY, 'value'))
  expect_identical(get_config(KEY), 'value')

  # set another value
  value <- head(iris, 5)
  expect_identical(set_config(KEY, value), 'value')
  expect_identical(get_config(KEY), value)
  
  # delete: with value
  expect_identical(delete_config(KEY), value)
  expect_null(get_config(KEY))
  
  # delete: no value
  expect_null(delete_config(KEY))
  expect_null(get_config(KEY))
})

# implements my view on when a condition is true, e.g. length(1) is true


# wrapper of sprintf that automatically collapses vectors
enhanced_sprintf <- function(fmt, ..., collapse = ',') {
  if (length(fmt) != 1) stop('Arg "format" must be a scalar (length 1)')

  dots <- list(...)
  if (!length(dots)) return(fmt)

  .collapse_args <- function(x) {
    if (length(x) > 1) {
      paste0(x, collapse = collapse) 
    } else {
      if (length(x) == 0) '' else x
    }
  }
  fixed_dots <- lapply(dots, .collapse_args)

  do.call(sprintf, c(list(fmt), fixed_dots))
}

# for building error messages, we must not crash on bad input
safe_enhanced_sprintf <- function(fmt, ...) {
  msg <- try(enhanced_sprintf(fmt, ...))
  if (is_error(msg)) {
    msg <- paste0('could not build error message: "', extract_error_message(msg), '"')
  }

  msg
}

extract_error_message <- function(err) {
  conditionMessage(attr(err, 'condition'))
}

is_condition_true <- function(cond) {
  if (length(cond) != 1) stop('condition must be a scalar')
  if (is.na(cond)) stop('condiction must not be missing (NA)')

  as.logical(cond)
}

# is_string <- function(x) length(x) == 1 && is.character(x)
is_nz_string <- function(x)  length(x) == 1 && is.character(x) && nzchar(x)

is_error <- function(x) inherits(x, "try-error")

stop_unless <- function(cond, format, ...) {
	if (missing(format)) {
		stop("Argument 'format' is missing in die_unless()")
	}
  if (!is_condition_true(cond)) {
    msg <- safe_enhanced_sprintf(format, ...)
    # N.B: we do not want the call to be part of the message
    stop(msg, call. = FALSE)
  }

  invisible()
}

stop_if <- function(cond, ...) {
  stop_unless(!cond, ...)
}

mute <- function(..., messages = TRUE, warnings = TRUE, output = TRUE) {
  msg <- if (messages) suppressMessages else identity
  warn <- if (warnings) suppressWarnings else identity
  out <- identity
  if (output) {
    out <- function(...) {
        utils::capture.output(msg <- utils::capture.output(...), type = 'message')
        msg
    }
  } 

  invisible(out(warn(msg(...))))
}

# sorting strings depend on locales, and as such may not be reproducible across platforms and settings
reproducible_sort <- function(...) {
  old_ctype <- Sys.getlocale('LC_CTYPE')
  old_collate <- Sys.getlocale('LC_COLLATE')
  on.exit({
    Sys.setlocale('LC_CTYPE', old_ctype)
    Sys.setlocale('LC_COLLATE', old_collate)
  }, add = TRUE)

  Sys.setlocale('LC_CTYPE', 'C')
  Sys.setlocale('LC_COLLATE', 'C')

  sort(...)
}

get_elements <- function(lst, idx) {
  lapply(lst, getElement, idx)
}

"%||%" <- function(a, b) {
  if (length(a) == 0) b else a
}

file_size <- function(path) {
  file.info(path, extra_cols = FALSE)$size
}

fast_unlist <- function(x, recursive = FALSE, use.names = FALSE) {
  unlist(x, recursive, use.names)
}

get_text_logger <- function(quiet = FALSE) { if (quiet) function(...) {} else cli::cli_text }

get_env <- function(name, default = '') {
  value <- Sys.getenv(name)
  if (length(value) && !is.na(value) && nzchar(value)) value else default
}

set_env <- function(var, value) {
  args <- list(value)
  names(args) <- var
  do.call(Sys.setenv, args)
}


set_option <- function(option, value) {
  args <- list(value)
  names(args) <- option
  do.call(options, args)
}

# subset a list with a s3 class without losing the class attribute
subset_s3_list <- function(lst, idx) {
  subset_lst <- lst[idx]
  class(subset_lst) <- class(lst)
  subset_lst
}

.subset_s3_list <- 
test_that(".subset_s3_list", {
  ### regular list
  lst <- as.list(1:5)
  expect_identical(subset_s3_list(lst, 2), lst[2])
  
  ###
  class(lst) <- "toto"
  res <- subset_s3_list(lst, 2)
  expect_identical(unclass(res), lst[2])
  expect_identical(class(res), "toto")
  
  res <- subset_s3_list(lst, 2:3)
  expect_identical(unclass(res), lst[2:3])
  expect_identical(class(res), "toto")
})



test_that("get_text_logger", {
  expect_identical(get_text_logger(), cli::cli_text)
  logger <- get_text_logger(quiet = TRUE)
  expect_identical(as.character(body(logger)), '{')
})

test_that("is_condition_true", {
  is_condition_true <- srcpkgs:::is_condition_true

  expect_true(is_condition_true(TRUE))
  expect_true(is_condition_true(length("A")))
  expect_true(is_condition_true(1L))
  
  expect_false(is_condition_true(FALSE))
  expect_false(is_condition_true(length(NULL)))
  expect_false(is_condition_true(0L))

  expect_error(is_condition_true(NA), 'missing')
  expect_error(is_condition_true(1:2), 'scalar')
  expect_error(is_condition_true(NULL), 'scalar')
})


test_that("enhanced_sprintf", 
{
  enhanced_sprintf <- srcpkgs:::enhanced_sprintf

  ### edge cases
  expect_identical(enhanced_sprintf(NA), NA)
  expect_identical(enhanced_sprintf(''), '')
  expect_identical(enhanced_sprintf('toto'), 'toto')
  expect_error(enhanced_sprintf(NULL), 'must be a scalar')
  expect_error(enhanced_sprintf(c('toto', 'titi')), 'must be a scalar')

  # empty sprintf input
  expect_identical(enhanced_sprintf('A%sB', NULL), 'AB')

  # normal behaviour
  expect_identical(enhanced_sprintf('toto%i %s %i', 1, 'titi', 5), 'toto1 titi 5')

  ### vectorized input
  # ints => must use %s
  expect_error(enhanced_sprintf('toto%i', 1:3), 'invalid format')
  expect_identical(enhanced_sprintf('toto %s', 1:3), 'toto 1,2,3')

  ### vectorized strings
  expect_identical(enhanced_sprintf('toto %s', LETTERS[1:3]), 'toto A,B,C')
  expect_identical(enhanced_sprintf("i = %i s = %s",3,'toto'), 'i = 3 s = toto')
})


test_that("safe_enhanced_sprintf", 
{
  safe_enhanced_sprintf <- srcpkgs:::safe_enhanced_sprintf


  # normal behaviour
  expect_identical(safe_enhanced_sprintf('toto%i %s %i', 1, 'titi', 5), 'toto1 titi 5')
  expect_identical(safe_enhanced_sprintf('toto %s', 1:3), 'toto 1,2,3')

  # safe behaviour
  expect_match(mute(safe_enhanced_sprintf("", stop('error'))), "could not build error message")

})

test_that("stop_unless", 
{
  stop_unless <- srcpkgs:::stop_unless

  ### no stop
  expect_error(stop_unless(TRUE, 'argh'), NA)
  expect_error(stop_unless(length(1), 'argh'), NA)
  # test if the message args are evaluated even if the condition is TRUE
  expect_error(stop_unless(TRUE, stop('beurk')), NA)

  ### expected stop
	expect_error(stop_unless(FALSE, 'argh'), 'argh')
  expect_error(stop_unless(length(NULL), 'argh'), 'argh')
  # sprintf message
	expect_error(stop_unless(FALSE, "i = %i s = %s",3,'toto'), "i = 3 s = toto")
  expect_error(stop_unless(FALSE, 'ints=%s', 1:3), 'ints=1,2,3')

  ### errors
  # error in the call --> bad params
  expect_error(stop_unless(TRUE), 'format')

  # error in the condition
  expect_error(stop_unless(1:2, 'argh'), 'scalar')
  expect_error(stop_unless(NA, 'argh'), 'missing')
  expect_error(stop_unless(NULL == 1, 'Argh'), 'scalar')
})

test_that("stop_if", 
{
  stop_if <- srcpkgs:::stop_if

  ### no stop
  expect_error(stop_if(FALSE, 'argh'), NA)
  expect_error(stop_if(length(NULL), 'argh'), NA)
  # test if the message args are evaluated even if the condition is TRUE
  expect_error(stop_if(FALSE, stop('beurk')), NA)

  ### expected stop
	expect_error(stop_if(TRUE, 'argh'), 'argh')
  expect_error(stop_if(length(1), 'argh'), 'argh')
  # sprintf message
	expect_error(stop_if(TRUE, "i = %i s = %s",3,'toto'), "i = 3 s = toto")
  expect_error(stop_if(TRUE, 'ints=%s', 1:3), 'ints=1,2,3')

  ### errors
  # error in the call --> bad params
  expect_error(stop_if(FALSE), 'format')

  # error in the condition
  expect_error(stop_if(1:2, 'argh'), 'scalar')
  expect_error(stop_if(NA, 'argh'), 'missing')
  expect_error(stop_if(NULL == 1, 'Argh'), 'scalar')
})

test_that("reproducible_sort", {
  expect_identical(reproducible_sort(rev(LETTERS)), LETTERS)
})


test_that("file_size", {
  setup_temp_dir()
  path <- 'toto'

  # empty file
  writeLines(character(0), path)
  expect_equal(file_size(path), 0)
  
  # at least one char
  writeLines('', path)
  expect_gt(file_size(path), 0)
})

# checks if the package has changed since the last pkg_write_md5sum() call
# if no md5sum file exists, return TRUE since we can not know, so we assume it has changed
pkg_has_changed <- function(pkg_path, md5sum = pkg_read_md5sum(pkg_path), quiet = FALSE) {
  if (!length(md5sum)) return(TRUE)

  rel_paths <- pkg_list_files(pkg_path)

  old <- setwd(pkg_path)
  on.exit(setwd(old), add = TRUE)

  change <- find_first_modified_file(rel_paths, md5sum)
  if (!length(change)) return(FALSE)

  if (!quiet) {
    file <- attr(change, 'name') # names() refuse to work here for an unknown reason
    cli::cli_text('package {.pkg {pkg_path}} has changed: {.file {file}} was {.str {change}}')
  }

  TRUE
}


# list files really part of the source package, i.e. the files that should be part of the build (
# (R CMD build) tarball
# N.B: by default exclude lots of files, but also make use of .Rbuildignore to exclude more
# N.B: return relative file paths
pkg_list_files <- function(pkg_path) {
  stop_unless(is_nz_string(pkg_path), "bad param 'pkg_path', must be a string")
  stop_unless(dir.exists(pkg_path), "pkg_path '%s' does not exist", pkg_path)

  ### exclusion
  exclude <- default_exclude_patterns()

  # use .Rbuildignore
  ignore_file <- file.path(pkg_path, ".Rbuildignore")
  if (file.exists(ignore_file))
    exclude <- append(exclude, readLines(ignore_file, warn = FALSE))
  exclude <- exclude[nzchar(exclude)]

  exclude_pattern <- paste0(paste0('(', exclude, ')'), collapse = '|')

  paths <- dir(pkg_path, recursive = TRUE, full.names = FALSE, no.. = TRUE)
  paths <- grep(exclude_pattern, paths, invert = TRUE, value = TRUE)

  paths
}

test_that("pkg_has_changed", {
  setup_temp_dir()
  pkg_create('.', 'AA')
  pkg_path <- file.path(getwd(), 'AA')

  ### no md5sum
  expect_true(pkg_has_changed(pkg_path))

  ### fresh md5sum
  pkg_write_md5sum(pkg_path)
  expect_false(pkg_has_changed(pkg_path))

  ### add a new file
  new_file_path <- file.path(pkg_path, 'R', 'toto.R')
  writeLines('coucou <- NULL', new_file_path)
  expect_true(pkg_has_changed(pkg_path, quiet = TRUE))

  # refresh
  pkg_write_md5sum(pkg_path)
  expect_false(pkg_has_changed(pkg_path))

  ### modify a file
  writeLines('coucou <- "hello"', new_file_path)
  expect_true(pkg_has_changed(pkg_path, quiet = TRUE))

  # refresh
  pkg_write_md5sum(pkg_path)
  expect_false(pkg_has_changed(pkg_path, quiet = TRUE))

  ### delete a file
  unlink(new_file_path)
  mute(expect_true(pkg_has_changed(pkg_path)))

  # refresh
  md5sum_path <- pkg_write_md5sum(pkg_path)
  expect_false(pkg_has_changed(pkg_path))

  ### delete the md5sum file
  unlink(md5sum_path)
  expect_true(pkg_has_changed(pkg_path))
})


test_that("pkg_list_files", {
  setup_temp_dir()

  pkg_create('.', 'mypkg')

  rel_paths <- pkg_list_files('mypkg')
  paths <- file.path('mypkg', rel_paths)

  common <- c('DESCRIPTION', 'R/dummy.R')
  expect_true(all(file.path('mypkg', common) %in% paths))
  expect_true(all(file.exists(paths)))

  ### ignored files
  .touch <- function(file)  writeLines('', file.path('mypkg', file))
  .present <- function(file) file %in% pkg_list_files('mypkg')
  
  .touch('non_hidden_root_level')
  expect_true(.present('non_hidden_root_level'))

  # add it to the .Rbuildignore file
  cat('^non_hidden_root_level$', file = 'mypkg/.Rbuildignore', append = TRUE)
  # now excluded
  expect_false(.present('non_hidden_root_level'))
  # N.B: .Rbuildignore is also automatically excluded
  expect_false(.present('.Rbuildignore'))

  ### hidden files: automatically excluded
  .touch('.hidden')
  expect_false(.present('.hidden'))
})

#' roxygenize a source package if needed 
#' 
#' - if the package has not changed (based on the md5sum file), does nothing
#' - otherwise roxygenise the package using roxygen2::roxygenise
#' - and update and save the new md5sum file
#' 
#' - N.B: has the side-effect of loading the package
#' 
#' @inheritParams params
#' @param force	      if force(d), do not use the md5-based system to detect package changes
#' @param ...         passed to  [devtools::document()]
#'
#' @return if the roxygenation has been performed
#' @export
#' @examples
#'  pkg <- setup_and_get_dummy_srcpkg()
#'  pkg_roxygenise(pkg$path)
pkg_roxygenise <- function(pkg_path, force = FALSE, quiet = FALSE,  ...) {
  if (!force && !pkg_needs_roxygen(pkg_path, quiet = quiet)) 
    return(invisible(FALSE))

  pkg_roxygenise_wrapper(pkg_path, quiet = quiet, ...)
  invisible(TRUE)
}

pkg_roxygenise_wrapper <- function(pkg_path, quiet = FALSE,  ...) {
  wrapper <- if (quiet) suppressMessages else function(...) { force(...) }
  wrapper(devtools::document(pkg_path, quiet = quiet, ...))
  pkg_write_md5sum(pkg_path)
}


# delete the roxygen-generated documentation files
pkg_delete_doc <- function(pkg_path) {
  old <- setwd(pkg_path)
  on.exit(setwd(old), add = TRUE)

  unlink('NAMESPACE')
  unlink('man', recursive = TRUE)
}

# even though a package has not changed, the documentation may never had been generated
pkg_has_no_doc <- function(pkg_path) {
  ### if no NAMESPACE nor man/ dir --> never roxygenized --> TRUE
  # N.B: I do not remember why we test the emptiness. Probably at one point was created empty
  namespace <- file.path(pkg_path, 'NAMESPACE')
  if (!file.exists(namespace) || file_size(namespace) == 0) return(TRUE)
  if (!dir.exists(file.path(pkg_path, 'man'))) return(TRUE)

  FALSE
}

# tests if the roxygen-generated documentation is outdated
pkg_needs_roxygen <- function(pkg_path, quiet = FALSE) {
  pkg_has_no_doc(pkg_path) || pkg_has_changed(pkg_path, quiet = quiet)
}

test_that("pkg_roxygenise", {
  setup_temp_dir()
  pkg <- pkg_create('.', 'AA')
  src_pkgs <- srcpkgs(list(pkg))
  on.exit(cleanup_dangling_srcpkgs(), add = TRUE)
  unloadNamespace('AA')
  cleanup_dangling_srcpkgs()

  roxy <- "
#' dummy roxy item
#' @param param1 desc1
#' @param param2 desc2
#' @return stuff
#' @export
dummy_function <- function(param1, param2) {}
  "
  writeLines(roxy, 'AA/R/roxy.R')

  ### before
  files <- pkg_list_files('AA')
  expect_true('DESCRIPTION' %in% files)
  expect_false('NAMESPACE' %in% files)

  ### pkg_roxygenise
  expect_false(pkg_is_loaded('AA'))

  expect_true(pkg_roxygenise('AA', quiet = TRUE))

  expect_true('NAMESPACE' %in% pkg_list_files('AA'))
  expect_true('man/dummy_function.Rd' %in% pkg_list_files('AA'))
  # N.B: as a side-effect, the package is now loaded
  expect_true(pkg_is_loaded('AA'))

  ### again
  expect_false(pkg_roxygenise('AA', quiet = TRUE))
  

  ### force regeneration
  cat('# srcpkgs rules!\n', file = 'AA/NAMESPACE', append = TRUE)
  content <- readLines('AA/NAMESPACE')
  expect_match(content, 'srcpkgs', all = FALSE)

  expect_true(pkg_roxygenise('AA', quiet = TRUE, force = TRUE))

  content <- readLines('AA/NAMESPACE')
  expect_no_match(content, 'srcpkgs', all = FALSE)

  ### pkg_delete_doc()
  pkg_delete_doc('AA')

  files <- pkg_list_files('AA')
  expect_false('NAMESPACE' %in% files)
  expect_false('man' %in% unique(dirname(files)))

  ### pkg_roxygenise_wrapper quiet=FALSE
  suppressMessages(pkg_roxygenise_wrapper('AA'))
  expect_true('NAMESPACE' %in% pkg_list_files('AA'))
})






test_that("pkg_needs_roxygen", {
  setup_temp_dir()
  pkg <- pkg_create('.', 'AA')
  src_pkgs <- srcpkgs(list(pkg))
  on.exit(cleanup_dangling_srcpkgs(), add = TRUE)

  refresh <- function() {
    pkg_roxygenise('AA', quiet = TRUE, force = TRUE)
    pkg_needs_roxygen('AA', quiet = TRUE)
  }

  ### never roxygenized
  expect_true(pkg_needs_roxygen('AA', quiet = TRUE))

  ### just been roxygenized
  expect_false(refresh())

  ### remove some files
  unlink('AA/NAMESPACE')
  expect_true(pkg_needs_roxygen('AA', quiet = TRUE))
  expect_false(refresh())

  unlink('AA/man', recursive = TRUE)
  expect_true(pkg_needs_roxygen('AA', quiet = TRUE))
  expect_false(refresh())

  ### add file
  writeLines('one <- 1L', 'AA/R/one.R')
  expect_true(pkg_needs_roxygen('AA', quiet = TRUE))
  expect_false(refresh())

  ### add line to existing file
  cat('### comment\n', file = 'AA/R/one.R', append = TRUE)
  expect_true(pkg_needs_roxygen('AA', quiet = TRUE))
  expect_false(refresh())

  ## delete_pkgs_doc
  pkg_delete_doc('AA')
  expect_true(pkg_needs_roxygen('AA', quiet = TRUE))
  expect_false(refresh())
})

#' list all characters
#' @return as a list
#' @importFrom elves elves
#' @export
lotr <- function() {
  list(
    hobbits = hobbits(),
    elves = elves(),
    dwarves = dwarves(),
    humans = humans(),
    ainur = ainur()
  )
}

#' @importFrom gimli gimli
dwarves <- function() {
  list(gimli = gimli())
}

#' @importFrom aragorn aragorn
humans <- function() {
  list(aragorn = aragorn())
}

#' @importFrom gandalf gandalf
ainur <- function() {
  list(gandalf = gandalf())
}

PKG_MD5_FILE <- '.SRCPKGS_MD5'


pkg_write_md5sum <- function(pkg_path, md5sum = pkg_md5sum(pkg_path)) {
  force(md5sum)
  md5sum_path <- file.path(pkg_path, PKG_MD5_FILE)
  write_md5sum_file(md5sum, md5sum_path)
  
  invisible(md5sum_path)
}

pkg_read_md5sum <- function(pkg_path) {
  md5sum_path <- file.path(pkg_path, PKG_MD5_FILE)
  if (!file.exists(md5sum_path)) return(NULL)
  read_md5sum_file(md5sum_path)
}


# compute the MD5sum of a source package
pkg_md5sum <- function(pkg_path) {
  paths <- pkg_list_files(pkg_path)

  old <- setwd(pkg_path)
  on.exit(setwd(old), add = TRUE)

  tools::md5sum(paths)
}


pkg_delete_md5sum <- function(pkg_path) {
  md5sum_path <- file.path(pkg_path, PKG_MD5_FILE)
  file.exists(md5sum_path) && unlink(md5sum_path)
}

# write a named character vector (as returned by tools::md5sum) to a MD5 checksum file
write_md5sum_file <- function(md5sum, path) {
  lines <- paste0(md5sum, '  ', names(md5sum))
  writeLines(lines, path)
}

read_md5sum_file <- function(path) {
  lines <- readLines(path)
  tokens <- strsplit(lines, '\\s+')
  mat <- do.call(rbind, tokens)

  structure(mat[, 1], names = mat[, 2])
}

find_first_modified_file <- function(paths, md5sum) {
  paths2 <- names(md5sum)
  
  new <- setdiff(paths, paths2)
  if (length(new)) {
    # there is a new file, not in md5sum
    return(structure('added', name = new[1]))
  }

  deleted <- setdiff(paths2, paths)
  if (length(deleted)) {
    # a file has been deleted, was in md5sum but no longer in paths
    return(structure('deleted', name = deleted[1]))
  }

  # check the md5 file by file
  tools_md5sum <- tools::md5sum
  for (path in paths) {
    if (tools_md5sum(path) != md5sum[path])
      return(structure('modified', name = path))
  }

  NULL
}


# copied from tools:::get_exclude_patterns() since it is not exported
default_exclude_patterns <- function() {
  c(
    "^\\.Rbuildignore$",
    "(^|/)\\.DS_Store$",
    "^\\.(RData|Rhistory)$",
    "~$", "\\.bak$", "\\.swp$",
    "(^|/)\\.#[^/]*$", "(^|/)#[^/]*#$",
    ## Outdated ...
    "^TITLE$", "^data/00Index$",
    "^inst/doc/00Index\\.dcf$",
    ## Autoconf
    "^config\\.(cache|log|status)$",
    "(^|/)autom4te\\.cache$", # ncdf4 had this in subdirectory 'tools'
    ## Windows dependency files
    "^src/.*\\.d$", "^src/Makedeps$",
    ## IRIX, of some vintage
    "^src/so_locations$",
    ## Sweave detrius
    "^inst/doc/Rplots\\.(ps|pdf)$"
    ## Karl: some custom additions
    ,"^Read-and-delete-me"
  )
}

test_that("pkg_write_md5sum", {
  setup_temp_dir()
  pkg_create('.', 'mypkg')

  md5sum <- pkg_md5sum('mypkg')

  files <- pkg_list_files('mypkg')
  expect_identical(names(md5sum), files)
  expect_true(all(nzchar(md5sum)))

  ### pkg_read_md5sum
  expect_null(pkg_read_md5sum('./mypkg/'))

  ### pkg_write_m5sum
  md5_path <- pkg_write_md5sum(file.path(getwd(), 'mypkg'))

  expect_identical(read_md5sum_file(md5_path), md5sum)
  expect_identical(pkg_read_md5sum('mypkg'), md5sum)
})


test_that("md5sum_file ", {
  setup_temp_dir()
  pkg_create('.', 'pkg', namespace = TRUE)

  paths <- file.path('pkg', pkg_list_files('pkg'))
  md5sum <- tools::md5sum(paths)
  
  ### write_md5sum_file
  write_md5sum_file(md5sum, 'MD5')

  # read back
  md5sum2 <- read_md5sum_file('MD5')
  expect_identical(md5sum2, md5sum)

  expect_null(find_first_modified_file(paths, md5sum2))
  
  # modify a md5
  x <- md5sum
  x[3] <- x[1]
  expect_identical(find_first_modified_file(paths, x), structure('modified', name = paths[3]))

  # delete a file from md5sums
  x <- md5sum[-2]
  expect_identical(find_first_modified_file(paths, x), structure('added', name = paths[2]))

  # add a file in md5sums
  x <- md5sum
  x['toto'] <- x[1]
  expect_identical(find_first_modified_file(paths, x), structure('deleted', name = 'toto'))
})

# meta data about loaded source packages



# fetches meta data about loaded source packages as a data frame, or NULL if no info
#  N.B: use the devtools meta data, and stores the package hash inside it
fetch_srcpkgs_meta <- function() {
  pkgs_names <- sort(dev_packages_fixed())
  pkgs <- lapply(pkgs_names, fetch_srcpkg_meta)
  pkgs <- pkgs[lengths(pkgs) > 0] %||% return(NULL)
  pkgs <- srcpkgs(pkgs)

  df <- as.data.frame(pkgs)

  df$attached <- !is.na(match(rownames(df), sub('package:', '', search())))

  df
}

# devtools::dev_packages() does not work any longer in devtools >= 2.5
dev_packages_fixed <- function() {
  loaded <- loadedNamespaces()
  have_meta <- sapply(loaded, function(x) !is.null(pkgload::dev_meta(x)))
  loaded[have_meta]
}

get_or_set_meta <- function(meta_env, value) {
  key <- 'SRCPKGS_META_KEY'
  if (missing(value)) {
    get0(key, envir = meta_env, inherits = FALSE)
  } else {
    assign(key, value, envir = meta_env)
  }
}

# returns the srcpkg object stored in the devtools meta env of any, or NULL
fetch_srcpkg_meta <- function(pkg_name) {
  if (!pkg_is_loaded(pkg_name)) return(NULL)
  meta_env <- pkgload::dev_meta(pkg_name)
  if (is.null(meta_env)) return(NULL)
  get_or_set_meta(meta_env)
}

# stores the srcpkg, or dies
store_srcpkg_meta <- function(srcpkg) {
  pkg_name <- srcpkg$package
  meta_env <- pkgload::dev_meta(pkg_name)
  stop_if(is.null(meta_env), 'error, no devtools meta env for package "%s"', pkg_name)
  get_or_set_meta(meta_env, srcpkg)
}

## N.B: in some cases, srcpkgs itself is registered in meta and that affects the tests
# so we remove it
{
  meta_env <- pkgload::dev_meta('srcpkgs')
  if (!is.null(meta_env)) get_or_set_meta(meta_env, NULL)
}

test_that("fetch_srcpkgs_meta", {
  setup_temp_dir()
  cleanup_dangling_srcpkgs()
  
  pkg1 <- pkg_create('.', 'pkg1')
  pkg2 <- pkg_create('.', 'pkg2')

  on.exit(cleanup_dangling_srcpkgs(), add = TRUE)

  ### no dev meta --> no srcpkg meta data
  # pkg not loaded (by devtools at least) 
  expect_null(fetch_srcpkg_meta('pkg1'))
  # pkg loaded (base) but not by devtools
  expect_null(fetch_srcpkg_meta('devtools'))

  expect_null(fetch_srcpkgs_meta())

  ## load pkg1
  on.exit(unloadNamespace('pkg1'), add = TRUE)
  devtools::load_all('pkg1', quiet = TRUE, attach = FALSE) 

  ## still no srcpkg meta
  expect_null(fetch_srcpkg_meta('pkg1'))
  expect_null(fetch_srcpkgs_meta())

  ## store the srcpkg in meta data
  srcpkg1 <- srcpkg(pkg1)
  store_srcpkg_meta(srcpkg1)

  # now that works
  expect_identical(fetch_srcpkg_meta('pkg1'), srcpkg1)

  # fetch_srcpkgs_meta() also works
  df <- fetch_srcpkgs_meta()

  expect_true(is.data.frame(df))
  expect_identical(df$package, 'pkg1')
  expect_false(df$attached)


  ### load pkg2, and attach it
  on.exit(unloadNamespace('pkg2'), add = TRUE)
  devtools::load_all('pkg2', quiet = TRUE, attach = TRUE) 

  # no meta data yet
  expect_identical(fetch_srcpkgs_meta()$package, 'pkg1')

  # store its meta data
  srcpkg2 <- srcpkg(pkg2)
  store_srcpkg_meta(srcpkg2)

  # check
  expect_identical(fetch_srcpkg_meta('pkg2'), srcpkg2)

  # fetch_srcpkgs_meta() also works
  df <- fetch_srcpkgs_meta()
  
  expect_true(is.data.frame(df))  
  expect_true(all(c('pkg1', 'pkg2') %in% df$package))
  df <- df[c('pkg1', 'pkg2'), ]
  expect_setequal(df$attached, c(TRUE, FALSE))

  # unload them
  unloadNamespace('pkg1')
  unloadNamespace('pkg2')

  # no longer works
  expect_null(fetch_srcpkg_meta('pkg1'))
})

#' installs a list of source packages
#'
#' - A source package can not be installed if its dependencies are not. 
#' - Will not reinstall packages if they are up-to-date
#' - will roxygenise packages if needed
#
#' @inheritParams params
#' @param only_deps   whether not to include `pkgids`, only their dependencies.
#' @param ...   passed to `devtools::install()`
#' @return the names of the packages actually installed
#' @export
pkgs_install <- function(pkgids, lib, src_pkgs = get_srcpkgs(), only_deps = FALSE,  quiet = TRUE, ...)
{
  stop_unless(length(pkgids), "No package to test")
  force(src_pkgs)
  pkgs <- as_srcpkgs(pkgids, src_pkgs)

  pkgs_to_install <- pkgs_deps(names(pkgs), src_pkgs, installed = FALSE)
  # N.B: must process them in reverse topological order
  pkgs_to_install <- rev(pkgs_to_install)

  if (!dir.exists(lib)) dir.create(lib, recursive = TRUE)

  .process_pkg <- function(pkg_name) {
    pkg <- src_pkgs[[pkg_name]]

    # is_loaded <- pkg_is_loaded(pkg_name)
    # pkg_roxygenise(pkg$path, quiet = quiet)
    pkg_load(pkg, src_pkgs, attach = FALSE, roxygen = TRUE, quiet = quiet)
    # if (!is_loaded) unloadNamespace(pkg_name)
    
    if (!pkg_needs_install(pkg$path, lib)) return(NULL)
    pkg_install_nodeps(pkg$path, lib, quiet = quiet, ...)
    pkg_name
  }
  res <- fast_unlist(lapply(pkgs_to_install, .process_pkg))
  if (!only_deps) {
    res2 <- fast_unlist(lapply(names(pkgs), .process_pkg))
    res <- c(res, res2)
  }

  res
}


# just install the package inconditionnally
# wrapper on devtools::install + store the MD5 sum
# N.B: do not roxygenize, nor modify the source package in any way
pkg_install_nodeps <- function(pkg_path, lib, md5 = pkg_md5sum(pkg_path), quiet = FALSE, quick = TRUE, ...)
{
  stop_unless(is_nz_string(pkg_path), "bad param 'pkg_path', must be a string")
  stop_unless(dir.exists(pkg_path), "pkg_path '%s' does not exist", pkg_path)
  stop_unless(dir.exists(lib), "directory '%s' does not exist", lib)

  pkg <- srcpkg(path = pkg_path)

  libpath <- c(lib, .libPaths())

  withr::with_libpaths(libpath, devtools::install(pkg_path, reload = FALSE, dependencies = FALSE, upgrade = FALSE, quick = quick, quiet = quiet, ...))

  # store the MD5 of the pkg for this installation
  installed_pkg_path <- file.path(lib, pkg$package)
  pkg_write_md5sum(installed_pkg_path, md5)

  invisible(md5)
}

pkg_needs_install <- function(pkg_path, lib, md5 = pkg_md5sum(pkg_path)) {
  pkg <- srcpkg(path = pkg_path)
  installed_pkg_path <- file.path(lib, pkg$package)

  force(md5)
  !identical(pkg_read_md5sum(installed_pkg_path), md5)
}

.pkgs_install <- 
test_that(".pkgs_install", {
  # G-->C-->B-->A--->stats, F-->D-->B, E-->A, Z, G--->utils
  src_pkgs <- examples_srcpkgs_complex_imports()
  LIB <- 'lib'

  ### a package with deps
  res <- pkgs_install("GG", LIB, src_pkgs = src_pkgs, quiet = TRUE)
  
  expect_identical(res, c("AA", "BB", "CC", "GG"))
  expect_equal(installed.packages(LIB)[, "Package"], res, ignore_attr = TRUE)

  # again
  expect_null( pkgs_install("GG", LIB, src_pkgs = src_pkgs, quiet = TRUE) )

  ### a package with some deps already installed
  res <- pkgs_install("FF", LIB, src_pkgs = src_pkgs, quiet = TRUE)
  expect_identical(res, c("DD", "FF"))

  # only deps
  expect_null( pkgs_install("EE", LIB, src_pkgs = src_pkgs, only_deps = TRUE, quiet = TRUE) )

  ### multiple packages
  res <- pkgs_install(names(src_pkgs), LIB, src_pkgs = src_pkgs, quiet = TRUE)
  expect_identical(res, c("EE", "ZZ"))
})


.pkg_install_nodeps <- 
test_that("pkg_install_nodeps", {
  # A->B->C->D, B->D->E, Z
  src_pkgs <- examples_srcpkgs_complex_deps()
  LIB <- 'toto'

  expect_error(pkg_install_nodeps(src_pkgs[[1]]$path, LIB), "does not exist")

  dir.create(LIB)

   # no NAMESPACE file yet
  expect_error(pkg_install_nodeps(src_pkgs$ZZ$path, LIB, quiet = TRUE), "failed")
  ## pkg_roxygenise --> generate the NAMESPACE
  pkg_roxygenise(src_pkgs$ZZ$path, quiet = TRUE)
  pkg_unload(src_pkgs$ZZ, quiet = TRUE)

  md5 <- pkg_install_nodeps(src_pkgs$ZZ$path, LIB, quiet = TRUE)
  
  expect_identical(installed.packages(LIB)[, "Package"], "ZZ")
  expect_identical(md5, pkg_md5sum(src_pkgs$ZZ$path))  
  expect_true(dir.exists(file.path(LIB, "ZZ")))

  expect_false(pkg_needs_install(src_pkgs$ZZ$path, LIB))

  ## modify the source package
  writeLines("# sorry", file.path(src_pkgs$ZZ$path, "R/toto.R"))
  expect_true(pkg_needs_install(src_pkgs$ZZ$path, LIB))
})

# N.B: the current dir should be set above that folder
library(devtools)
load_all('srcpkgs', export_all = FALSE)

library(lotr)

get_srcpkgs()



pkg_load('lotr')
pkg_load('hobbits')
hobbits()
pkg_unload('hobbits')

execute_plan <- function(plan, src_pkgs, quiet = FALSE, ...) {
  if (is_loaders_hack_enabled()) {
    unhack_r_loaders()
    on.exit(hack_r_loaders(), add = TRUE)
  }

  if (!length(plan)) return()
  logger <- get_text_logger(quiet)
  for (i in seq_len(nrow(plan))) {
    row <- plan[i, , drop = TRUE]
    params <- row$params[[1]]
    logger('executing {.strong {row$action} on {.pkg {row$package}}}')
    switch(row$action, 
      unload = execute_action_unload(row$package, quiet = quiet, ...),
      load = do.call(execute_action_load, 
        c(list(as_srcpkg(row$package, src_pkgs), quiet = quiet, ...), params)),
      stop_if(TRUE, 'unknown action: "%s"', row$action)
    )
  }
}

execute_action_unload <- function(pkg_name, quiet, ...) {
  pkgload::unregister(pkg_name)
  # N.B: shoud not happen with source packages
  # # check since sometimes it may fail (e.g. for rlang)
  # if (pkg_is_loaded(pkg_name)) {
  #   if (!quiet) 
  #     cli::cli_text('failed to {.emph unregister} {.pkg pkg_name}, retrying with {.fn unload}')
  #   pkgload::unload(pkg_name)

  #   if (pkg_is_loaded(pkg_name)) {
  #     if (!quiet) 
  #       cli::cli_text('failed to {.emph unload} {.pkg pkg_name}, giving up...')
  #   }
  # }
}

execute_action_load <- function(pkg, quiet, attach = FALSE, roxygen = FALSE, ...) {
  pkg_load_wrapper(pkg, attach = attach, roxygen = roxygen, quiet = quiet, ...)
}

# Note: this should be read-only
dummy_srcpkg_path <- function() {
  system.file("extdata/examples/dummy_srcpkg", package = 'srcpkgs')
}

#' installs the dummy srcpkg in a temp location
#'
#' Intended for testing and to write examples
#'
#' @param dest  where to install the dummy srcpkg
#' @return the package as a `srcpkg` object
#' @export
#' @examples
#' pkg <- setup_and_get_dummy_srcpkg()
#' print(pkg)
setup_and_get_dummy_srcpkg <- function(dest = tempfile()) {
  pkg_path <- dummy_srcpkg_path()
  if (!dir.exists(dest)) dir.create(dest)
  file.copy(pkg_path, dest, recursive = TRUE, overwrite = TRUE)

  as_srcpkg(file.path(dest, basename(pkg_path)), src_pkgs = NULL)
}

.dummy_srcpkg_path <- 
test_that("dummy_srcpkg_path", { 
  expect_true(dir.exists(dummy_srcpkg_path()))
})


.setup_and_get_dummy_srcpkg <- 
test_that("setup_and_get_dummy_srcpkg", {

  pkg <- setup_and_get_dummy_srcpkg()
  on.exit(unlink(dirname(pkg$path), recursive = TRUE), add = TRUE)

  expect_s3_class(pkg, "srcpkg")
  expect_identical(pkg$package, "dummy.srcpkg")

  ## what if the directory and package already exist
  pkg2 <- setup_and_get_dummy_srcpkg(dirname(pkg$path))
  expect_identical(pkg2, pkg)

  ## testing
  srcpkgs <- srcpkgs(list(pkg))
  res <- pkg_test(pkg, src_pkgs = srcpkgs, reporter = "silent")

  expect_s3_class(res, "pkg_test")
  df <- summary(res)

  expect_setequal(df$failed, c(1, 0))

  # ## checking
  # chk <- pkg_check(pkg, src_pkgs = srcpkgs, error_on = "never", quiet = TRUE)

  # expect_s3_class(chk, "pkg_check")
})

# # @param attach          whether the package should be attached to the search path

#' loads or reloads if needed a source package, taking care of its dependencies
#'
#' N.B: the defaults are different from [devtools::load_all()]: the helpers are not loaded, only 
#' the functions tagged as *exported* are actually exported. The intended goal is to make it as similar
#' to the behaviour of the R loaders.
#' 
#' 
#' This the workhorse function of the package, called by [library()] and [loadNamespace()] 
#' when hacked (cf [hack_r_loaders()].
#' 
#' This function will check that all dependent packages are up-to-date, and
#' document and reload them as needed.
#'
#' To be able to properly load a package, its dependent source packages
#' must be loaded in proper order.
#' i.e. if A-->B-->C, the load order must be C, B, A
#'
#' @inheritParams params
#' @inheritParams pkgload::load_all
#' @inheritDotParams devtools::load_all
#' @param suggests whether to load suggested packages. if TRUE, the suggested are processed
#'    like imports
#' @param roxygen whether to automatically roxygenise packages (if needed)
#' @param ... passed to 
#' @return the load plan as a data frame, or NULL if there is nothing to do.
#' @export
#' @examples
#' root <- tempfile()
#' pkg <- setup_and_get_dummy_srcpkg(root)
#' reset(root)
#' # load and attach a package
#' pkg_load(pkg)
#' 
#' # just load, do not attach it (~ loadNamespace())
#' pkg_unload(pkg)
#' pkg_load(pkg, attach = FALSE)
#' 
#' # do some changes, to a source package or any of its depencies or dependents
#' pkg_unload(pkg)
#' plan <- pkg_load(pkg, dry_run = TRUE)
#' # then you can inspect the plan actions
pkg_load <- function(pkgid,
  src_pkgs = get_srcpkgs(),
  attach = TRUE,
  suggests = FALSE,
  roxygen = TRUE,
  helpers = FALSE,
  export_all = FALSE, 
  quiet = FALSE,
  dry_run = FALSE,
  ...)
{
  force(src_pkgs)
  pkg <- as_srcpkg(pkgid, src_pkgs)
  pkg_name <- pkg$package
  stop_unless(pkg_name %in% names(src_pkgs), 
    'package "%s" is not managed by srcpkgs, check `get_srcpkgs()`', pkg_name)

  ### store the state of source packages BEFORE
  # df0 <- fetch_srcpkgs_meta()

  plan <- pkg_load_full_plan(pkg_name, src_pkgs, roxygen = roxygen, attach = attach, 
    suggests = suggests, quiet = quiet, ...)

  if (!dry_run) {
     if (length(plan)) 
     execute_plan(plan, src_pkgs, quiet = quiet, helpers = helpers, export_all = export_all)
  }


  ### currently the plan does not include an "attach" action. So in the case where the package
  ### is already loaded and up to date, we still need to attach it
  if (attach && !pkg_is_attached(pkg_name))
    attachNamespace(pkg_name)

  invisible(plan)
}


pkg_load_full_plan <- function(pkg_name, src_pkgs, loaded = loadedNamespaces(), 
  outdated = NULL, attach = TRUE, roxygen = TRUE, quiet = FALSE, ...) 
{
  mat <- graph_from_srcpkgs(src_pkgs, ...)
  deps <- graph_get_all_dependencies(mat, pkg_name)
  ordering <- rev(c(pkg_name, deps))

  # find out which packages are outdated
  if (is.null(outdated)) {
    .is_outdated <- function(x) pkg_is_outdated(src_pkgs[[x]]$path, roxygen = roxygen, quiet = quiet)
    outdated <- ordering[sapply(ordering, .is_outdated)]
  }

  # we need to unload: all ordering already loaded AND outdated
  need_unload <- intersect(ordering, intersect(outdated, loaded))
  plana <- NULL
  if (length(need_unload)) plana <- unload_plan(need_unload, mat, loaded = loaded)

  # we need to load:
  # - all unloaded packages (no side-effects), even if not part of ordering
  # - all not loaded packages from ordering
  not_loaded <- setdiff(ordering, loaded)
  pkgs_to_load <- union(plana$package, not_loaded)
  planb <- NULL
  if (length(pkgs_to_load)) {
    planb <- load_plan(pkgs_to_load, mat)
    planb <- planb[planb$package %in% pkgs_to_load, , drop = FALSE]

    # add params
    planb$params <- replicate(nrow(planb), list())

    if (roxygen) {
      idx <- which(planb$package %in% outdated)
      for (i in idx) planb$params[[i]] <- list(roxygen = TRUE)
    }

    ## packages listed as Depends must be attached
    deps_mat <- graph_from_srcpkgs(src_pkgs[planb$package], imports = FALSE, suggests = FALSE)
    to_attach <- colnames(deps_mat)[colSums(deps_mat) > 0]
    if (attach)
      to_attach <- intersect(planb$package, append(to_attach, pkg_name))

    # attach requested package if needed
    if (length(to_attach)) {
      pkg_name_idx <- match(to_attach, planb$package)
      for (i in pkg_name_idx)
        planb$params[[i]] <- append(planb$params[[i]], list(attach = TRUE))
    }
  }

  if (length(plana)) plana$params <- replicate(nrow(plana), list())
  plan <- rbind.data.frame(plana, planb, make.row.names = FALSE)
  if (!nrow(plan)) plan <- NULL

  plan
}

# plan to load pkg_names in the right order inconditionally (even if already loaded)
load_plan <- function(pkg_names, mat) {
  plan <- unload_plan(pkg_names, t(mat), loaded = colnames(mat)) %||% return(NULL)

  plan$action <- 'load'

  plan
}


# return TRUE iff the package should be updated: roxigenised, (re)loaded
pkg_is_outdated <- function(pkg_path, roxygen = TRUE, quiet = FALSE) {
  (roxygen && pkg_has_no_doc(pkg_path)) || pkg_has_changed(pkg_path, quiet = quiet)
}

# N.B: should only be called if the package is not loaded or has changed
pkg_load_wrapper <- function(pkg, roxygen = TRUE, attach = FALSE, helpers = FALSE, export_all = FALSE, quiet = FALSE, ...) {
  old_options <- options(rlib_restart_package_not_found = FALSE)
  on.exit(options(old_options), add = TRUE)

  if (roxygen) pkg_roxygenise_wrapper(pkg$path, quiet = quiet)

  devtools::load_all(pkg$path, export_all = export_all, 
    helpers = helpers, quiet = quiet, attach = attach, ...)

  pkg_write_md5sum(pkg$path)
  store_srcpkg_meta(pkg)
}

E <- list()
R <- list(roxygen = TRUE)
RA <- list(roxygen = TRUE, attach = TRUE)
A <- list(attach = TRUE)

.pkg_load <- 
test_that("pkg_load", {
  cleanup_test_packages()

  ### trivial example A->B
  src_pkgs <- examples_srcpkgs_basic()

  NONE <- character()
  ALL <- names(src_pkgs)
  unloadNamespace('AA')
  ###
  plan <- pkg_load('AA', src_pkgs, quiet = TRUE)

  expect_gt(nrow(plan), 1) # may be > 2 if BB already loaded for instance (by another test)
  expect_true(pkg_is_loaded('AA'))
  expect_true(pkg_is_attached('AA'))
  expect_true(pkg_is_loaded('BB'))
  expect_false(pkg_is_attached('BB'))
  
  ### again
  expect_null(pkg_load('AA', src_pkgs, quiet = TRUE))

  # dirty package BB
  writeLines('#hello', 'BB/R/hello.R')

  plan <- pkg_load('AA', src_pkgs, quiet = TRUE)
  expect_equal(nrow(plan), 4)
})

.load_plan <- 
test_that("load_plan", {
  ### A->C->D, B->D
  mat <- graph_from_strings('A->C->D', 'B->D')
  NONE <- character()
  ALL <- LETTERS
  # N.B: A,B,D,C would have also worked
  plan <- load_plan(c('A', 'B'), mat)
  expect_identical(unique(plan$action), 'load')
  expect_identical(plan$package, c('D', 'B', 'C', 'A'))

  expect_identical(load_plan('A', mat)$package, c('D', 'C', 'A'))
  expect_identical(load_plan('B', mat)$package, c('D', 'B'))
  expect_identical(load_plan('C', mat)$package, c('D', 'C'))
  expect_identical(load_plan('D', mat)$package, 'D')
})

# fix regression: we need to attach the package it is already loaded and up to date but not
# attached
.pkg_load__attach <- 
test_that("pkg_load() - attach", {
  cleanup_test_packages()
  ### trivial example A->B
  src_pkgs <- examples_srcpkgs_basic()

  pkg_load('AA', src_pkgs, quiet = TRUE)


  expect_true(pkg_is_attached('AA'))
  pkg_detach('AA')
  expect_false(pkg_is_attached('AA'))

  pkg_load('AA', src_pkgs, quiet = TRUE)
  expect_true(pkg_is_attached('AA'))
})


.pkg_load_full_plan__complex_deps <- 
test_that("pkg_load_full_plan() - examples_srcpkgs_complex_deps", {
  src_pkgs <- examples_srcpkgs_complex_deps()

  NONE <- character()
  ALL <- names(src_pkgs)

  ##################################################
  ### load A
  # none loaded

  plan <- pkg_load_full_plan('AA', src_pkgs, loaded = NONE, outdated = NONE)

  expect_identical(plan$package, c("EE", "DD", "CC", "BB", "AA"))
  expect_identical(unique(plan$action), 'load')
  expect_identical(plan$params, list(A, E, A, E, A))
  
  # all loaded
  plan <- pkg_load_full_plan('AA', src_pkgs, loaded = ALL, outdated = NONE)
  expect_null(plan)

  # all loaded and outdated
  plan <- pkg_load_full_plan('AA', src_pkgs, loaded = ALL, outdated = ALL)

  expected <- data.frame(
    package = c("AA", "BB", "CC", "DD", "EE", "EE", "DD", "CC", "BB", "AA"), 
    action = c(rep("unload", 5), rep("load", 5)))
  expected$params <- list(E, E, E, E, E, RA, R, RA, R, RA)
  expect_identical(plan, expected)
  
  # all loaded, C outdated
  plan <- pkg_load_full_plan('AA', src_pkgs, loaded = ALL, outdated = 'CC')
  expected <- data.frame(
    package = c("AA", "BB", "CC", "CC", "BB", "AA"), 
    action = c(rep("unload", 3), c("load", "load", "load")))
  expected$params <- list(E, E, E, RA, E, A)
  expect_identical(plan, expected)

  ### load Z
  plan <- pkg_load_full_plan('ZZ', src_pkgs, loaded = NONE, outdated = ALL)
  expected <- data.frame(package = 'ZZ', action = "load")
  expected$params <- list(RA)
  expect_identical(plan, expected)

  ### load B
  plan <- pkg_load_full_plan('BB', src_pkgs, loaded = NONE, outdated = NONE)
  expected <- data.frame(package = c("EE", "DD", "CC", "BB"), action = "load")
  expected$params <- list(A, E, E, A)
  expect_identical(plan, expected)

  # all loaded, E outdated
  plan <- pkg_load_full_plan('BB', src_pkgs, loaded = ALL, outdated = 'EE')
  expected <- data.frame(
    package = c("AA", "BB", "CC", "DD", "EE", "EE", "DD", "CC", "BB", "AA"), 
    action = c(rep("unload", 5), rep("load",  5)))
  expected$params <- list(E, E, E, E, E, RA, E, A, A, E)
  expect_identical(plan, expected)
})

.pkg_load_full_plan__star <- 
test_that("pkg_load_full_plan() - star example ", {
  src_pkgs <- examples_srcpkgs_star()

  NONE <- character()
  ALL <- names(src_pkgs)
  ###############################################################################################
  ### load E
  # none loaded, no suggests
  plan <- pkg_load_full_plan('EE', src_pkgs, loaded = NONE, outdated = ALL)

  # N.B: no 'CC' which is in suggests only
  expect_identical(plan$package, c('AA', 'BB', 'DD', 'EE'))
  expect_identical(unique(plan$action), 'load')
  expect_identical(plan$params,  list(R, RA, RA, RA))

  # none loaded, with suggests
  plan <- pkg_load_full_plan('EE', src_pkgs, loaded = NONE, outdated = ALL, suggests = TRUE)

  expect_identical(plan$package, c('AA', 'BB', 'CC', 'DD', 'EE'))
  expect_identical(unique(plan$action), 'load')
  expect_identical(plan$params,  list(R, RA, R, RA, RA))
  
  # all loaded, none outdated
  plan <- pkg_load_full_plan('EE', src_pkgs, loaded = ALL, outdated = NONE, suggests = TRUE)
  expect_null(plan)

  # all loaded, all outdated
  plan <- pkg_load_full_plan('EE', src_pkgs, loaded = ALL, outdated = ALL, suggests = TRUE)
  expected <- data.frame(
    package = c("EE", "AA", "BB", "CC", "DD", "AA", "BB", "CC", "DD", "EE"),
    action = c(rep("unload", 5), rep("load", 5)))
  expected$params <- list(E, E, E, E, E, R, RA, R, RA, RA)
  expect_identical(plan, expected)

  # all loaded, B outdated
  plan <- pkg_load_full_plan('EE', src_pkgs, loaded = ALL, outdated = 'BB')
  expected <- data.frame(
    package = c("EE",  "BB", "BB", "EE"),
    action = c("unload", "unload", "load",  "load"))
  expected$params <- list(E, E, RA, A)
  expect_identical(plan, expected)

  ### load other, like AA
  # all loaded, none outdated
  plan <- pkg_load_full_plan('AA', src_pkgs, loaded = ALL, outdated = NONE)
  expect_null(plan)

  ### all loaded, AA outdated
  plan <- pkg_load_full_plan('AA', src_pkgs, loaded = ALL, outdated = 'AA')
  expected <- data.frame(
    package = c("EE",  "AA", "AA", "EE"),
    action = c("unload", "unload", "load",  "load"))
  expected$params <- list(E, E, RA, E)
  expect_identical(plan, expected)

  ### all loaded, EE outdated -> nothing to do, E is a parent
  plan <- pkg_load_full_plan('AA', src_pkgs, loaded = ALL, outdated = 'EE')
  expect_null(plan)
})

.pkg_load_full_plan__transitivity <- 
test_that("pkg_load_full_plan() - transitivity ", {
  # the deps graph is A-->B--C
  setup_temp_dir()

  pkg_create('.', 'pkgA', imports  = 'pkgB')
  pkg_create('.', 'pkgB', imports  = 'pkgC')
  pkg_create('.','pkgC', suggests = 'roxygen2')
  cleanup_dangling_srcpkgs()
  on.exit(cleanup_dangling_srcpkgs(), add = TRUE)
  src_pkgs <- find_srcpkgs('.')
  NONE <- character()
  ALL <- names(src_pkgs)

  ################### load A #################################################
  ### all loaded and outdated
  plan <- pkg_load_full_plan('pkgA', src_pkgs, loaded = ALL, outdated = ALL)
  expected <- data.frame(
    package = c("pkgA", "pkgB", "pkgC", "pkgC", "pkgB", "pkgA"),
    action = c("unload", "unload", "unload", "load", "load", "load"))
  expected$params <- list(E, E, E, R, R, RA)
  expect_identical(plan, expected)

  ### all loaded, none outdated
  expect_null(pkg_load_full_plan('pkgA', src_pkgs, loaded = ALL, outdated = NONE))

  ### all loaded, A outdated
  plan <- pkg_load_full_plan('pkgA', src_pkgs, loaded = ALL, outdated = 'pkgA')
  expected <- data.frame(package = c("pkgA", "pkgA"), action = c("unload", "load"))
  expected$params <- list(E, RA)
  expect_identical(plan, expected)
  
  ### all loaded, C outdated
  plan <- pkg_load_full_plan('pkgA', src_pkgs, loaded = ALL, outdated = 'pkgC')
  expected <- data.frame(
    package = c("pkgA", "pkgB", "pkgC", "pkgC", "pkgB", "pkgA"),
    action = c("unload", "unload", "unload", "load", "load", "load"))
  expected$params <- list(E, E, E, R, E, A)
  expect_identical(plan, expected)
  
  ### none loaded, none outdated
  plan <- pkg_load_full_plan('pkgA', src_pkgs, loaded = NONE, outdated = NONE)
  expected <- data.frame(package = c("pkgC", "pkgB", "pkgA"), action = "load")
  expected$params <- list(E, E, A)
  expect_identical(plan, expected)
  
  ### none loaded, all outdated
  plan <- pkg_load_full_plan('pkgA', src_pkgs, loaded = NONE, outdated = ALL)
  expected <- data.frame(package = c("pkgC", "pkgB", "pkgA"), action = "load")
  expected$params <- list(R, R, RA)
  expect_identical(plan, expected)

  ### none loaded, C outdated
  plan <- pkg_load_full_plan('pkgA', src_pkgs, loaded = NONE, outdated = 'pkgC')
  expected <- data.frame(package = c("pkgC", "pkgB", "pkgA"), action = "load")
  expected$params <- list(R, E, A)
  expect_identical(plan, expected)

  ### A loaded, B outdated (N.B: that should not be possible)
  plan <- pkg_load_full_plan('pkgA', src_pkgs, loaded = 'pkgA', outdated = 'pkgB')
  expected <- data.frame(package = c("pkgC", "pkgB"), action = "load")
  expected$params <- list(E, R)
  expect_identical(plan, expected)

  ################### load C #################################################
  ### all loaded and outdated
  plan <- pkg_load_full_plan('pkgC', src_pkgs, loaded = ALL, outdated = ALL)
  expected <- data.frame(
    package = c("pkgA", "pkgB", "pkgC", "pkgC", "pkgB", "pkgA"),
    action = c(rep("unload", 3), rep("load", 3)))
  expected$params <- list(E, E, E, RA, R, R)
  expect_identical(plan, expected)

  ### all loaded, only B outdated. B is a parent --> no need to reload C even if its parent is outdated
  expect_null(pkg_load_full_plan('pkgC', src_pkgs, loaded = ALL, outdated = 'pkgB'))

  ### all loaded, C outdated
  plan <- pkg_load_full_plan('pkgC', src_pkgs, loaded = ALL, outdated = 'pkgC')
  expected <- data.frame(
    package = c("pkgA", "pkgB", "pkgC", "pkgC", "pkgB", "pkgA"),
    action = c(rep("unload", 3), rep("load", 3)))
  expected$params <- list(E, E, E, RA, E, E)
  expect_identical(plan, expected)
})


.pkg_is_outdated <- 
test_that("pkg_is_outdated", {
  setup_temp_dir()
  pkg_create('.', 'AA')
  cleanup_dangling_srcpkgs()
  on.exit(cleanup_dangling_srcpkgs(), add = TRUE)
  PKG <- srcpkg(path = 'AA')

  ### no doc
  expect_true(pkg_is_outdated('AA', quiet = FALSE))
  
  ### loaded via srcpkg: --> up-to-date
  pkg_load_wrapper(PKG, quiet = TRUE, attach = FALSE)
  expect_false(pkg_is_outdated('AA', quiet = FALSE))

  # unloaded --> still up-to-date
  pkg_unload(PKG, quiet = TRUE)
  expect_false(pkg_is_outdated('AA', quiet = FALSE))

  # no roxygen
  pkg_delete_doc(PKG$path)
  expect_true(pkg_is_outdated('AA', quiet = FALSE))

  # roxygenise..
  pkg_roxygenise_wrapper(PKG$path, quiet = TRUE)
  expect_false(pkg_is_outdated('AA', quiet = TRUE))

  # modify a source
  writeLines('# hello', 'AA/R/toto.R')
  expect_true(pkg_is_outdated('AA', quiet = TRUE))

  # loaded via load_all --> needs reload
  pkg_unload(PKG, quiet = TRUE)
  pkg_delete_md5sum(PKG$path)
  devtools::load_all(PKG, quiet = TRUE, attach = FALSE)

  expect_true(pkg_is_outdated(PKG$path, quiet = FALSE))
})



# test_that("pkg_needs_reload", {
#   setup_temp_dir()
#   pkg_create('.', 'AA')
#   on.exit(pkg_unload('AA', quiet = TRUE), add = TRUE)
#   pkg_unload('AA', quiet = TRUE)
#   PKG <- srcpkg(path = 'AA')
  
#   ### not loaded yet
#   expect_true(pkg_needs_reload('AA'))

#   ### loaded via srcpkg: 
#   pkg_just_load_pkg(PKG, quiet = TRUE, attach = FALSE)
#   expect_false(pkg_needs_reload('AA'))

#   # unloaded
#   pkg_unload(PKG, quiet = TRUE)
#   expect_true(pkg_needs_reload('AA'))

#   # loaded, but no roxygen
#   pkg_just_load_pkg(PKG, quiet = TRUE, attach = FALSE)
#   pkg_delete_doc(PKG$path)
#   expect_true(pkg_needs_reload('AA', quiet = TRUE))

#   # unloaded
#   pkg_unload(PKG, quiet = TRUE)
#   expect_true(pkg_needs_reload('AA'))

#   # loaded via load_all --> needs reload
#   devtools::load_all(PKG, quiet = TRUE)
#   expect_true(pkg_needs_reload('AA'))
#   pkg_unload(PKG, quiet = TRUE)

#   # loaded via library
#   expect_false(pkg_needs_reload('base'))
# })

#' unloads a package, unloading its dependent packages if needed
#'
#' To be able to unload properly a package, all the packages that depend
#' even indirectly on it should be unloaded first.
#'
#' N.B: this function also works for non source packages.
#'  
#' @inheritParams params
#' @param loaded  the loaded packages, useful for testing.
#' @return a data frame of the unloaded package names, and whether they were
#'  attached, invisibly or NULL if the package is not loaded
#' @export
#' @examples
#' root <- tempfile()
#' pkg <- setup_and_get_dummy_srcpkg(root)
#' reset(root)
#' pkg_load(pkg)
#' 
#' pkg_unload(pkg)
pkg_unload <- function(pkg_or_name, 
  src_pkgs = get_srcpkgs(), dry_run = FALSE, loaded = loadedNamespaces(), quiet = FALSE)
{
  pkg_name <- as_pkg_name(pkg_or_name)
  # is it a srcpkg or not?
  plan <- NULL
  force(src_pkgs)
  if (pkg_name %in% names(src_pkgs)) {
    plan <- srcpkg_unload_plan(pkg_name, src_pkgs, loaded = loaded)
  } else {
    plan <- non_srcpkg_unload_plan(pkg_name, loaded = loaded)
  }

  if (!dry_run) execute_plan(plan, src_pkgs, quiet = quiet)

  invisible(plan)
}

srcpkg_unload_plan <- function(pkg_name, src_pkgs, 
  loaded = loadedNamespaces(), mat = graph_from_srcpkgs(src_pkgs)) 
{
  if (!pkg_name %in% loaded) return(NULL)
  force(mat)
  # mat <- sub_graph(mat, loaded)
  unload_plan(pkg_name, mat, loaded = loaded)
}

non_srcpkg_unload_plan <- function(pkg_name, loaded = loadedNamespaces()) 
{
  if (!pkg_name %in% loaded) return(NULL)

  deps <- find_loaded_packages_namespace_imports(loaded)
  deps <- lapply(deps, intersect, loaded)
  mat <- graph_from_deps(deps)

  unload_plan(pkg_name, mat, loaded = loaded)
}

# plan to unload pkg_names. will not unload packages not loaded
unload_plan <- function(pkg_names, mat, loaded = loadedNamespaces()) {
  if (!nrow(mat)) return(NULL)
  
  .deps <- function(x) c(graph_get_all_dependents(mat, x), x)
  deps_lst <- lapply(pkg_names, .deps)

  # N.B: can not be mepty
  deps <- unique(fast_unlist(deps_lst))

  ordering <- graph_topo_sort_nodes(mat, deps)
  plan <- data.frame(package = ordering, action = 'unload')
  plan <- plan[plan$package %in% loaded, , drop = FALSE]

  if (!nrow(plan)) return(NULL)

  plan
}


# N.B: namespace-imports != description-imports
# find the namespace-imports of all loaded packages (but "base" of course) as a dependency list
# does not report "base" as an import as well 
find_loaded_packages_namespace_imports <- function(loaded = loadedNamespaces()) {
  pkg_names <- setdiff(loaded, 'base')

  .fetch_imports <- function(x) {
    y <- pkg_list_ns_imports(x)
    reproducible_sort(setdiff(y, 'base'))
  }
  imports <- lapply(pkg_names, .fetch_imports)
  names(imports) <- pkg_names

  imports[reproducible_sort(names(imports))]
}

test_that("execute_plan", {
  src_pkgs <- examples_srcpkgs_complex_deps()

  mat <- graph_from_srcpkgs(src_pkgs)

  pkg_unload('EE', src_pkgs, quiet = TRUE) # in case

  ###
  plan <- pkg_load('AA', src_pkgs, roxygen = FALSE, quiet = TRUE)

  mat2 <- sub_graph(mat, loadedNamespaces())
  expect_equal(nrow(plan), nrow(mat2))

  uplan <- unload_plan('EE', mat2)
  expect_equal(nrow(uplan), 5)

  execute_plan(uplan, src_pkgs, quiet = TRUE)

  ### edge cases
  plan <- data.frame(package = 'AA', action = '')
  expect_error(execute_plan(plan, src_pkgs, quiet = TRUE), 'unknown action')
})

test_that("non_srcpkg_unload_plan", {
  plan <- non_srcpkg_unload_plan('rlang', loaded = c('devtools', 'pkgload', 'rlang', 'stats'))
  expect_identical(plan$package, c('devtools', 'pkgload', 'rlang'))
  expect_identical(unique(plan$action), 'unload')
})


test_that("srcpkg_unload_plan", {
  src_pkgs <- examples_srcpkgs_complex_deps()

  ### nothing to unload
  expect_null(srcpkg_unload_plan('EE', src_pkgs, loaded = NULL))
  
  ### simulate all loaded 
  # unload E --> need to unload all but Z
  plan <- srcpkg_unload_plan('EE', src_pkgs, loaded = names(src_pkgs))
  expect_identical(plan$package, c('AA', 'BB', 'CC', 'DD', 'EE'))
  expect_identical(unique(plan$action), 'unload')

  # unload Z -> only Z
  plan <- srcpkg_unload_plan('ZZ', src_pkgs, loaded = names(src_pkgs))
  expect_identical(plan$package, 'ZZ')
  # same for A
  plan <- srcpkg_unload_plan('AA', src_pkgs, loaded = names(src_pkgs))
  expect_identical(plan$package, 'AA')
  # for D
  plan <- srcpkg_unload_plan('DD', src_pkgs, loaded = names(src_pkgs))
  expect_identical(plan$package, c('AA', 'BB', 'CC', 'DD'))

  ### not all loaded
  expect_null(srcpkg_unload_plan('EE', src_pkgs, loaded = 'DD'))
  
  plan <- srcpkg_unload_plan('EE', src_pkgs, loaded = c('DD', 'EE'))
  expect_identical(plan$package, c('DD', 'EE'))

  # this should not be possible, since A depends on B. nonetheless we should then unload also A
  plan <- srcpkg_unload_plan('EE', src_pkgs, loaded = c('AA', 'DD', 'EE'))
  expect_identical(plan$package, c('AA', 'DD', 'EE'))
})


test_that("unload_plan", {
  ### A->B-C, A->C
  mat <- graph_from_strings('A->B->C', 'B->D')
  NONE <- character()
  ALL <- c('A', 'B', 'C', 'D')
  # N.B: A,B,D,C would have also worked

  expect_identical(unload_plan(c('C', 'D'), mat, loaded = ALL)$package, c('A', 'B', 'C', 'D'))

  expect_identical(unload_plan('C', mat, loaded = ALL)$package, c('A', 'B', 'C'))
  expect_identical(unload_plan('D', mat, loaded = ALL)$package, c('A', 'B', 'D'))
  expect_identical(unload_plan('B', mat, loaded = ALL)$package, c('A', 'B'))

  # using loaded=
  expect_null(unload_plan(c('C', 'D'), mat, loaded = NONE))
  expect_null(unload_plan('D', mat, loaded = NONE))
  expect_identical(unload_plan(c('C', 'D'), mat, loaded = c('C', 'D'))$package, c('C', 'D'))
  expect_identical(unload_plan(c('C', 'D'), mat, loaded = c('C', 'D'))$package, c('C', 'D'))

  ### empty matrix
  mat <- matrix(0L, 0, 0)
  expect_null(unload_plan('A', mat, loaded = ALL))

  ### trivial: A
  mat <- graph_from_strings('A->A')
  mat[1,1] <- 0L

  plan <- unload_plan('A', mat, loaded = ALL)
  expect_identical(plan, data.frame(package = 'A', action = 'unload'))

  ### simple : A-> B
  mat <- graph_from_strings('A->B')

  expect_identical(unload_plan('A', mat, loaded = ALL), data.frame(package = 'A', action = 'unload'))
  expect_identical(unload_plan('B', mat, loaded = ALL), data.frame(package = c('A','B'), action = 'unload'))

  ###  A-> B -> C
  mat <- graph_from_strings('A->B->C')

  expect_identical(unload_plan('A', mat, loaded = ALL), data.frame(package = 'A', action = 'unload'))
  expect_identical(unload_plan('B', mat, loaded = ALL), data.frame(package = c('A','B'), action = 'unload'))
  expect_identical(unload_plan('C', mat, loaded = ALL), 
    data.frame(package = c('A','B', 'C'), action = rep('unload', 3)))
  
  # 
  mat <- graph_from_strings('A->B->C', 'B->D->C')
  
  expect_identical(unload_plan('A', mat, loaded = ALL), data.frame(package = 'A', action = 'unload'))
  expect_identical(unload_plan('D', mat, loaded = ALL), 
    data.frame(package = c('A','B', 'D'), action = 'unload'))
  expect_identical(unload_plan('C', mat, loaded = ALL), 
    data.frame(package = c('A','B', 'D', 'C'), action = 'unload'))
})


test_that("pkg_unload", {
  all_src_pkgs <- examples_srcpkgs_complex_deps()

  ### srcpkg

  .unload <- function(x, dry_run = TRUE, quiet = TRUE, src_pkgs = all_src_pkgs, ...) {
    force(src_pkgs)
    my_src_pkgs <- src_pkgs
    pkg_unload(x, src_pkgs = my_src_pkgs, dry_run = dry_run, quiet = quiet, ...)
  }

  # not loaded
  expect_null(.unload('AA', loaded = NULL))

  # all loaded
  plan <- .unload('EE', loaded = names(all_src_pkgs))
  expect_identical(plan$package, c('AA', 'BB', 'CC', 'DD', 'EE'))
  
  plan <- .unload('BB', loaded = names(all_src_pkgs))
  expect_identical(plan$package, c('AA', 'BB'))

  # actually load/unload something
  for(pkg in all_src_pkgs) .unload(pkg, dry_run = FALSE)

  devtools::load_all('EE', export_all = FALSE, attach = TRUE, quiet = TRUE)
  devtools::load_all('DD', export_all = FALSE, attach = FALSE, quiet = TRUE)
 
  plan <- .unload('EE', dry_run = FALSE)

  expect_identical(plan$package, c('DD', 'EE'))
  expect_false(pkg_is_loaded('DD'))
  expect_false(pkg_is_loaded('EE'))
  ### non srcpkg

  # not loaded
  expect_null(.unload('rlang', loaded = NULL))

  # all loaded
  plan <- .unload('rlang', loaded = c('devtools', 'pkgload', 'rlang'))
  expect_identical(plan$package, c('devtools', 'pkgload', 'rlang'))
})


test_that("find_loaded_packages_namespace_imports", {
  setup_temp_dir()
  on.exit(cleanup_dangling_srcpkgs(), add = TRUE)
  cleanup_dangling_srcpkgs()

  pkg_create('.', 'AA', imports = c('tools', 'devtools'), namespace = TRUE)
  pkg_create('.', 'BB', imports = 'AA', namespace = TRUE)
  pkg_create('.', 'CC', imports = 'BB', depends = 'AA', namespace = TRUE)
  pkg_create('.', 'descimports', imports = c('testthat', 'utils'), namespace = FALSE)

  # our test packages are not loaded yet
  imports <- find_loaded_packages_namespace_imports()

  expect_true(is.list(imports))
  # not yet loaded
  expect_false(any(c('AA', 'BB', 'CC', 'descimports') %in% names(imports)))
  # testthat should be there
  expect_true('rlang' %in% imports[['testthat']])

  # now load the packages
  on.exit({
    unloadNamespace('CC')
    unloadNamespace('BB')
    unloadNamespace('AA')
    unloadNamespace('descimports')
  }, add = TRUE)

  # N.B: attaching AA because it is a "depends" of CC
  devtools::load_all('AA', export_all = FALSE, attach = TRUE, quiet = TRUE)
  devtools::load_all('BB', export_all = FALSE, attach = FALSE, quiet = TRUE)
  devtools::load_all('CC', export_all = FALSE, attach = FALSE, quiet = TRUE)
  devtools::load_all('descimports', export_all = FALSE, attach = FALSE, quiet = TRUE)

  imports <- find_loaded_packages_namespace_imports()

  expect_true(all(c('AA', 'BB', 'CC') %in% names(imports)))

  expect_null(imports[['descimports']])
  expect_identical(imports[['AA']],  c('devtools', 'tools')) # N.B: sorted
  expect_identical(imports[['BB']],  'AA')
  expect_identical(imports[['CC']],  'BB') # N.B; 'AA' is not listed since it is a "Depends"
})

test_that("load/unload POC - NAMESPACE imports", {
  setup_temp_dir()

  pkg_create('.', 'AA')
  pkg_create('.', 'BB', imports = 'AA', namespace = TRUE)

  on.exit({
    unloadNamespace('BB')
    unloadNamespace('AA')
  }, add = TRUE)

  # BB can not be loaded because AA is a NS-import and is not loaded
  expect_error(devtools::load_all('BB', export_all = FALSE, quiet = TRUE), 'required')
  
  # load AA...
  devtools::load_all('AA', export_all = FALSE,  quiet = TRUE)

  # and now BB CAN be loaded
  devtools::load_all('BB', export_all = FALSE, quiet = TRUE)

  ### unload
  # for the same reason, AA can NOT be unloaded because it is NS-imported by BB
  expect_error(unloadNamespace('AA'), 'cannot be unloaded')

  # but if we first unload BB...
  unloadNamespace('BB')

  # then we CAN unload AA
  expect_error(unloadNamespace('AA'), NA)
})


# cf "load/unload POC - NAMESPACE imports" above
test_that("load/unload POC - DESCRIPTION imports", {
  setup_temp_dir()

  pkg_create('.', 'AA')
  pkg_create('.', 'BB', imports = 'AA', namespace = FALSE)

  on.exit({
    unloadNamespace('BB')
    unloadNamespace('AA')
  }, add = TRUE)

  ### load: same behaviour as the NS-imports
  #  BB can not be loaded because AA is an import and is not loaded nor installed
  expect_error(devtools::load_all('BB', export_all = FALSE, quiet = TRUE), 'required')

  # load AA...
  devtools::load_all('AA', export_all = FALSE,  quiet = TRUE)

  # and now BB CAN be loaded
  devtools::load_all('BB', export_all = FALSE, quiet = TRUE)

  ### unload: different behaviour !
  # AA CAN be unloaded because it is only DESC-imported by BB
  expect_error(unloadNamespace('AA'), NA)

  # then we CAN unload BB
  expect_error(unloadNamespace('BB'), NA)
})


test_that("load/unload POC - DESCRIPTION depends", {
  setup_temp_dir()

  pkg_create('.', 'AA')
  pkg_create('.', 'BB', depends = 'AA')

  on.exit({
    unloadNamespace('BB')
    unloadNamespace('AA')
  }, add = TRUE)

  ### load: same behaviour as the NS-imports
  #  BB can not be loaded because AA is an import and is not loaded nor installed
  expect_error(devtools::load_all('BB', export_all = FALSE, quiet = TRUE), 'required')

  # load AA...
  devtools::load_all('AA', export_all = FALSE, attach = FALSE,  quiet = TRUE)
  expect_false(pkg_is_attached('AA'))

  # BB can be loaded BUT with a WARNING since AA is not attached
  expect_warning(devtools::load_all('BB', export_all = FALSE, quiet = TRUE), 'AA')

  # attach AA
  unloadNamespace('BB')
  attachNamespace('AA')
  expect_true(pkg_is_attached('AA'))
  
  # NOW BB CAN be loaded because AA is attached
  expect_error(devtools::load_all('BB', export_all = FALSE, quiet = TRUE), NA)

  ### unload: 
  expect_error(pkg_detach('AA'), 'required')

  expect_error(unloadNamespace('BB'), NA)

  # now we can detach AA
  expect_error(pkg_detach('AA'), NA)

  # or unload it
  expect_error(unloadNamespace('AA'), NA)
})

PREVIOUS_SETTINGS <- settings()
withr::defer(restore_init(PREVIOUS_SETTINGS), teardown_env())

if (is_loaders_hack_enabled()) {
  unhack_r_loaders()
  withr::defer(hack_r_loaders(), teardown_env())
}

withr::local_envvar(
  R_USER_CACHE_DIR = tempfile(),
  .local_envir = teardown_env()
)

#' shared params
#'
#' @param dry_run       whether not to actually execute any action having side-effects
#' @param lib         	directory where to install and find installed pkgs
#' @param pkg						a package as a "srcpkg" object
#' @param pkgs					packages as a "srcpkgs" object
#' @param pkg_name			the package name, as a character
#' @param pkg_or_name   a package name or object ("package" or "srcpkg")
#' @param pkg_path			the package path, as a character
#' @param pkgid         a package name, path or package object
#' @param pkgids        a list of package ids (names, paths or object), or a
#' 											srcpkgs object. Also accept a singleton package object
#' @param md5         	the MD5 hash of the source package
#' @param progress 			whether to display a progress bar
#' @param roxygen       whether to roxygenize
#' @param src_pkgs      a collection of source packages as a `srckgs` object.
#' @param srcpkgs_paths paths to the source packages folders
#' @param root		      directory from where to search for source packages
#' @param quiet         whether to be quiet/silent
#' @param test_filter		a pattern to select the testthat tests to run.
#'                      Test files are names test-xxxxx.R where xxxxx is the
#'                      test name. Only test files whose name match the pattern
#'                      will be run.
#' @param test_parallel	whether to run the package tests in parallel
#'
#' @name params
#' @keywords internal
#'
NULL

#' @export
summary.pkg_check <- function(object, ...) {
  df <- data.frame(package = object$package)

  cols <- c("errors", "warnings", "notes")
  df[cols] <- lengths(object[cols])

  df$time <- object$duration

  df
}

#' @export
as.logical.pkg_check <- function(x, ...) {
   df <- summary(x)
   df$error == 0
}

#' @export
as.data.frame.pkg_check <- function(x, ...) {
  summary.pkg_check(x)
}

#' tests a package - runs R CMD check 
#'
#' This function will check a source package.
#'
#' @inheritParams params
#' @param error_on              passed to `devtools::check()`
#' @param check_system_clock    if FALSE, disable the `_R_CHECK_SYSTEM_CLOCK_` check. This check
#'                              sometimes fail because of firewalls...
#' @param ...                   passed to `devtools::check()`
#' @return the results as a `pkg_test` object, or NULL if no tests found
#' @export
pkg_check <- function(pkgid, src_pkgs = get_srcpkgs(), lib = ".check",  roxygen = TRUE,
  quiet = FALSE, error_on = "error",  check_system_clock = FALSE, ...) 
{  
  force(src_pkgs)
  pkg <- as_srcpkg(pkgid, src_pkgs)

  if (roxygen) pkg_roxygenise(pkg$path, quiet = TRUE)

  res <- withr::with_envvar(c(`_R_CHECK_SYSTEM_CLOCK_` = as.integer(check_system_clock)), 
    devtools::check(pkg, error_on = error_on, check_dir = lib, quiet = quiet, ...))
  
  class(res) <- c('pkg_check', class(res)) 
  invisible(res)
}

print_text_table <- function(df, title = NULL, header_style = "bold", border_style = "double-single", ...) {
  if (length(title)) cli::cli_h1(title)
  ct <- cli_table(df, header_style = header_style, border_style = border_style, ...)
  cat(ct, sep = "\n")
}

---
title: "Testing and checking packages with srcpkgs"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Testing and checking packages with srcpkgs}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---


```{r setup, include = FALSE}
## N.B: we need to setup and make available the demo collection of packages "srcpkgs_lotr_demo"
# setup a temp dir
root_dir <- tempfile()
dir.create(root_dir)
file.copy('srcpkgs_lotr_demo', root_dir, recursive = TRUE)
workdir <- file.path(root_dir, 'srcpkgs_lotr_demo')
knitr::opts_knit$set(root.dir = workdir) # set directory for other chunks
withr::local_envvar(R_USER_CACHE_DIR = tempfile())
```

I will demonstrate srcpkgs using a dummy collection of source packages: https://github.com/kforner/srcpkgs_lotr_demo
It consists currently in 11 related packages, with internal dependencies.
The dependencies are implemented by a mix of Imports, Imports with namespace imports and Depends.


## listing the LOTR collection of packages with srcpkgs

```{r srcpkgs_setup, results = "markup"}

library(srcpkgs)
reset(root_dir)
print(names(get_srcpkgs()))
# cat(clitable::cli_table(as.data.frame(get_srcpkgs())), sep = "\n")
```


## testing the collection 

### no tests yet

The LOTR collection does not come with any test. 
Let's see what happens then...

```{r no_tests, results = "markup"}
print(pkgs_test(reporter = "silent"))
```

So no tests (`nb == 0`) but the testing was successful since no test failed...


### adding dummy tests to the packages

Let's add programmatically some dummy tests to our packages.

```{r dummy_tests}
add_dummy_test_to_srcpkg <- function(srcpkg, with_failures = TRUE, with_errors = TRUE, with_warnings = TRUE) {
  withr::local_dir(srcpkg$path)
  dir.create("tests/testthat", recursive = TRUE, showWarnings = FALSE)

  .write_test <- function(name, code, test = name) {
    writeLines(sprintf(r"-----{
    test_that("%s", {
      %s
    })
    }-----", name, code), sprintf("tests/testthat/test-%s.R", test))
  }

  .write_test("success", "expect_true(TRUE)")
  if (with_failures) {
    .write_test("failure", "expect_true(FALSE)")
    .write_test("mixed", "expect_true(FALSE);expect_true(TRUE)")
  }
  .write_test("skip", 'skip("skipping");expect_true(FALSE)')
  if (with_errors) .write_test("errors", 'expect_true(TRUE);stop("Arghh");expect_true(TRUE)')
  if (with_warnings)  .write_test("warning", 'expect_true(FALSE);warning("watch out");expect_true(FALSE)')
  if (with_failures && with_errors)
    writeLines(r"-----{
    test_that("misc1", {
      expect_true(FALSE)
      expect_true(TRUE)
    })
    test_that("misc2", {
      expect_true(FALSE)
      skip("skipping")
    })
    test_that("misc3", {
      expect_true(TRUE)
      expect_true(TRUE)
    })
    test_that("misc4", {
      expect_true(TRUE)
      warning("fais gaffe")
      stop("aie")
      expect_true(TRUE)
    })
    }-----", "tests/testthat/test-misc.R")

  writeLines(sprintf(r"-----{
    library(testthat)
    library(%s)

    test_check("%s")
  }-----", srcpkg$package, srcpkg$package), "tests/testthat.R")
}
i <- 0
for (pkg in get_srcpkgs()) {
  add_dummy_test_to_srcpkg(pkg, i %% 3 == 1, i %% 7 == 1, i %% 5 == 1)
  i <- i + 1
}
```


### testing

Now let's test again.

```{r with_tests, results = "markup"}
# N.B: we use the silent testthat reporter because we only want to get the results as tables
test_results <- pkgs_test(reporter = "silent")
print(test_results)
```

Note that in markdown we can not have the ANSI colors and formatting. Here's a screenshot
![screenshot](pkgs_test.png)


### using the test results

The test results are stored as a `pkgs_test` object, which is a list named after the packages, of `pkg_test` objects
which are a subclass of `testthat_results`.
You can manipulate them with S3 methods:

  - `as.data.frame()` - converts the results to a data frame with one row per package
  - `summary()` - converts the results to a one-row data frame that summarizes the results for the collection of packages
  - `as.logical()` - tells if the overall testing of the collection was successful
  - `print()` - prints the results as pretty tables

These S3 methods are also implemented for `pkg_test` objects.

```{r pkgs_test_methods, results = "markup"}
print(as.data.frame(test_results))
print(summary(test_results))
print(as.logical(test_results))

print(test_results$bilbo)
print(as.data.frame(test_results$lotr))
print(summary(test_results$lotr))
print(as.logical(test_results$aragorn))
```


## checking the collection 

Checking is very similar to testing except that it takes much longer!
So we'll use a small subset of the collection.

```{r subset}
small_collection <- get_srcpkgs(filter = "elves|galadriel|legolas")
```

### fixing the collection: declare testthat as dependency
```{r fixing}
.fix_description <- function(path, lst) {
  df <- read.dcf(path, all = TRUE)
  df2 <- utils::modifyList(df, lst)
  write.dcf(df2, path)
}
for (pkg in small_collection) {
  .fix_description(file.path(pkg$path, "DESCRIPTION"), list(Suggests = "testthat"))
}
```


### checking the packages

```{r checking, results = "markup"}
check_results <- pkgs_check(small_collection, quiet = TRUE)
print(check_results)
```

### using the check results

The check results are stored as a `pkgs_check` object, which is a list named after the packages, of `pkg_check` objects
which are a subclass of `rcmdcheck`.

As with `pkgs_test` results, you can manipulate them with S3 methods:

  - `as.data.frame()` - converts the results to a data frame with one row per package
  - `summary()` - converts the results to a one-row data frame that summarizes the results for the collection of packages
  - `as.logical()` - tells if the overall testing of the collection was successful
  - `print()` - prints the results as pretty tables

```{r pkgs_check_methods, results = "markup"}
print(as.data.frame(check_results))
print(summary(check_results))
print(as.logical(check_results))

print(check_results$bilbo)
print(as.data.frame(check_results$lotr))
print(summary(check_results$lotr))
print(as.logical(check_results$aragorn))
```

---
title: "why would you need srcpkgs?"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{why would you need srcpkgs?}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, echo = FALSE, results='hide'}
# suppressPackageStartupMessages(library(srcpkgs))

# setup a temp dir
root_dir <- tempfile()
dir.create(root_dir)
file.copy('srcpkgs_lotr_demo', root_dir, recursive = TRUE)
workdir <- file.path(root_dir, 'srcpkgs_lotr_demo')
knitr::opts_knit$set(root.dir = workdir) # set directory for other chunks
```

I will demonstrate srcpkgs using a dummy collection of source packages: https://github.com/kforner/srcpkgs_lotr_demo

## overview of the srcpkgs_lotr_demo collection

It consists currently in 11 related packages, with a tree structure:

  - lotr - the highest level package
    * elves
      * legolas
      * galadriel
    * hobbits
      * bilbo
      * frodo
    * gimli
    * aragorn
    * gandalf
    * elrond

N.B: `lotr` depends on all other packages, except for `elrond` (not yet).

The dependencies are implemented by a mix of Imports, Imports with namespace imports and Depends.

## using devtools


```{r devtools_setup, results = 'hide'}
suppressPackageStartupMessages(library(devtools))
pkgs <- c('bilbo', 'frodo', 'hobbits', 'legolas', 'galadriel', 'elves', 'gimli', 'aragorn', 'gandalf', 'lotr')
```

### loading

`devtools` is designed to manage a single source package. Let's use it to load our `lotr` source package:

```{r devtools_load, error = TRUE}
load_all('lotr')
```

--> devtools can NOT load `lotr` since it can not possibly find the `hobbits` package, which is a dependency.

Let's help him:
```{r devtools_load2, error = TRUE}
load_all('hobbits')
```

--> same problem

Here is how we must load the packages, following the dependencies order.
Note that we also need to roxygenize them (using `document()`)
```{r devtools_load3, message = FALSE}
document('frodo')
load_all('frodo')
document('bilbo')
load_all('frodo')
document('hobbits')
load_all('hobbits')

document('legolas')
load_all('legolas')
document('galadriel')
load_all('galadriel')
document('elves')
load_all('elves')

document('gimli')
load_all('gimli')
document('aragorn')
load_all('aragorn')
document('gandalf')
load_all('gandalf')
```


and finally we can load it
```{r devtools_load_final}
document('lotr')
load_all('lotr')

# use it
str(lotr())
```


### editing and reloading

Let's modify one of the direct dependency of `lotr`, e.g. the `hobbits` package. 
Currently:

```{r devtools_edit1}
names(lotr()$hobbits)
```


```{r devtools_edit2}
lines <- readLines('hobbits/R/main.R')
cat(lines, sep = '\n')
```

Edit `hobbits/R/main.R` and comment out bilbo, which comes from the `bilbo` package in **Depends**.

```{r devtools_edit3}
edited_lines <- grep('bilbo', lines, invert = TRUE, value = TRUE)
writeLines(edited_lines, 'hobbits/R/main.R')
```

Let's try to apply our changes:

```{r devtools_edit4}
load_all('lotr')
names(lotr()$hobbits)
```
--> `load_all()` can not properly reload our package, since it does not know that 
a dependency has been modified.

To apply the changes:
```{r devtools_edit5}
load_all('hobbits')
names(lotr()$hobbits)
```
--> now it works. Note that this is because `load_all()` is now able to force the unload of a package
even though it is needed by another package:


```{r devtools_edit6, error = TRUE}
unloadNamespace('hobbits')
```

But:
```{r devtools_edit7}
devtools::unload('hobbits')
```

Note that now `lotr` is **broken**:
```{r devtools_edit8, error = TRUE}
lotr()
```

Let's fix it
```{r devtools_edit9, error = TRUE}
load_all('hobbits')
names(lotr())
```

## using srcpkgs

```{r srcpkgs_setup}
library(srcpkgs)
old <- options(width = 200)
print(get_srcpkgs())
```

### unloading

the `srcpkgs::pkg_unload()` takes into account the dependencies between the source packages.

```{r srcpkgs_unload1}
plan <- pkg_unload('bilbo')
print(plan)
```

unload all our packages to start from a clean state. We will also 
```{r srcpkgs_unload2}
for (pkg in get_srcpkgs()) pkg_unload(pkg)
```


### loading

`srcpkgs::pkg_load()` takes care of everything: it roxygenizes the packages if needed, and load them 
in the appropriate order:

```{r srcpkgs_load1}
plan <- pkg_load('lotr')
print(plan)
print(names(lotr()))
```

### editing and reloading

Let's edit the `frodo` package, and change the weapon from `sting` to `sword`:

```{r srcpkgs_edit1}
lotr()$hobbits$frodo

lines <- readLines('frodo/R/main.R')
cat(lines, sep = '\n')

edited_lines <- sub('sting', 'sword', lines)
writeLines(edited_lines, 'frodo/R/main.R')
```

Now let's ask `srcpkgs` to make sure the `lotr` package is up-to-date: 

```{r srcpkgs_edit2}
plan <- pkg_load('lotr')
print(plan)
```

--> It figured out that `frodo` was modified, and needed to be reloaded, and for that 
all its dependents needed to be also properly unloaded and re-loaded.

```{r srcpkgs_edit3}
lotr()$hobbits$frodo$weapons
```

```{r cleanup, include = FALSE}
options(old)
```

---
title: "Getting started"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# use the LOTR dummy collection of source packages

To illustrate the use of the `srcpkgs`, we need some source packages.
An easy way to experiment with `srcpkgs` is with the https://github.com/kforner/srcpkgs_lotr_demo repository:

  - install the `srcpkgs` package (cf https://kforner.github.io/srcpkgs/#installation)
  - check out the https://github.com/kforner/srcpkgs_lotr_demo repository:

```
git clone https://github.com/kforner/srcpkgs_lotr_demo.git
cd srcpkgs_lotr_demo
```

then start a R session in that folder (type `R`):

```
# load srcpkgs
>library(srcpkgs)
```


# get_srcpkgs()

The `get_srcpkgs()` fetches the list of source packages that are currently managed by `srcpkgs`.
You should see all packages from the `srcpkgs_lotr_demo` repository.

```
# let's see what source packages are discovered:
> get_srcpkgs()
            package version                                                path                     imports depends suggests
aragorn     aragorn     0.1   /workspaces/srcpkgs_pkg/srcpkgs_lotr_demo/aragorn                                             
bilbo         bilbo     0.1     /workspaces/srcpkgs_pkg/srcpkgs_lotr_demo/bilbo                                             
elrond       elrond     0.1    /workspaces/srcpkgs_pkg/srcpkgs_lotr_demo/elrond                                             
elves         elves     0.1     /workspaces/srcpkgs_pkg/srcpkgs_lotr_demo/elves         galadriel,\nlegolas                 
frodo         frodo     0.1     /workspaces/srcpkgs_pkg/srcpkgs_lotr_demo/frodo                                             
galadriel galadriel     0.1 /workspaces/srcpkgs_pkg/srcpkgs_lotr_demo/galadriel                                             
gandalf     gandalf     0.1   /workspaces/srcpkgs_pkg/srcpkgs_lotr_demo/gandalf                                             
gimli         gimli     0.1     /workspaces/srcpkgs_pkg/srcpkgs_lotr_demo/gimli                                             
hobbits     hobbits     0.1   /workspaces/srcpkgs_pkg/srcpkgs_lotr_demo/hobbits                       frodo   bilbo         
legolas     legolas     0.1   /workspaces/srcpkgs_pkg/srcpkgs_lotr_demo/legolas                                             
lotr           lotr     0.1      /workspaces/srcpkgs_pkg/srcpkgs_lotr_demo/lotr elves,gimli,aragorn,gandalf hobbits  
```

This function returns a `srcpkgs` object, that is a named list of `srcpkg` objects.
You can `print` that object, or transform it in a data frame using `as.data.frame()`.

```
pkgs <- get_srcpkgs()
print(pkgs)

df <- as.data.frame(pkgs)
```

# loading and reloading packages

## hacked R loaders

Since by default, the R loaders should be automatically *hacked* by `srcpkgs` (cf `hack_r_loaders()`),
 you may simply load and/or ttach these packages using `library()`, `loadNamespace()`, `getNamespace`, `::` and so on:

```
library(elves)
loadNamespace('bilbo')
frodo::frodo()
```

The rationale for hacking the R loaders is not just for convenience. It's also for writing R code that
will still work once the source packages are installed.

## using pkg_load()

The `pkg_load()` is the workhorse function to load packages. A major difference with 
`devtools::load_all()` defaults is that it does not source the *test helpers*, nor does it 
export all objects (cf `export_all`). The rationale is to stick as much as possible to the standard
R loaders behaviour. 

You may use a package name, path, package or `srcpkgs` object to designate the package to load:

```
pkg_load('lotr')
pkg_load('/workspaces/srcpkgs_pkg/srcpkgs_lotr_demo/lotr')
pkg_load(pkgs$lotr)
```

### updating

When you execute `pkg_load()` on a package already loaded, it will check for changes not only in that 
package but also in all its dependencies (among the source packages), and will do what's needed to roxygenize, unload, reload
until your package is up-to-date inside your R session.



# reset and settings

At `srcpkgs` startup time, i.e. when the `library(srcpkgs)` statement is executed, the R loaders are
automatically hacked (cf `hack_r_loaders()`) UNLESS the hack is inhibited using 
the option `srcpkgs.inhibit_r_loaders_hack` or the environment variable `SRCPKGS.INHIBIT_R_LOADERS_HACK`.

Then the fist time `get_srcpkgs()` will be executed (for example indirectly by calling `pkg_load()`), 
`srcpkgs` will figure out the **project root**, and from this root will search for source packages, 
using a heuristic. Since that can take a bit of I/O time, the project root and the paths of found 
source packages are cached. 

You may check those using the `settings()` function, and can change them using the `reset()` function.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkg_unload.R
\name{pkg_unload}
\alias{pkg_unload}
\title{unloads a package, unloading its dependent packages if needed}
\usage{
pkg_unload(
  pkg_or_name,
  src_pkgs = get_srcpkgs(),
  dry_run = FALSE,
  loaded = loadedNamespaces(),
  quiet = FALSE
)
}
\arguments{
\item{pkg_or_name}{a package name or object ("package" or "srcpkg")}

\item{src_pkgs}{a collection of source packages as a \code{srckgs} object.}

\item{dry_run}{whether not to actually execute any action having side-effects}

\item{loaded}{the loaded packages, useful for testing.}

\item{quiet}{whether to be quiet/silent}
}
\value{
a data frame of the unloaded package names, and whether they were
attached, invisibly or NULL if the package is not loaded
}
\description{
To be able to unload properly a package, all the packages that depend
even indirectly on it should be unloaded first.
}
\details{
N.B: this function also works for non source packages.
}
\examples{
root <- tempfile()
pkg <- setup_and_get_dummy_srcpkg(root)
reset(root)
pkg_load(pkg)

pkg_unload(pkg)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkg_load.R
\name{pkg_load}
\alias{pkg_load}
\title{loads or reloads if needed a source package, taking care of its dependencies}
\usage{
pkg_load(
  pkgid,
  src_pkgs = get_srcpkgs(),
  attach = TRUE,
  suggests = FALSE,
  roxygen = TRUE,
  helpers = FALSE,
  export_all = FALSE,
  quiet = FALSE,
  dry_run = FALSE,
  ...
)
}
\arguments{
\item{pkgid}{a package name, path or package object}

\item{src_pkgs}{a collection of source packages as a \code{srckgs} object.}

\item{attach}{Whether to attach a package environment to the search
path. If \code{FALSE} \code{load_all()} behaves like \code{loadNamespace()}. If
\code{TRUE} (the default), it behaves like \code{library()}. If \code{FALSE},
the \code{export_all}, \code{export_imports}, and \code{helpers} arguments have
no effect.}

\item{suggests}{whether to load suggested packages. if TRUE, the suggested are processed
like imports}

\item{roxygen}{whether to automatically roxygenise packages (if needed)}

\item{helpers}{if \code{TRUE} loads \pkg{testthat} test helpers.}

\item{export_all}{If \code{TRUE} (the default), export all objects.
If \code{FALSE}, export only the objects that are listed as exports
in the NAMESPACE file.}

\item{quiet}{whether to be quiet/silent}

\item{dry_run}{whether not to actually execute any action having side-effects}

\item{...}{
  Arguments passed on to \code{\link[devtools:load_all]{devtools::load_all}}
  \describe{
    \item{\code{path}}{Path to a package, or within a package.}
    \item{\code{reset}}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} This is no longer supported
because preserving the namespace requires unlocking its environment, which
is no longer possible in recent versions of R.}
    \item{\code{recompile}}{DEPRECATED. force a recompile of DLL from source code, if
present. This is equivalent to running \code{\link[pkgbuild:clean_dll]{pkgbuild::clean_dll()}} before
\code{load_all()}}
  }}
}
\value{
the load plan as a data frame, or NULL if there is nothing to do.
}
\description{
N.B: the defaults are different from \code{\link[devtools:load_all]{devtools::load_all()}}: the helpers are not loaded, only
the functions tagged as \emph{exported} are actually exported. The intended goal is to make it as similar
to the behaviour of the R loaders.
}
\details{
This the workhorse function of the package, called by \code{\link[=library]{library()}} and \code{\link[=loadNamespace]{loadNamespace()}}
when hacked (cf \code{\link[=hack_r_loaders]{hack_r_loaders()}}.

This function will check that all dependent packages are up-to-date, and
document and reload them as needed.

To be able to properly load a package, its dependent source packages
must be loaded in proper order.
i.e. if A-->B-->C, the load order must be C, B, A
}
\examples{
root <- tempfile()
pkg <- setup_and_get_dummy_srcpkg(root)
reset(root)
# load and attach a package
pkg_load(pkg)

# just load, do not attach it (~ loadNamespace())
pkg_unload(pkg)
pkg_load(pkg, attach = FALSE)

# do some changes, to a source package or any of its depencies or dependents
pkg_unload(pkg)
plan <- pkg_load(pkg, dry_run = TRUE)
# then you can inspect the plan actions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgs_check.R
\name{pkgs_check}
\alias{pkgs_check}
\title{checks a list of source packages}
\usage{
pkgs_check(
  pkgids = names(filter_srcpkgs(src_pkgs, filter)),
  src_pkgs = get_srcpkgs(),
  filter = NULL,
  lib = ".check",
  quiet = FALSE,
  fail_on_error = FALSE,
  ...
)
}
\arguments{
\item{pkgids}{a list of package ids (names, paths or object), or a
srcpkgs object. Also accept a singleton package object}

\item{src_pkgs}{a collection of source packages as a \code{srckgs} object.}

\item{filter}{filter out the packages to check using this pattern}

\item{lib}{directory where to install and find installed pkgs}

\item{quiet}{whether to be quiet/silent}

\item{fail_on_error}{whether to die if there is at least an error or warning in the checks}

\item{...}{passed to \code{pkg_check}}
}
\value{
the results as a \code{pkgs_test} object
}
\description{
checks a list of source packages
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reset.R
\name{reset}
\alias{reset}
\title{resets the \code{srcpkgs} settings}
\usage{
reset(root = find_project_root(), srcpkgs_paths = find_srcpkgs_paths(root))
}
\arguments{
\item{root}{directory from where to search for source packages}

\item{srcpkgs_paths}{paths to the source packages folders}
}
\value{
the settings (cf \code{\link[=settings]{settings()}}) invisibly
}
\description{
With this function, you can reset or set precisely the settings.
}
\examples{
# reset to appropriate defaults based on your current directory
old <- reset()

# explictly set the project root
reset(root = tempdir())

# explictly set the source package paths (very unlikely)
reset(srcpkgs_paths = c('pkgs/mypkg1', 'pkgs/mypkg2'))

# restore previous settings
reset(root = old$root, srcpkgs_paths = old$srcpkgs_paths)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{legolas}
\alias{legolas}
\title{provides legolas}
\usage{
legolas()
}
\value{
as a list
}
\description{
provides legolas
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{hobbits}
\alias{hobbits}
\title{list the available hobbits}
\usage{
hobbits()
}
\value{
hobbit names as character
}
\description{
list the available hobbits
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reset.R
\name{settings}
\alias{settings}
\title{informs about the settings currently used by \code{srcpkgs}}
\usage{
settings()
}
\value{
a named list of:
\itemize{
\item initialized: whether the settings are initialized (as triggered by \code{\link[=get_srcpkgs]{get_srcpkgs()}})
\item root: the project root
\item srcpkgs_paths: the paths of the source packages to manage
\item hack_r_loaders_installed: whether the R loaders are hacked
\item hack_r_loaders_enabled: whether the R loaded hack is in action (internal use)
}
}
\description{
informs about the settings currently used by \code{srcpkgs}
}
\examples{
print(settings())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{lotr}
\alias{lotr}
\title{list all characters}
\usage{
lotr()
}
\value{
as a list
}
\description{
list all characters
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{aragorn}
\alias{aragorn}
\title{provides aragorn}
\usage{
aragorn()
}
\value{
as a list
}
\description{
provides aragorn
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkg_check.R
\name{pkg_check}
\alias{pkg_check}
\title{tests a package - runs R CMD check}
\usage{
pkg_check(
  pkgid,
  src_pkgs = get_srcpkgs(),
  lib = ".check",
  roxygen = TRUE,
  quiet = FALSE,
  error_on = "error",
  check_system_clock = FALSE,
  ...
)
}
\arguments{
\item{pkgid}{a package name, path or package object}

\item{src_pkgs}{a collection of source packages as a \code{srckgs} object.}

\item{lib}{directory where to install and find installed pkgs}

\item{roxygen}{whether to roxygenize}

\item{quiet}{whether to be quiet/silent}

\item{error_on}{passed to \code{devtools::check()}}

\item{check_system_clock}{if FALSE, disable the \verb{_R_CHECK_SYSTEM_CLOCK_} check. This check
sometimes fail because of firewalls...}

\item{...}{passed to \code{devtools::check()}}
}
\value{
the results as a \code{pkg_test} object, or NULL if no tests found
}
\description{
This function will check a source package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkg_utils.R
\name{pkg_create}
\alias{pkg_create}
\title{creates and populates a R package-like folder programmatically, useful for writing tests}
\usage{
pkg_create(
  dir,
  name,
  functions = list(dummy = function() "DUMMY"),
  imports = NULL,
  depends = NULL,
  suggests = NULL,
  namespace = FALSE,
  roxygen_imports = FALSE,
  ignore_hidden_files = TRUE
)
}
\arguments{
\item{dir}{the directory in which to create the package, as a string}

\item{name}{the package name, as a string}

\item{functions}{a named list of functions to add to the package}

\item{imports}{the "imports" dependencies}

\item{depends}{the "depends" dependencies}

\item{suggests}{the "suggests" dependencies}

\item{namespace}{whether to write the namespace file (currently only applicable to the imports.
N.B: if the namespace file is generated, roxygen will refuse to update it}

\item{roxygen_imports}{whether to write the roxygen statements to defined the imports}

\item{ignore_hidden_files}{whether to create a \code{.Rbuildignore} file to ignore hidden files.}
}
\value{
the srcpkg instance, invisibly
}
\description{
basically a wrapper around \code{utils::package.skeleton()}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgs_deps.R
\name{pkgs_deps}
\alias{pkgs_deps}
\title{computes the dependencies of some (source) packages}
\usage{
pkgs_deps(
  pkgids,
  src_pkgs = get_srcpkgs(),
  source = TRUE,
  installed = TRUE,
  imports = TRUE,
  depends = TRUE,
  suggests = TRUE,
  reverse = FALSE
)
}
\arguments{
\item{pkgids}{a list of package ids (names, paths or object), or a
srcpkgs object. Also accept a singleton package object}

\item{src_pkgs}{a collection of source packages as a \code{srckgs} object.}

\item{source}{whether to report source packages}

\item{installed}{whether to report installed (non-source) packages}

\item{imports}{whether to only consider \code{imports} dependencies}

\item{depends}{whether to only consider \code{depends} dependencies}

\item{suggests}{whether to only consider \code{suggests} dependencies}

\item{reverse}{whether to compute reverse dependencies instead}
}
\value{
the dependencies, as a character vector, topologically sorted
}
\description{
computes the dependencies of some (source) packages
}
\examples{
pkg <- setup_and_get_dummy_srcpkg()
deps_src <- pkgs_deps(pkg, installed = FALSE)
deps_inst <- pkgs_deps(pkg, source = FALSE)
print(get_srcpkgs())
deps_rev <- pkgs_deps(pkg, reverse = TRUE, suggests = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{elrond}
\alias{elrond}
\title{provides elrond}
\usage{
elrond()
}
\value{
as a list
}
\description{
provides elrond
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{galadriel}
\alias{galadriel}
\title{provides galadriel}
\usage{
galadriel()
}
\value{
as a list
}
\description{
provides galadriel
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkg_test.R
\name{pkg_test}
\alias{pkg_test}
\title{tests a package - runs its unit tests}
\usage{
pkg_test(
  pkgid,
  filter = NULL,
  src_pkgs = get_srcpkgs(),
  export_all = TRUE,
  quiet = TRUE,
  ...
)
}
\arguments{
\item{pkgid}{a package name, path or package object}

\item{filter}{filter in the tests to run. cf \code{testthat::test_dir()}}

\item{src_pkgs}{a collection of source packages as a \code{srckgs} object.}

\item{export_all}{passed to \code{\link[=pkg_load]{pkg_load()}}. Enables the test functions to easily access to non-exported
functions. Caveat: If the pkg is already loaded and up-to-date with export_all=FALSE, it will not work.}

\item{quiet}{whether to be quiet/silent}

\item{...}{passed to \code{testthat::test_dir()}}
}
\value{
the results as a \code{pkg_test} object, which is an empty listL if no tests were found
}
\description{
This function will test a source package using \code{testthat},
making sure the package and its source package dependencies are up-to-date and loaded
}
\examples{
 root <- tempfile()
 pkg <- setup_and_get_dummy_srcpkg(root)
 reset(root)
 res <- pkg_test(pkg)
 print(res)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{gimli}
\alias{gimli}
\title{provides gimli}
\usage{
gimli()
}
\value{
as a list
}
\description{
provides gimli
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkg_roxygenise.R
\name{pkg_roxygenise}
\alias{pkg_roxygenise}
\title{roxygenize a source package if needed}
\usage{
pkg_roxygenise(pkg_path, force = FALSE, quiet = FALSE, ...)
}
\arguments{
\item{pkg_path}{the package path, as a character}

\item{force}{if force(d), do not use the md5-based system to detect package changes}

\item{quiet}{whether to be quiet/silent}

\item{...}{passed to  \code{\link[devtools:document]{devtools::document()}}}
}
\value{
if the roxygenation has been performed
}
\description{
\itemize{
\item if the package has not changed (based on the md5sum file), does nothing
\item otherwise roxygenise the package using roxygen2::roxygenise
\item and update and save the new md5sum file
}
}
\details{
\itemize{
\item N.B: has the side-effect of loading the package
}
}
\examples{
 pkg <- setup_and_get_dummy_srcpkg()
 pkg_roxygenise(pkg$path)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/examples.R
\name{setup_and_get_dummy_srcpkg}
\alias{setup_and_get_dummy_srcpkg}
\title{installs the dummy srcpkg in a temp location}
\usage{
setup_and_get_dummy_srcpkg(dest = tempfile())
}
\arguments{
\item{dest}{where to install the dummy srcpkg}
}
\value{
the package as a \code{srcpkg} object
}
\description{
Intended for testing and to write examples
}
\examples{
pkg <- setup_and_get_dummy_srcpkg()
print(pkg)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{frodo}
\alias{frodo}
\title{provides frodo}
\usage{
frodo()
}
\value{
as a list
}
\description{
provides frodo
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{bilbo}
\alias{bilbo}
\title{provides bilbo}
\usage{
bilbo()
}
\value{
as a list
}
\description{
provides bilbo
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgs_install.R
\name{pkgs_install}
\alias{pkgs_install}
\title{installs a list of source packages}
\usage{
pkgs_install(
  pkgids,
  lib,
  src_pkgs = get_srcpkgs(),
  only_deps = FALSE,
  quiet = TRUE,
  ...
)
}
\arguments{
\item{pkgids}{a list of package ids (names, paths or object), or a
srcpkgs object. Also accept a singleton package object}

\item{lib}{directory where to install and find installed pkgs}

\item{src_pkgs}{a collection of source packages as a \code{srckgs} object.}

\item{only_deps}{whether not to include \code{pkgids}, only their dependencies.}

\item{quiet}{whether to be quiet/silent}

\item{...}{passed to \code{devtools::install()}}
}
\value{
the names of the packages actually installed
}
\description{
\itemize{
\item A source package can not be installed if its dependencies are not.
\item Will not reinstall packages if they are up-to-date
\item will roxygenise packages if needed
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/srcpkgs-package.R
\docType{package}
\name{srcpkgs-package}
\alias{srcpkgs-package}
\title{srcpkgs: R Source Packages Manager}
\description{
Manage a collection/library of R source packages. Discover, document, load, test source packages. Enable to use those packages as if they were actually installed. Quickly reload only what is needed on source code change. Run tests and checks in parallel.
}
\section{Features}{


\code{srcpkgs} main objective is to ease development on any project
that uses a collection of R source packages (a library).
It is able to figure out which dependencies are source packages, and is able
to quickly detect changes in any of the used source packages.
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/kforner/srcpkgs}
  \item Report bugs at \url{https://github.com/kforner/srcpkgs/issues}
}

}
\author{
\strong{Maintainer}: Karl Forner \email{karl.forner@gmail.com} [copyright holder]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hack_r_loaders.R
\name{hack_r_loaders}
\alias{hack_r_loaders}
\title{instruments the R loaders to make them aware of source packages}
\usage{
hack_r_loaders()
}
\value{
no return value, called for side-effects
}
\description{
hacks \code{library()} and \code{loadNamespace()} using the base R tracer function \code{trace()}.
\code{library(pkg)} will basically call \code{pkg_load(pkg)} if the source package \code{pkg}
is managed by \strong{srcpkgs}
}
\details{
N.B: usually you do not need to call that function explicitly. The function is reentrant.
}
\section{Package startup}{


At package startup (actually \code{.OnAttach()}), \code{hack_r_loaders()} will be automatically called to hack
the R loaders, UNLESS this is inhibited via the option \code{srcpkgs.inhibit_r_loaders_hack} or the
environment variable \code{SRCPKGS.INHIBIT_R_LOADERS_HACK}. You may set any value like TRUE, "TRUE", 1 or "1".
}

\examples{
\dontrun{
# hack library
hack_r_loaders()

# unhack
unhack_r_loaders()

# prevent automatic hacking when srcpkgs is loaded
options(srcpkgs.inhibit_r_loaders_hack=TRUE)
# or
Sys.setenv(SRCPKGS.INHIBIT_R_LOADERS_HACK="1")
library(srcpkgs)
}
}
\seealso{
\code{\link[=unhack_r_loaders]{unhack_r_loaders()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find_srcpkgs.R
\name{find_srcpkgs}
\alias{find_srcpkgs}
\title{finds all available source packages starting from the project root}
\usage{
find_srcpkgs(
  root = get_project_root(),
  srcpkgs_paths = find_srcpkgs_paths(root, prune = prune),
  prune = TRUE
)
}
\arguments{
\item{root}{directory from where to search for source packages}

\item{srcpkgs_paths}{paths to the source packages folders}

\item{prune}{whether to report packages contained inside another package (e.g. in tests/)}
}
\value{
a "srcpkgs" object (or NULL if none found), a named list of "srcpkg" objects, that essentially are
devtools "package" objects. The list is named after the package names.
}
\description{
N.B: the \emph{hidden} files and directories are ignored.
In general, this function is not used directly, instead you should use \code{\link[=get_srcpkgs]{get_srcpkgs()}}
}
\examples{
pkg <- setup_and_get_dummy_srcpkg()
pkgs <- find_srcpkgs(dirname(pkg$path))
print(pkgs)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find_srcpkgs.R
\name{get_srcpkgs}
\alias{get_srcpkgs}
\title{get the current source packages list}
\usage{
get_srcpkgs(filter = NULL)
}
\arguments{
\item{filter}{a pattern to filter the source packages}
}
\value{
the source packages as a "scrpkgs" object, cf \code{\link[=find_srcpkgs]{find_srcpkgs()}}, or NULL if none
}
\description{
The first call to this function will trigger the initialization of the package ((cf \code{\link[=reset]{reset()}}).
Since it is used by mostly all user-facing load-related functions, this enables a runtime initialization,
as opposed to a load-time initialization. So for example
you may load \code{srcpkgs}, then change the current directory to your project.
Then the first load will setup the settings.
}
\details{
For optimization, the paths to discovered source packages are cached (cf \code{\link[=reset]{reset()}} and \code{\link[=settings]{settings()}}.
This function will reparse the DESCRIPTION for any change.
If you add or delete a source package, you must reset the source package paths using \code{\link[=reset]{reset()}}

This function is useful for troubleshooting, to understand what are the source packages discovered
and managed by \code{srcpkgs}
}
\examples{
# setup a srcpkg. We need reset because it is not discoverable from the current directory
pkg <- setup_and_get_dummy_srcpkg()
reset(dirname(pkg$path))

print(get_srcpkgs())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/0_params.R
\name{params}
\alias{params}
\title{shared params}
\arguments{
\item{dry_run}{whether not to actually execute any action having side-effects}

\item{lib}{directory where to install and find installed pkgs}

\item{pkg}{a package as a "srcpkg" object}

\item{pkgs}{packages as a "srcpkgs" object}

\item{pkg_name}{the package name, as a character}

\item{pkg_or_name}{a package name or object ("package" or "srcpkg")}

\item{pkg_path}{the package path, as a character}

\item{pkgid}{a package name, path or package object}

\item{pkgids}{a list of package ids (names, paths or object), or a
srcpkgs object. Also accept a singleton package object}

\item{md5}{the MD5 hash of the source package}

\item{progress}{whether to display a progress bar}

\item{roxygen}{whether to roxygenize}

\item{src_pkgs}{a collection of source packages as a \code{srckgs} object.}

\item{srcpkgs_paths}{paths to the source packages folders}

\item{root}{directory from where to search for source packages}

\item{quiet}{whether to be quiet/silent}

\item{test_filter}{a pattern to select the testthat tests to run.
Test files are names test-xxxxx.R where xxxxx is the
test name. Only test files whose name match the pattern
will be run.}

\item{test_parallel}{whether to run the package tests in parallel}
}
\description{
shared params
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hack_r_loaders.R
\name{unhack_r_loaders}
\alias{unhack_r_loaders}
\title{untraces library() and loadNamespace()}
\usage{
unhack_r_loaders()
}
\value{
no return value, called for side-effects
}
\description{
The function is reentrant.
}
\examples{
\dontrun{
unhack_r_loaders()
}
}
\seealso{
\code{\link[=hack_r_loaders]{hack_r_loaders()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/srcpkgs.R
\name{srcpkgs}
\alias{srcpkgs}
\title{creates a new "srcpkgs" object}
\usage{
srcpkgs(pkgs = lapply(paths, devtools::as.package), paths = NULL)
}
\arguments{
\item{pkgs}{an existing srcpkgs object (no op), or a list of
source package-like objects}

\item{paths}{a list of source package paths as a character vector or list}
}
\value{
a \code{srcpkgs} object (a list named after the package names)
}
\description{
creates a new "srcpkgs" object
}
\examples{
# build dummy source packages
pkg1 <- setup_and_get_dummy_srcpkg()
pkg2 <- pkg1
pkg2$package <- "dummy.srcpkg2"

print(srcpkgs(list(pkg1, pkg2)))
print(srcpkgs(paths = pkg1$path))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgs_test.R
\name{pkgs_test}
\alias{pkgs_test}
\title{tests a list of source packages}
\usage{
pkgs_test(
  pkgids = names(filter_srcpkgs(src_pkgs, filter)),
  src_pkgs = get_srcpkgs(),
  filter = NULL,
  quiet = TRUE,
  ...
)
}
\arguments{
\item{pkgids}{a list of package ids (names, paths or object), or a
srcpkgs object. Also accept a singleton package object}

\item{src_pkgs}{a collection of source packages as a \code{srckgs} object.}

\item{filter}{filter out the packages to test using this pattern}

\item{quiet}{whether to be quiet/silent}

\item{...}{passed to \code{pkg_test}}
}
\value{
the results as a \code{pkgs_test} object
}
\description{
tests a list of source packages
}
\examples{
## create a dummy collection of srcpkgs by replicating the dummy srcpkg
 pkg <- setup_and_get_dummy_srcpkg()
 pkgs <- srcpkgs(list(pkg, pkg))

 res <- pkgs_test(pkgs, error_on = "never")
 print(res)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkg_utils.R
\name{pkg_list_attached}
\alias{pkg_list_attached}
\title{lists the packages that are attached, i.e. present in the R search() path}
\usage{
pkg_list_attached()
}
\value{
the names of attached package name as a character vector
}
\description{
lists the packages that are attached, i.e. present in the R search() path
}
\examples{
print(sort(pkg_list_attached()))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{gandalf}
\alias{gandalf}
\title{provides gandalf}
\usage{
gandalf()
}
\value{
as a list
}
\description{
provides gandalf
}
